/*
 * Copyright (C) 2008 Pleyo.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials http://gaara-fr.com/index2.phpprovided with the distribution.
 * 3.  Neither the name of Pleyo nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PLEYO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL PLEYO OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "PlatformKeyboardEvent.h"

#include "KeyboardCodes.h"
#include "BALBase.h"
#include "Logging.h"
#include "TextEncoding.h"
#include "Unicode.h"
#include "CString.h"

#include <directfb_keyboard.h>

#include <stdio.h>

namespace WKAL {

typedef struct {
    int           keyCode;
    int           keyText;
	const char*   keyDesc;
    int           keyDFB;
    const char*   nameDFB;
} KeymapEntry;

static KeymapEntry keycodeMap[] =
{
     { VK_A, 'a', "A", DIKI_A, "A" },
     { VK_B, 'b', "B", DIKI_B, "B" }, 
     { VK_C, 'c', "C", DIKI_C, "C" },
     { VK_D, 'd', "D", DIKI_D, "D" }, 
     { VK_E, 'e', "E", DIKI_E, "E" }, 
     { VK_F, 'f', "F", DIKI_F, "F" }, 
     { VK_G, 'g', "G", DIKI_G, "G" }, 
     { VK_H, 'h', "H", DIKI_H, "H" }, 
     { VK_I, 'i', "I", DIKI_I, "I" }, 
     { VK_J, 'j', "J", DIKI_J, "J" }, 
     { VK_K, 'k', "K", DIKI_K, "K" }, 
     { VK_L, 'l', "L", DIKI_L, "L" },
     { VK_M, 'm', "M", DIKI_M, "M" },
     { VK_N, 'n', "N", DIKI_N, "N" },
     { VK_O, 'o', "O", DIKI_O, "O" },
     { VK_P, 'p', "P", DIKI_P, "P" },
     { VK_Q, 'q', "Q", DIKI_Q, "Q" },
     { VK_R, 'r', "R", DIKI_R, "R" },
     { VK_S, 's', "S", DIKI_S, "S" },
     { VK_T, 't', "T", DIKI_T, "T" },
     { VK_U, 'u', "U", DIKI_U, "U" },
     { VK_V, 'v', "V", DIKI_V, "V" },
     { VK_W, 'w', "W", DIKI_W, "W" },
     { VK_X, 'x', "X", DIKI_X, "X" },
     { VK_Y, 'y', "Y", DIKI_Y, "Y" },
     { VK_Z, 'z', "Z", DIKI_Z, "Z" },
     { VK_0, '0', "0", DIKI_0, "0" },
     { VK_1, '1', "1", DIKI_1, "1" },
     { VK_2, '2', "2", DIKI_2, "2" },
     { VK_3, '3', "3", DIKI_3, "3" },
     { VK_4, '4', "4", DIKI_4, "4" },
     { VK_5, '5', "5", DIKI_5, "5" },
     { VK_6, '6', "6", DIKI_6, "6" },
     { VK_7, '7', "7", DIKI_7, "7" },
     { VK_8, '8', "8", DIKI_8, "8" },
     { VK_9, '9', "9", DIKI_9, "9" },
     { VK_F1, 0, "F1", DIKI_F1, "F1" },
     { VK_F2, 0, "F2", DIKI_F2, "F2" },
     { VK_F3, 0, "F3", DIKI_F3, "F3" },
     { VK_F4, 0, "F4", DIKI_F4, "F4" },
     { VK_F5, 0, "F5", DIKI_F5, "F5" },
     { VK_F6, 0, "F6", DIKI_F6, "F6" },
     { VK_F7, 0, "F7", DIKI_F7, "F7" },
     { VK_F8, 0, "F8", DIKI_F8, "F8" },
     { VK_F9, 0, "F9", DIKI_F9, "F9" },
     { VK_F10, 0, "F10", DIKI_F10, "F10" },
     { VK_F11, 0, "F11", DIKI_F11, "F11" },
     { VK_F12, 0, "F12", DIKI_F12, "F12" },
     { VK_SHIFT, 0, "Shift", DIKI_SHIFT_L, "SHIFT_L" },
     { VK_SHIFT, 0, "Shift", DIKI_SHIFT_R, "SHIFT_R" },
     { VK_CONTROL, 0, "Ctrl", DIKI_CONTROL_L, "CONTROL_L" },
     { VK_CONTROL, 0, "Ctrl", DIKI_CONTROL_R, "CONTROL_R" },
     { VK_MENU, 0, "Alt", DIKI_ALT_L, "ALT_L" },
     { VK_MENU, 0, "Alt", DIKI_ALT_R, "ALT_R" },
     { VK_MENU, 0, "Meta", DIKI_META_L, "META_L" },
     { VK_MENU, 0, "Meta", DIKI_META_R, "META_R" },
     { VK_MENU, 0, "Super", DIKI_SUPER_L, "SUPER_L" },
     { VK_MENU, 0, "Super", DIKI_SUPER_R, "SUPER_R" },
     { VK_MENU, 0, "Hyper", DIKI_HYPER_L, "HYPER_L" },
     { VK_MENU, 0, "Hyper", DIKI_HYPER_R, "HYPER_R" },
     { VK_CAPITAL, 0, "CAPS_LOCK", DIKI_CAPS_LOCK, "CAPS_LOCK" },
     { VK_NUMLOCK, 0, "NUM_LOCK", DIKI_NUM_LOCK, "NUM_LOCK" },
     { VK_SCROLL,  0, "SCROLL_LOCK", DIKI_SCROLL_LOCK, "SCROLL_LOCK" },
     { VK_ESCAPE,  0, "U+001B", DIKI_ESCAPE, "ESCAPE" },
     { VK_LEFT,    0, "Left", DIKI_LEFT, "LEFT" },
     { VK_RIGHT,   0, "Right", DIKI_RIGHT, "RIGHT" },
     { VK_UP,      0, "Up", DIKI_UP, "UP" },
     { VK_DOWN,    0, "Down", DIKI_DOWN, "DOWN" },
     { VK_TAB,     0, "U+0009", DIKI_TAB, "TAB" },
     { VK_RETURN,  '\r', "Enter", DIKI_ENTER, "ENTER" },
     { VK_SPACE,   ' ', "U+0020", DIKI_SPACE, "SPACE" },
     { VK_BACK,    0, "Back", DIKI_BACKSPACE, "BACKSPACE" },
     { VK_INSERT,  0, "Insert", DIKI_INSERT, "INSERT" },
     { VK_DELETE,  0, "Delete", DIKI_DELETE, "DELETE" },
     { VK_HOME,    0, "Home", DIKI_HOME, "HOME" },
     { VK_END,     0, "End", DIKI_END, "END" },
     { VK_PRIOR,   0, "PageUp", DIKI_PAGE_UP, "PAGE_UP" },
     { VK_NEXT,    0, "PageDown", DIKI_PAGE_DOWN, "PAGE_DOWN" },
     { VK_PRINT,   0, "PrintScreen", DIKI_PRINT, "PRINT" },
     { VK_PAUSE,   0, "Pause", DIKI_PAUSE, "PAUSE" },
     { VK_UNKNOWN,  0, "Quote Left", DIKI_QUOTE_LEFT, "QUOTE_LEFT" },
     { VK_OEM_MINUS, 0, "Minus", DIKI_MINUS_SIGN, "MINUS_SIGN" },
     { VK_UNKNOWN,  0,   "Equals", DIKI_EQUALS_SIGN, "EQUALS_SIGN" },
     { VK_UNKNOWN,  0, "Open Bracket", DIKI_BRACKET_LEFT, "BRACKET_LEFT" },
     { VK_UNKNOWN,  0, "Close Bracket", DIKI_BRACKET_RIGHT, "BRACKET_RIGHT" },
     { VK_UNKNOWN, 0, "Black Slash", DIKI_BACKSLASH, "BACKSLASH" },
     { VK_UNKNOWN, 0, "Semi Colon", DIKI_SEMICOLON, "SEMICOLON" },
     { VK_UNKNOWN, 0, "Quote Right", DIKI_QUOTE_RIGHT, "QUOTE_RIGHT" },
     { VK_OEM_COMMA, ',', "Comma", DIKI_COMMA, "COMMA" },
     { VK_OEM_PERIOD, '.', "Period", DIKI_PERIOD, "PERIOD" },
     { VK_UNKNOWN, 0, "Slash", DIKI_SLASH, "SLASH" },
     { VK_UNKNOWN, 0, "Less", DIKI_LESS_SIGN, "LESS_SIGN" },
     { VK_UNKNOWN, 0, "Divide", DIKI_KP_DIV, "KP_DIV" },
     { VK_UNKNOWN, 0, "Multiply", DIKI_KP_MULT, "KP_MULT" },
     { VK_UNKNOWN, 0, "Sub", DIKI_KP_MINUS, "KP_MINUS" },
     { VK_UNKNOWN, 0, "Add", DIKI_KP_PLUS, "KP_PLUS" },
     { VK_RETURN, '\r', "Enter", DIKI_KP_ENTER, "KP_ENTER" },
     { VK_SPACE, ' ', "U+0020", DIKI_KP_SPACE, "KP_SPACE" },
     { VK_TAB, 0, "U+0009", DIKI_KP_TAB, "KP_TAB" },
     { VK_F1, 0, "F1", DIKI_KP_F1, "KP_F1" },
     { VK_F2, 0, "F2", DIKI_KP_F2, "KP_F2" },
     { VK_F3, 0, "F3", DIKI_KP_F3, "KP_F3" },
     { VK_F4, 0, "F4", DIKI_KP_F4, "KP_F4" },
     { VK_UNKNOWN, 0, "Equals", DIKI_KP_EQUAL, "KP_EQUAL" },
     { VK_UNKNOWN, '/', "Separater", DIKI_KP_SEPARATOR, "KP_SEPARATOR" },
     { VK_DECIMAL, '.', "Decimal", DIKI_KP_DECIMAL, "KP_DECIMAL" },
     { VK_0, '0', "0", DIKI_KP_0, "KP_0" },
     { VK_1, '1', "1", DIKI_KP_1, "KP_1" },
     { VK_2, '2', "2", DIKI_KP_2, "KP_2" },
     { VK_3, '3', "3", DIKI_KP_3, "KP_3" },
     { VK_4, '4', "4", DIKI_KP_4, "KP_4" },
     { VK_5, '5', "5", DIKI_KP_5, "KP_5" },
     { VK_6, '6', "6", DIKI_KP_6, "KP_6" },
     { VK_7, '7', "7", DIKI_KP_7, "KP_7" },
     { VK_8, '8', "8", DIKI_KP_8, "KP_8" },
     { VK_9, '9', "9", DIKI_KP_9, "KP_9" },
     { VK_UNKNOWN, 0, "VK_UNKNOWN", DIKI_UNKNOWN, "UNKNOWN" }
};

static KeymapEntry keysymbolMap[] =
{
     { VK_BACK, 0, "Back", DIKS_BACKSPACE, "BACKSPACE" },
     { VK_TAB,  0, "U+0009", DIKS_TAB, "TAB" },
     { VK_RETURN, '\r', "Enter", DIKS_RETURN, "RETURN" },
     { VK_UNKNOWN, 0, "Cancel", DIKS_CANCEL, "CANCEL" },
     { VK_ESCAPE, 0, "U+001B", DIKS_ESCAPE, "ESCAPE" },
     { VK_SPACE, ' ', "U+0020", DIKS_SPACE, "SPACE" },
     { VK_UNKNOWN, 0, "Exclamation Mark", DIKS_EXCLAMATION_MARK, "EXCLAMATION_MARK" },
     { VK_UNKNOWN, 0, "Quotation", DIKS_QUOTATION, "QUOTATION" },
     { VK_UNKNOWN, 0, "Number Sign", DIKS_NUMBER_SIGN, "NUMBER_SIGN" },
     { VK_UNKNOWN, '$', "Dollar", DIKS_DOLLAR_SIGN, "DOLLAR_SIGN" },
     { VK_UNKNOWN, '%', "Percent Sign", DIKS_PERCENT_SIGN, "PERCENT_SIGN" },
     { VK_UNKNOWN, 0, "Ampersand", DIKS_AMPERSAND, "AMPERSAND" },
     { VK_UNKNOWN, 0, "Apostrophe", DIKS_APOSTROPHE, "APOSTROPHE" },
     { VK_UNKNOWN, '(', "Parenthesis Left", DIKS_PARENTHESIS_LEFT, "PARENTHESIS_LEFT" },
     { VK_UNKNOWN, ')', "Parenthesis Right", DIKS_PARENTHESIS_RIGHT, "PARENTHESIS_RIGHT" },
     { VK_UNKNOWN, '*', "Asterisk", DIKS_ASTERISK, "ASTERISK" },
     { VK_UNKNOWN, '+', "Add", DIKS_PLUS_SIGN, "PLUS_SIGN" },
     { VK_UNKNOWN, ',', "Comma", DIKS_COMMA, "COMMA" },
     { VK_UNKNOWN, '-', "Minus", DIKS_MINUS_SIGN, "MINUS_SIGN" },
     { VK_OEM_PERIOD, '.', "Period", DIKS_PERIOD, "PERIOD" },
     { VK_UNKNOWN, 0, "Slash", DIKS_SLASH, "SLASH" },
     { VK_0, '0', "0", DIKS_0, "0" },
     { VK_1, '1', "1", DIKS_1, "1" },
     { VK_2, '2', "2", DIKS_2, "2" },
     { VK_3, '3', "3", DIKS_3, "3" },
     { VK_4, '4', "4", DIKS_4, "4" },
     { VK_5, '5', "5", DIKS_5, "5" },
     { VK_6, '6', "6", DIKS_6, "6" },
     { VK_7, '7', "7", DIKS_7, "7" },
     { VK_8, '8', "8", DIKS_8, "8" },
     { VK_9, '9', "9", DIKS_9, "9" },
     { VK_UNKNOWN, 0, "Colon", DIKS_COLON, "COLON" },
     { VK_UNKNOWN, 0, "Semicolon", DIKS_SEMICOLON, "SEMICOLON" },
     { VK_UNKNOWN, '<', "Less Than Sign", DIKS_LESS_THAN_SIGN, "LESS_THAN_SIGN" },
     { VK_UNKNOWN, '=', "Equals Sign", DIKS_EQUALS_SIGN, "EQUALS_SIGN" },
     { VK_UNKNOWN, '>', "Greater Than Sign", DIKS_GREATER_THAN_SIGN, "GREATER_THAN_SIGN" },
     { VK_UNKNOWN, '?', "Question Mark", DIKS_QUESTION_MARK, "QUESTION_MARK" },
     { VK_UNKNOWN, '@', "At", DIKS_AT, "AT" },
     { VK_A, 'A', "A", DIKS_CAPITAL_A, "CAPITAL_A" },
     { VK_B, 'B', "B", DIKS_CAPITAL_B, "CAPITAL_B" },
     { VK_C, 'C', "C", DIKS_CAPITAL_C, "CAPITAL_C" },
     { VK_D, 'D', "D", DIKS_CAPITAL_D, "CAPITAL_D" },
     { VK_E, 'E', "E", DIKS_CAPITAL_E, "CAPITAL_E" },
     { VK_F, 'F', "F", DIKS_CAPITAL_F, "CAPITAL_F" },
     { VK_G, 'G', "G", DIKS_CAPITAL_G, "CAPITAL_G" },
     { VK_H, 'H', "H", DIKS_CAPITAL_H, "CAPITAL_H" },
     { VK_I, 'I', "I", DIKS_CAPITAL_I, "CAPITAL_I" },
     { VK_J, 'J', "J", DIKS_CAPITAL_J, "CAPITAL_J" },
     { VK_K, 'K', "K", DIKS_CAPITAL_K, "CAPITAL_K" },
     { VK_L, 'L', "L", DIKS_CAPITAL_L, "CAPITAL_L" },
     { VK_M, 'M', "M", DIKS_CAPITAL_M, "CAPITAL_M" },
     { VK_N, 'N', "N", DIKS_CAPITAL_N, "CAPITAL_N" },
     { VK_O, 'O', "O", DIKS_CAPITAL_O, "CAPITAL_O" },
     { VK_P, 'P', "P", DIKS_CAPITAL_P, "CAPITAL_P" },
     { VK_Q, 'Q', "Q", DIKS_CAPITAL_Q, "CAPITAL_Q" },
     { VK_R, 'R', "R", DIKS_CAPITAL_R, "CAPITAL_R" },
     { VK_S, 'S', "S", DIKS_CAPITAL_S, "CAPITAL_S" },
     { VK_T, 'T', "T", DIKS_CAPITAL_T, "CAPITAL_T" },
     { VK_U, 'U', "U", DIKS_CAPITAL_U, "CAPITAL_U" },
     { VK_V, 'V', "V", DIKS_CAPITAL_V, "CAPITAL_V" },
     { VK_W, 'W', "W", DIKS_CAPITAL_W, "CAPITAL_W" },
     { VK_X, 'X', "X", DIKS_CAPITAL_X, "CAPITAL_X" },
     { VK_Y, 'Y', "Y", DIKS_CAPITAL_Y, "CAPITAL_Y" },
     { VK_Z, 'Z', "Z", DIKS_CAPITAL_Z, "CAPITAL_Z" },
     { VK_UNKNOWN, '[', "Square Bracket Left", DIKS_SQUARE_BRACKET_LEFT, "SQUARE_BRACKET_LEFT" },
     { VK_UNKNOWN, '\\', "Back Slash", DIKS_BACKSLASH, "BACKSLASH" },
     { VK_UNKNOWN, ']', "Square Bracket Right", DIKS_SQUARE_BRACKET_RIGHT, "SQUARE_BRACKET_RIGHT" },
     { VK_UNKNOWN, 0, "Circumflex Accent", DIKS_CIRCUMFLEX_ACCENT, "CIRCUMFLEX_ACCENT" },
     { VK_UNKNOWN, '_', "Underscore", DIKS_UNDERSCORE, "UNDERSCORE" },
     { VK_UNKNOWN, 0, "Grave Accent", DIKS_GRAVE_ACCENT, "GRAVE_ACCENT" },
     { VK_A, 'a', "A", DIKS_SMALL_A, "SMALL_A" },
     { VK_B, 'b', "B", DIKS_SMALL_B, "SMALL_B" },
     { VK_C, 'c', "C", DIKS_SMALL_C, "SMALL_C" },
     { VK_D, 'd', "D", DIKS_SMALL_D, "SMALL_D" },
     { VK_E, 'e', "E", DIKS_SMALL_E, "SMALL_E" },
     { VK_F, 'f', "F", DIKS_SMALL_F, "SMALL_F" },
     { VK_G, 'g', "G", DIKS_SMALL_G, "SMALL_G" },
     { VK_H, 'h', "H", DIKS_SMALL_H, "SMALL_H" },
     { VK_I, 'i', "I", DIKS_SMALL_I, "SMALL_I" },
     { VK_J, 'j', "J", DIKS_SMALL_J, "SMALL_J" },
     { VK_K, 'k', "K", DIKS_SMALL_K, "SMALL_K" },
     { VK_L, 'l', "L", DIKS_SMALL_L, "SMALL_L" },
     { VK_M, 'm', "M", DIKS_SMALL_M, "SMALL_M" },
     { VK_N, 'n', "N", DIKS_SMALL_N, "SMALL_N" },
     { VK_O, 'o', "O", DIKS_SMALL_O, "SMALL_O" },
     { VK_P, 'p', "P", DIKS_SMALL_P, "SMALL_P" },
     { VK_Q, 'q', "Q", DIKS_SMALL_Q, "SMALL_Q" },
     { VK_R, 'r', "R", DIKS_SMALL_R, "SMALL_R" },
     { VK_S, 's', "S", DIKS_SMALL_S, "SMALL_S" },
     { VK_T, 't', "T", DIKS_SMALL_T, "SMALL_T" },
     { VK_U, 'u', "U", DIKS_SMALL_U, "SMALL_U" },
     { VK_V, 'v', "V", DIKS_SMALL_V, "SMALL_V" },
     { VK_W, 'w', "W", DIKS_SMALL_W, "SMALL_W" },
     { VK_X, 'x', "X", DIKS_SMALL_X, "SMALL_X" },
     { VK_Y, 'y', "Y", DIKS_SMALL_Y, "SMALL_Y" },
     { VK_Z, 'z', "Z", DIKS_SMALL_Z, "SMALL_Z" },
     { VK_UNKNOWN, 0, "Curly Bracket Left", DIKS_CURLY_BRACKET_LEFT, "CURLY_BRACKET_LEFT" },
     { VK_UNKNOWN, 0, "Vertical Bar", DIKS_VERTICAL_BAR, "VERTICAL_BAR" },
     { VK_UNKNOWN, 0, "Curly Bracket Right", DIKS_CURLY_BRACKET_RIGHT, "CURLY_BRACKET_RIGHT" },
     { VK_UNKNOWN,  0, "Tilde", DIKS_TILDE, "TILDE" },
     { VK_DELETE,   0, "Delete", DIKS_DELETE, "DELETE" },
     { VK_LEFT,     0, "Left", DIKS_CURSOR_LEFT, "CURSOR_LEFT" },
     { VK_RIGHT,    0, "Right", DIKS_CURSOR_RIGHT, "CURSOR_RIGHT" },
     { VK_UP,       0, "Up", DIKS_CURSOR_UP, "CURSOR_UP" },
     { VK_DOWN,     0, "Down", DIKS_CURSOR_DOWN, "CURSOR_DOWN" },
     { VK_INSERT, 0, "Insert", DIKS_INSERT, "INSERT" },
     { VK_HOME,   0, "Home", DIKS_HOME, "HOME" },
     { VK_END,    0, "End", DIKS_END, "END" },
     { VK_PRIOR,  0, "PageUp", DIKS_PAGE_UP, "PAGE_UP" },
     { VK_NEXT,   0, "PageDown", DIKS_PAGE_DOWN, "PAGE_DOWN" },
     { VK_PRINT,  0, "PrintScreen", DIKS_PRINT, "PRINT" },
     { VK_PAUSE,  0, "Pause", DIKS_PAUSE, "PAUSE" },
     { VK_UNKNOWN, 0, "Ok", DIKS_OK, "OK" },
     { VK_RETURN, '\r', "Enter", DIKS_SELECT, "SELECT" },
     { VK_UNKNOWN, 0, "Goto", DIKS_GOTO, "GOTO" },
     { VK_UNKNOWN, 0, "Clear", DIKS_CLEAR, "CLEAR" },
     { VK_UNKNOWN, 0, "Power", DIKS_POWER, "POWER" },
     { VK_UNKNOWN, 0, "Power2", DIKS_POWER2, "POWER2" },
     { VK_UNKNOWN, 0, "Option", DIKS_OPTION, "OPTION" },
     { VK_MENU,   0, "Menu", DIKS_MENU, "MENU" },
     { VK_UNKNOWN, 0, "Help", DIKS_HELP, "HELP" },
     { VK_UNKNOWN, 0, "Info", DIKS_INFO, "INFO" },
     { VK_UNKNOWN, 0, "Time", DIKS_TIME, "TIME" },
     { VK_UNKNOWN, 0, "Vendor", DIKS_VENDOR, "VENDOR" },
     { VK_UNKNOWN, 0, "Archive", DIKS_ARCHIVE, "ARCHIVE" },
     { VK_UNKNOWN, 0, "Program", DIKS_PROGRAM, "PROGRAM" },
     { VK_UNKNOWN, 0, "Channel", DIKS_CHANNEL, "CHANNEL" },
     { VK_UNKNOWN, 0, "Favorites", DIKS_FAVORITES, "FAVORITES" },
     { VK_UNKNOWN, 0, "EPG", DIKS_EPG, "EPG" },
     { VK_UNKNOWN, 0, "PVR", DIKS_PVR, "PVR" },
     { VK_UNKNOWN, 0, "MHP", DIKS_MHP, "MHP" },
     { VK_UNKNOWN, 0, "Language", DIKS_LANGUAGE, "LANGUAGE" },
     { VK_F6, 0, "TITLE", DIKS_TITLE, "TITLE" },
     { VK_UNKNOWN, 0, "SUBTITLE", DIKS_SUBTITLE, "SUBTITLE" },
     { VK_UNKNOWN, 0, "ANGLE", DIKS_ANGLE, "ANGLE" },
     { VK_UNKNOWN, 0, "ZOOM", DIKS_ZOOM, "ZOOM" },
     { VK_UNKNOWN, 0, "MODE", DIKS_MODE, "MODE" },
     { VK_UNKNOWN, 0, "KEYBOARD", DIKS_KEYBOARD, "KEYBOARD" },
     { VK_UNKNOWN, 0, "PC", DIKS_PC, "PC" },
     { VK_UNKNOWN, 0, "SCREEN", DIKS_SCREEN, "SCREEN" },
     { VK_UNKNOWN, 0, "TV",  DIKS_TV, "TV" },
     { VK_UNKNOWN, 0, "TV2", DIKS_TV2, "TV2" },
     { VK_UNKNOWN, 0, "VCR", DIKS_VCR, "VCR" },
     { VK_UNKNOWN, 0, "VCR2", DIKS_VCR2, "VCR2" },
     { VK_UNKNOWN, 0, "SAT", DIKS_SAT, "SAT" },
     { VK_UNKNOWN, 0, "SAT2", DIKS_SAT2, "SAT2" },
     { VK_UNKNOWN, 0, "CD", DIKS_CD, "CD" },
     { VK_UNKNOWN, 0, "TAPE", DIKS_TAPE, "TAPE" },
     { VK_UNKNOWN, 0, "RADIO", DIKS_RADIO, "RADIO" },
     { VK_UNKNOWN, 0, "TUNER", DIKS_TUNER, "TUNER" },
     { VK_UNKNOWN, 0, "PLAYER", DIKS_PLAYER, "PLAYER" },
     { VK_UNKNOWN, 0, "TEXT", DIKS_TEXT, "TEXT" },
     { VK_UNKNOWN, 0, "DVD", DIKS_DVD, "DVD" },
     { VK_UNKNOWN, 0, "AUX", DIKS_AUX, "AUX" },
     { VK_UNKNOWN, 0, "MP3", DIKS_MP3, "MP3" },
     { VK_UNKNOWN, 0, "PHONE", DIKS_PHONE, "PHONE" },
     { VK_UNKNOWN, 0, "AUDIO", DIKS_AUDIO, "AUDIO" },
     { VK_UNKNOWN, 0, "VIDEO", DIKS_VIDEO, "VIDEO" },
     { VK_UNKNOWN, 0, "INTERNET", DIKS_INTERNET, "INTERNET" },
     { VK_UNKNOWN, 0, "MAIL", DIKS_MAIL, "MAIL" },
     { VK_UNKNOWN, 0, "NEWS", DIKS_NEWS, "NEWS" },
     { VK_UNKNOWN, 0, "DIRECTORY", DIKS_DIRECTORY, "DIRECTORY" },
     { VK_UNKNOWN, 0, "LIST", DIKS_LIST, "LIST" },
     { VK_UNKNOWN, 0, "CALCULATOR", DIKS_CALCULATOR, "CALCULATOR" },
     { VK_UNKNOWN, 0, "MEMO", DIKS_MEMO, "MEMO" },
     { VK_UNKNOWN, 0, "CALENDAR", DIKS_CALENDAR, "CALENDAR" },
     { VK_UNKNOWN, 0, "EDITOR", DIKS_EDITOR, "EDITOR" },
     { VK_UNKNOWN, 0, "RED", DIKS_RED, "RED" },
     { VK_UNKNOWN, 0, "GREEN", DIKS_GREEN, "GREEN" },
     { VK_UNKNOWN, 0, "RELLOW", DIKS_YELLOW, "YELLOW" },
     { VK_UNKNOWN, 0, "BLUE", DIKS_BLUE, "BLUE" },
     { VK_UNKNOWN, 0, "CHANNEL_UP", DIKS_CHANNEL_UP, "CHANNEL_UP" },
     { VK_UNKNOWN, 0, "CHANNEL_DOWN", DIKS_CHANNEL_DOWN, "CHANNEL_DOWN" },
     { VK_BROWSER_BACK, 0, "BROWSER_BACK", DIKS_BACK, "BACK" },
     { VK_UNKNOWN, 0, "FORWARD", DIKS_FORWARD, "FORWARD" },
     { VK_UNKNOWN, 0, "FIRST", DIKS_FIRST, "FIRST" },
     { VK_UNKNOWN, 0, "LAST", DIKS_LAST, "LAST" },
     { VK_UNKNOWN, 0, "VOLUME_UP", DIKS_VOLUME_UP, "VOLUME_UP" },
     { VK_UNKNOWN, 0, "VOLUME_DOWN", DIKS_VOLUME_DOWN, "VOLUME_DOWN" },
     { VK_UNKNOWN, 0, "MUTE", DIKS_MUTE, "MUTE" },
     { VK_UNKNOWN, 0, "AB", DIKS_AB, "AB" },
     { VK_UNKNOWN, 0, "PLAYPAUSE", DIKS_PLAYPAUSE, "PLAYPAUSE" },
     { VK_UNKNOWN, 0, "PLAY", DIKS_PLAY, "PLAY" },
     { VK_UNKNOWN, 0, "STOP", DIKS_STOP, "STOP" },
     { VK_UNKNOWN, 0, "RESTART", DIKS_RESTART, "RESTART" },
     { VK_UNKNOWN, 0, "SLOW", DIKS_SLOW, "SLOW" },
     { VK_UNKNOWN, 0, "FAST", DIKS_FAST, "FAST" },
     { VK_UNKNOWN, 0, "RECORD", DIKS_RECORD, "RECORD" },
     { VK_UNKNOWN, 0, "EJECT", DIKS_EJECT, "EJECT" },
     { VK_UNKNOWN, 0, "SHUFFLE", DIKS_SHUFFLE, "SHUFFLE" },
     { VK_UNKNOWN, 0, "REWIND", DIKS_REWIND, "REWIND" },
     { VK_UNKNOWN, 0, "FASTFORWARD", DIKS_FASTFORWARD, "FASTFORWARD" },
     { VK_UNKNOWN, 0, "PREVIOUS", DIKS_PREVIOUS, "PREVIOUS" },
     { VK_UNKNOWN, 0, "NEXT", DIKS_NEXT, "NEXT" },
     { VK_UNKNOWN, 0, "BEGIN", DIKS_BEGIN, "BEGIN" },
     { VK_UNKNOWN, 0, "DIGITS", DIKS_DIGITS, "DIGITS" },
     { VK_UNKNOWN, 0, "TEEN", DIKS_TEEN, "TEEN" },
     { VK_UNKNOWN, 0, "TWEN", DIKS_TWEN, "TWEN" },
     { VK_UNKNOWN, 0, "BREAK", DIKS_BREAK, "BREAK" },
     { VK_UNKNOWN, 0, "EXIT", DIKS_EXIT, "EXIT" },
     { VK_UNKNOWN, 0, "SETUP", DIKS_SETUP, "SETUP" },
     { VK_UNKNOWN, 0, "CursorLeftUp", DIKS_CURSOR_LEFT_UP, "CURSOR_LEFT_UP" },
     { VK_UNKNOWN, 0, "CursorLeftDown", DIKS_CURSOR_LEFT_DOWN, "CURSOR_LEFT_DOWN" },
     { VK_UNKNOWN, 0, "CursorUpRight", DIKS_CURSOR_UP_RIGHT, "CURSOR_UP_RIGHT" },
     { VK_UNKNOWN, 0, "CursorDownRight", DIKS_CURSOR_DOWN_RIGHT, "CURSOR_DOWN_RIGHT" },
     { VK_F1, 0, "F1", DIKS_F1, "F1" },
     { VK_F2, 0, "F2", DIKS_F2, "F2" },
     { VK_F3, 0, "F3", DIKS_F3, "F3" },
     { VK_F4, 0, "F4", DIKS_F4, "F4" },
     { VK_F5, 0, "F5", DIKS_F5, "F5" },
     { VK_F6, 0, "F6", DIKS_F6, "F6" },
     { VK_F7, 0, "F7", DIKS_F7, "F7" },
     { VK_F8, 0, "F8", DIKS_F8, "F8" },
     { VK_F9, 0, "F9", DIKS_F9, "F9" },
     { VK_F10, 0, "F10", DIKS_F10, "F10" },
     { VK_F11, 0, "F11", DIKS_F11, "F11" },
     { VK_F12, 0, "F12", DIKS_F12, "F12" },
     { VK_SHIFT, 0, "Shift", DIKS_SHIFT, "SHIFT" },
     { VK_CONTROL, 0, "Ctrl", DIKS_CONTROL, "CONTROL" },
     { VK_MENU, 0, "Alt", DIKS_ALT, "ALT" },
     { VK_UNKNOWN, 0, "ALTGR", DIKS_ALTGR, "ALTGR" },
     { VK_MENU, 0, "Meta", DIKS_META, "META" },
     { VK_UNKNOWN, 0, "SUPER", DIKS_SUPER, "SUPER" },
     { VK_UNKNOWN, 0, "HYPER", DIKS_HYPER, "HYPER" },
     { VK_CAPITAL, 0, "CAPS_LOCK", DIKS_CAPS_LOCK, "CAPS_LOCK" },
     { VK_NUMLOCK, 0, "NUM_LOCK", DIKS_NUM_LOCK, "NUM_LOCK" },
     { VK_SCROLL, 0, "SCROLL_LOCK", DIKS_SCROLL_LOCK, "SCROLL_LOCK" },
     { VK_UNKNOWN, 0, "DEAD_ABOVEDOT", DIKS_DEAD_ABOVEDOT, "DEAD_ABOVEDOT" },
     { VK_UNKNOWN, 0, "DEAD_ABOVRING", DIKS_DEAD_ABOVERING, "DEAD_ABOVERING" },
     { VK_UNKNOWN, 0, "DEAD_ACUTE", DIKS_DEAD_ACUTE, "DEAD_ACUTE" },
     { VK_UNKNOWN, 0, "DEAD_BREVE", DIKS_DEAD_BREVE, "DEAD_BREVE" },
     { VK_UNKNOWN, 0, "DEAD_CAROW", DIKS_DEAD_CARON, "DEAD_CARON" },
     { VK_UNKNOWN, 0, "DEAD_CEDILLA", DIKS_DEAD_CEDILLA, "DEAD_CEDILLA" },
     { VK_UNKNOWN, 0, "DEAD_CIRCUMFLEX", DIKS_DEAD_CIRCUMFLEX, "DEAD_CIRCUMFLEX" },
     { VK_UNKNOWN, 0, "DEAD_DIAERESIS", DIKS_DEAD_DIAERESIS, "DEAD_DIAERESIS" },
     { VK_UNKNOWN, 0, "DEAD_DOUBLEACUTE", DIKS_DEAD_DOUBLEACUTE, "DEAD_DOUBLEACUTE" },
     { VK_UNKNOWN, 0, "DEAD_GRAVE", DIKS_DEAD_GRAVE, "DEAD_GRAVE" },
     { VK_UNKNOWN, 0, "DEAD_IOTA", DIKS_DEAD_IOTA, "DEAD_IOTA" },
     { VK_UNKNOWN, 0, "DEAD_MACRON", DIKS_DEAD_MACRON, "DEAD_MACRON" },
     { VK_UNKNOWN, 0, "DEAD_OGONEK", DIKS_DEAD_OGONEK, "DEAD_OGONEK" },
     { VK_UNKNOWN, 0, "DEAD_SEMIVOICED_SOUND", DIKS_DEAD_SEMIVOICED_SOUND, "DEAD_SEMIVOICED_SOUND" },
     { VK_UNKNOWN, 0, "DEAD_TILDE", DIKS_DEAD_TILDE, "DEAD_TILDE" },
     { VK_UNKNOWN, 0, "DEAD_VOICED_SOUND", DIKS_DEAD_VOICED_SOUND, "DEAD_VOICED_SOUND" },
     { VK_UNKNOWN, 0, "CUSTOM0", DIKS_CUSTOM0, "CUSTOM0" },
     { VK_UNKNOWN, 0, "CUSTOM1", DIKS_CUSTOM1, "CUSTOM1" },
     { VK_UNKNOWN, 0, "CUSTOM2", DIKS_CUSTOM2, "CUSTOM2" },
     { VK_UNKNOWN, 0, "CUSTOM3", DIKS_CUSTOM3, "CUSTOM3" },
     { VK_UNKNOWN, 0, "CUSTOM4", DIKS_CUSTOM4, "CUSTOM4" },
     { VK_UNKNOWN, 0, "CUSTOM5", DIKS_CUSTOM5, "CUSTOM5" },
     { VK_UNKNOWN, 0, "CUSTOM6", DIKS_CUSTOM6, "CUSTOM6" },
     { VK_UNKNOWN, 0, "CUSTOM7", DIKS_CUSTOM7, "CUSTOM7" },
     { VK_UNKNOWN, 0, "CUSTOM8", DIKS_CUSTOM8, "CUSTOM8" },
     { VK_UNKNOWN, 0, "CUSTOM9", DIKS_CUSTOM9, "CUSTOM9" },
     { VK_UNKNOWN, 0, "DIKS_NULL", DIKS_NULL, "NULL" }
};

static int compare_keymap(const void *a, const void *b)
{
    int *key_dfb = (int*)a;
    KeymapEntry* keyEntry = (KeymapEntry*)b;

    return *key_dfb - keyEntry->keyDFB;
}

// Keep this in sync with the other platform event constructors
// TODO: m_balEventKey should be refcounted
PlatformKeyboardEvent::PlatformKeyboardEvent(BalEventKey* _event)
{
    int modifiers = 0;
    KeymapEntry* symbolEntry = NULL;
    KeymapEntry* identifierEntry = NULL;
    KeymapEntry* entry = NULL;
    int key_symbol = -1;

    m_balEventKey = _event;
    switch(_event->clazz)
    {
        case DFEC_WINDOW:
            {
                DFBWindowEvent *winEvent = (DFBWindowEvent*)_event;

                char* szYouTubeTVMode = getenv("YouTubeTVMode");
                if (szYouTubeTVMode && strcmp(szYouTubeTVMode, "swf") == 0)
                {
					printf("--> Enter into YouTube TV mode, %s() %d@%s.\n", __FUNCTION__, __LINE__, __FILE__);
                	if(winEvent->key_symbol == DIKS_NEXT || winEvent->key_symbol == DIKS_PREVIOUS)
					{
                		winEvent->key_symbol = DIKS_TAB;
						printf("--> The symbol has been converted into TAB.\n");
					}
                }

                if (winEvent->type == DWET_KEYDOWN)
                {
                    m_type = KeyDown;               
                } else {
                    m_type = KeyUp;
                }
                modifiers = winEvent->modifiers;
                symbolEntry = (KeymapEntry*)bsearch(&winEvent->key_symbol,
                        keysymbolMap, sizeof(keysymbolMap) / sizeof(keysymbolMap[0]) - 1, sizeof(keysymbolMap[0]),
                        compare_keymap);
                identifierEntry = (KeymapEntry*)bsearch(&winEvent->key_id,
                        keycodeMap, sizeof(keycodeMap) / sizeof(keycodeMap[0]) - 1, sizeof(keycodeMap[0]),
                        compare_keymap);
                if (symbolEntry)
                    key_symbol = winEvent->key_symbol;
                else if (identifierEntry)
                    key_symbol = winEvent->key_id;
                    
                {
                    printf("DFEC_WINDOW: symbol = %x, key_id = %x, key_code = %x ", winEvent->key_symbol, winEvent->key_id, winEvent->key_code);
                    if (symbolEntry != NULL)
                        printf(" symbol name = %s ", symbolEntry->nameDFB);
                    if (identifierEntry != NULL)
                        printf(" id name = %s ", identifierEntry->nameDFB);
                    printf("\n");
                }
            }
            break;
        case DFEC_INPUT:
            {
                DFBInputEvent *inputEvent = (DFBInputEvent*)_event;
                if (inputEvent->type == DIET_KEYPRESS)                
                {
                    m_type = KeyDown;
                } else {
                    m_type = KeyUp;
                }
                modifiers = inputEvent->modifiers;
                symbolEntry = (KeymapEntry*)bsearch(&inputEvent->key_symbol,
                        keysymbolMap, sizeof(keysymbolMap) / sizeof(keysymbolMap[0]) - 1, sizeof(keysymbolMap[0]),
                        compare_keymap);
                identifierEntry = (KeymapEntry*)bsearch(&inputEvent->key_id,
                        keycodeMap, sizeof(keycodeMap) / sizeof(keycodeMap[0]) - 1, sizeof(keycodeMap[0]),
                        compare_keymap);
                if (symbolEntry)
                    key_symbol = inputEvent->key_symbol;
                else if (identifierEntry)
                    key_symbol = inputEvent->key_id;

                {
                    printf("DFEC_INPUT: symbol = %x, key_id = %x, key_code = %x ", inputEvent->key_symbol, inputEvent->key_id, inputEvent->key_code);
                    if (symbolEntry != NULL)
                        printf(" symbol name = %s ", symbolEntry->nameDFB);
                    if (identifierEntry != NULL)
                        printf(" id name = %s ", identifierEntry->nameDFB);
                    printf("\n");
                }

            }
            break;
        default:
            ASSERT(0);
    }

    m_autoRepeat = false;
    m_isKeypad = false;
    m_shiftKey = modifiers & DIMM_SHIFT;
    m_ctrlKey = modifiers & DIMM_CONTROL;
    m_altKey =  modifiers & DIMM_ALT;
    m_metaKey = modifiers & DIMM_META;

    if (symbolEntry != NULL)
        entry = symbolEntry;
    else
        entry = identifierEntry;
    m_windowsVirtualKeyCode = VK_UNKNOWN;
    if (entry != NULL)
    {
        m_windowsVirtualKeyCode = entry->keyCode;

        UChar aSrc[2];

        if (key_symbol != -1 && DFB_KEY_TYPE(key_symbol) == DIKT_UNICODE)
            aSrc[0] = key_symbol;
        else {
            aSrc[0] = entry->keyText;

            if (symbolEntry == NULL && m_shiftKey)
            {
                if (aSrc[0] >= 'a' && aSrc[0] <= 'z')
                    aSrc[0] += ('A' - 'a');
                else if (aSrc[0] >= 'A' && aSrc[0] <= 'Z')
                    aSrc[0] += ('a' - 'A');
            }
        }
        aSrc[1] = 0;

        WebCore::String aText(aSrc);
        WebCore::String aUnmodifiedText(aSrc);
        WebCore::String aKeyIdentifier(entry->keyDesc);

        m_text = aText;
        m_unmodifiedText = aText;
        m_keyIdentifier = aKeyIdentifier;
    }
}

void PlatformKeyboardEvent::disambiguateKeyDownEvent(Type type, bool backwardCompatibilityMode)
{
    // Can only change type from KeyDown to RawKeyDown or Char, as we lack information for other conversions.
    ASSERT(m_type == KeyDown);
    m_type = type;

    if (backwardCompatibilityMode)
        return;

    if (type == RawKeyDown) {
        m_text = String();
        m_unmodifiedText = String();
    } else {
        m_keyIdentifier = String();
        m_windowsVirtualKeyCode = 0;
    }
}

bool PlatformKeyboardEvent::currentCapsLockState()
{
    NotImplemented();
    return false;
}

BalEventKey* PlatformKeyboardEvent::balEventKey() const
{
    return m_balEventKey;
}

}
