/*
 * Copyright (C) 2006 Apple Computer, Inc.
 * Copyright (C) 2006 Michael Emmel mike.emmel@gmail.com
 * Copyright (C) 2007, 2008 Alp Toker <alp@atoker.com>
 * Copyright (C) 2007 Holger Hans Peter Freyther
 * Copyright (C) 2009 Igalia S.L.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "FontPlatformData.h"

#include "CString.h"
#include "PlatformString.h"
#include "FontDescription.h"

#include <cairo-ft.h>
#include <cairo.h>

namespace WebCore {

FontPlatformData::FontPlatformData(const FontDescription& fontDescription, const AtomicString& familyName)
    : m_size(fontDescription.computedPixelSize())
    , m_syntheticBold(false)
    , m_syntheticOblique(false)
    , m_font(0)
    , m_scaledFont(0)
    , m_isFallbackFont(false)
{
    FontPlatformData::init();

    static const cairo_font_options_t* defaultOptions = cairo_font_options_create();
    cairo_matrix_t ctm;
    cairo_matrix_t fontMatrix;
    char filename[1024];

    cairo_matrix_init_identity(&ctm);
    cairo_matrix_init_scale(&fontMatrix, fontDescription.computedPixelSize(), fontDescription.computedPixelSize());

    static AtomicString timesStr("DroidSansFallback");
    if (familyName == timesStr) {
    snprintf(filename, sizeof(filename), "%s/fonts/DroidSansFallback.ttf", getenv("VA_RUNTIME_ROOT"));
        m_isFallbackFont = true;
    } else
        snprintf(filename, sizeof(filename), "%s/fonts/DroidSans.ttf", getenv("VA_RUNTIME_ROOT"));

    m_font = cairo_ft_font_face_create_for_ft_file(filename, 0, 0);    
    m_scaledFont = cairo_scaled_font_create(m_font, &fontMatrix, &ctm, defaultOptions);

    printf("font %s, size = %f, face_refcount = %d, scaled_refcount = %d\n", filename, m_size, cairo_font_face_get_reference_count(m_font), cairo_scaled_font_get_reference_count(m_scaledFont));
}

FontPlatformData::FontPlatformData(float size, bool bold, bool italic)
    : m_size(size)
    , m_syntheticBold(bold)
    , m_syntheticOblique(italic)
    , m_font(0)
    , m_scaledFont(0)
    , m_isFallbackFont(false)
{
}

FontPlatformData::FontPlatformData(cairo_font_face_t* fontFace, int size, bool bold, bool italic)
    : m_size(size)
    , m_syntheticBold(bold)
    , m_syntheticOblique(italic)
    , m_font(0)
    , m_scaledFont(0)
    , m_isFallbackFont(false)
{
    static const cairo_font_options_t* defaultOptions = cairo_font_options_create();
    cairo_matrix_t fontMatrix;
    cairo_matrix_t ctm;

    cairo_matrix_init_scale(&fontMatrix, size, size);
    cairo_matrix_init_identity(&ctm);

    m_font = cairo_font_face_reference(fontFace);
    m_scaledFont = cairo_scaled_font_create(m_font, &fontMatrix, &ctm, defaultOptions);
}

FontPlatformData& FontPlatformData::operator=(const FontPlatformData& other)
{
    // Check for self-assignment.
    if (this == &other)
        return *this;

    m_size = other.m_size;
    m_syntheticBold = other.m_syntheticBold;
    m_syntheticOblique = other.m_syntheticOblique;

    if (other.m_font && other.m_font != hashTableDeletedFontValue())
        cairo_font_face_reference(other.m_font);
    if (m_font && m_font != hashTableDeletedFontValue())
        cairo_font_face_destroy(m_font);
    m_font = other.m_font;

    if (other.m_scaledFont)
        cairo_scaled_font_reference(other.m_scaledFont);
    if (m_scaledFont)
        cairo_scaled_font_destroy(m_scaledFont);
    m_scaledFont = other.m_scaledFont;
    m_isFallbackFont = other.m_isFallbackFont;

    return *this;
}

FontPlatformData::FontPlatformData(const FontPlatformData& other)
    : m_font(0)    
    , m_scaledFont(0)
{
    *this = other;
}

bool FontPlatformData::init()
{
    return true;
}

FontPlatformData::~FontPlatformData()
{
    if (m_font && m_font != hashTableDeletedFontValue()) {
        cairo_font_face_destroy(m_font);
        m_font = 0;
    }

    if (m_scaledFont) {
        cairo_scaled_font_destroy(m_scaledFont);
        m_scaledFont = 0;
    }
}

bool FontPlatformData::isFixedPitch()
{
    return false;
}

void FontPlatformData::setFont(cairo_t* cr) const
{
    ASSERT(m_scaledFont);

    cairo_set_scaled_font(cr, m_scaledFont);
}

bool FontPlatformData::operator==(const FontPlatformData& other) const
{
    return (m_font == other.m_font && m_scaledFont == other.m_scaledFont);
}

#ifndef NDEBUG
String FontPlatformData::description() const
{
    return String();
}
#endif

}
