/*
 * Copyright (C) 2008 Pleyo.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Pleyo nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PLEYO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL PLEYO OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "config.h"
#include "Logging.h"
#include "Font.h"

#include "GraphicsContext.h"
#include "BALBase.h"
#include "SimpleFontData.h"
#include "fonts/pixelfont.h"
#include "DebugPrintf.h"

#if PLATFORM(SDL)
#include "SDL.h"
#endif

namespace WebCore {

#define IS_HIGH_SURROGATE(u)  ((UChar)(u) >= (UChar)0xd800 && (UChar)(u) <= (UChar)0xdbff)
#define IS_LOW_SURROGATE(u)  ((UChar)(u) >= (UChar)0xdc00 && (UChar)(u) <= (UChar)0xdfff)

bool Font::canReturnFallbackFontsForComplexText()
{
    return false;
}

void Font::drawComplexText(GraphicsContext* context, const TextRun& run, const FloatPoint& point, int from, int to) const
{
    BalNotImplemented();
}


float Font::floatWidthForComplexText(const TextRun& run, HashSet<const SimpleFontData*>* fallbackFonts) const
{
    BalNotImplemented();
    return 0.0;
}

int Font::offsetForPositionForComplexText(const TextRun& run, int x, bool includePartialGlyphs) const
{
    BalNotImplemented();
    return 0;
}

FloatRect Font::selectionRectForComplexText(const TextRun& run, const IntPoint& point, int h, int from, int to) const
{
    BalNotImplemented();
    return FloatRect();
}

void Font::drawGlyphs(GraphicsContext* context, const SimpleFontData* font, const GlyphBuffer& glyphBuffer,
                      int from, int numGlyphs, const FloatPoint& point) const
{
    // Set the text color to use for drawing.
    Color penColor = context->fillColor();
    unsigned pixelColor = (penColor.alpha() << 24) | (penColor.red() << 16) | (penColor.green() << 8) | penColor.blue();
    GlyphBufferGlyph* glyphs = const_cast<GlyphBufferGlyph*>(glyphBuffer.glyphs(from));
    PixelFontType *fontPixel = font->m_platformData.m_pixelFont;
    uint8_t w = fontPixel->width;
    uint8_t h = fontPixel->height;
    IntRect dstRect(0, 0, w * numGlyphs , h);

#if PLATFORM(SDL)
    SDL_Surface* surface = context->platformContext();
    Uint32 rmask, gmask, bmask, amask;
    /* SDL interprets each pixel as a 32-bit number, so our masks must depend
    on the endianness (byte order) of the machine */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    rmask = 0xff000000;
    gmask = 0x00ff0000;
    bmask = 0x0000ff00;
    amask = 0x000000ff;
#else
    rmask = 0x00ff0000;
    gmask = 0x0000ff00;
    bmask = 0x000000ff;
    amask = 0xff000000;
#endif
    /* Use SWSURFACE to improve the performance of pixel level access. */
    SDL_Surface* imgBuffer = SDL_CreateRGBSurface(SDL_SWSURFACE, w * numGlyphs, h, 32, rmask, gmask, bmask, amask);
    //the surface can be null if the width or the height are too big, it's a sdl limitation.
    if (!imgBuffer)
        return;

    SDL_LockSurface(imgBuffer);
    uint32_t* baseAddr = (uint32_t*) imgBuffer->pixels;
    //Vector<unsigned> *glyphRGBABuffer = new Vector<unsigned>(h * w * numGlyphs);
    IntPoint aPoint(static_cast<uint16_t> (point.x()), static_cast<uint16_t> (point.y() - h));
    for (int j = 0; j < h; j++) {
        for (int i = 0; i < numGlyphs; i++) {
            for (int k = 0; k < w; k++) {
                uint16_t glyphOffset = glyphs[i] * h + j;
                if (fontPixel->buffer[glyphOffset] & (1 << (w - k)))
                    baseAddr[(j * numGlyphs * w) + k + (i * w)] = pixelColor;
                else
                    baseAddr[(j * numGlyphs * w) + k + (i * w)] = 0;
            }
        }
    }
    SDL_UnlockSurface(imgBuffer);

    SDL_Rect sdlSrc, sdlDest;
    sdlDest.x = aPoint.x() + context->origin().x();
    sdlDest.y = aPoint.y() + context->origin().y();
    sdlDest.w = dstRect.width();
    sdlDest.h = dstRect.height();
    sdlSrc.w = dstRect.width();
    sdlSrc.h = dstRect.height();
    sdlSrc.x = dstRect.x();
    sdlSrc.y = dstRect.y();

    D_INFO("{%d, %d, %d, %d} --> {%d, %d}\n",
            sdlSrc.x, sdlSrc.y, sdlSrc.w, sdlSrc.h,
            sdlDest.x, sdlDest.y
           );

    SDL_BlitSurface(imgBuffer, &sdlSrc, surface, &sdlDest);    

    /* use SDL_Flip only if double buffering is available */
    /*if (surface->flags & SDL_DOUBLEBUF) {
       SDL_Flip(surface);
    }
    else {
       SDL_UpdateRect(surface, sdlDest.x + origin().width(), sdlDest.y + origin().height(), sdlDest.w, sdlDest.h);
    }*/
    SDL_FreeSurface(imgBuffer);

#elif PLATFORM(DFB) //PLATFORM(DFB)

#if !PLATFORM(DFB_CAIRO)
    IDirectFB* dfb = NULL;
    DFBGraphicsEngine* surface = context->platformGraphicsEngine();
    DFBSurfaceDescription desc;
    IDirectFBSurface *imgBuffer = NULL;
    unsigned char* pixels;
    int pitch;
    IntPoint aPoint(static_cast<uint16_t> (point.x()), static_cast<uint16_t> (point.y() - h));
 
    OWB_DFBCHECK(DirectFBCreate(&dfb));
    desc.flags = (DFBSurfaceDescriptionFlags)(DSDESC_WIDTH | DSDESC_HEIGHT | DSDESC_PIXELFORMAT | DSDESC_CAPS);
    desc.caps = DSCAPS_SYSTEMONLY;
    desc.width = w * numGlyphs;
    desc.height = h;
    desc.pixelformat = DSPF_ARGB;
    OWB_DFBCHECK(dfb->CreateSurface(dfb, &desc, &imgBuffer));

    OWB_DFBCHECK(imgBuffer->Lock(imgBuffer, DSLF_WRITE, (void**)&pixels, &pitch));
    for (int j = 0; j < h; j++)
    {
        uint32_t *baseAddr = (uint32_t*)(pixels + j * pitch);
        for (int i = 0; i < numGlyphs; i++)
        {
            for (int k = 0; k < w; k++)
            {
                uint16_t glyphOffset = glyphs[i] * h + j;
                if (fontPixel->buffer[glyphOffset] & (1 << (w - k)))
                    baseAddr[k + (i * w)] = pixelColor;
                else
                    baseAddr[k + (i * w)] = 0;
            }
        }
    }
    OWB_DFBCHECK(imgBuffer->Unlock(imgBuffer));

    D_INFO(" {%d, %d, %d, %d} --> {%d, %d} \n", 
            0, 0, w * numGlyphs, h,
            aPoint.x() + context->origin().x(),
            aPoint.y() + context->origin().y()
            );

    surface->bitBlit(aPoint.x() + context->origin().x(),
            aPoint.y() + context->origin().y(),
            w * numGlyphs,
            h,
            imgBuffer,
            0,
            0);

    OWB_SAFE_RELEASE(imgBuffer);
    OWB_SAFE_RELEASE(dfb);
#else /// PLATFORM(DFB_CAIRO) 
    IntPoint aPoint(static_cast<uint16_t> (point.x()), static_cast<uint16_t> (point.y()) - h);
    
    cairo_t* cr = context->platformContext();
    cairo_save(cr);

    cairo_surface_t *glyphSurface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, w*numGlyphs, h);

    if (!glyphSurface)
    {
        printf(" failed to create glyph surface {%d, %d}\n", w*numGlyphs, h);
        return;
    }
    unsigned char* pixels = cairo_image_surface_get_data (glyphSurface);
    int pitch = cairo_image_surface_get_stride (glyphSurface);

    for (int j = 0; j < h; j++)
    {
        uint32_t *baseAddr = (uint32_t*)(pixels + j * pitch);
        for (int i = 0; i < numGlyphs; i++)
        {
            for (int k = 0; k < w; k++)
            {
                uint16_t glyphOffset = glyphs[i] * h + j;
                if (fontPixel->buffer[glyphOffset] & (1 << (w - k)))
                    baseAddr[k + (i * w)] = pixelColor;
                else
                    baseAddr[k + (i * w)] = 0;
            }
        }
    }

    cairo_pattern_t* pattern = cairo_pattern_create_for_surface(glyphSurface);
    cairo_pattern_set_filter(pattern, CAIRO_FILTER_NEAREST);

    float scaleX = 1.0;
    float scaleY = 1.0;
    cairo_matrix_t matrix = { scaleX, 0, 0, scaleY, 0, 0 };
    cairo_pattern_set_matrix(pattern, &matrix);

    // Draw the image.
    cairo_translate(cr, aPoint.x(), aPoint.y());
    cairo_set_source(cr, pattern);
//     cairo_rectangle(cr, 0, 0, w*numGlyphs, h);
//     cairo_clip(cr);
    cairo_paint_with_alpha(cr, context->getAlpha());

    cairo_restore (cr);

    cairo_pattern_destroy(pattern);
    cairo_surface_destroy(glyphSurface);    
#endif /// !PLATFORM(DFB_CAIRO)

#else
    NotImplemented();
#endif

}

}
