/*
 * Copyright (C) 2008 Pleyo.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Pleyo nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PLEYO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL PLEYO OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "config.h"
#include "FontPlatformData.h"

#include "CString.h"
#include "PlatformString.h"
#include "Logging.h"
#include "FontDescription.h"
#include <ft2build.h>
#include FT_FREETYPE_H

#if USE(FONTCONFIG)
#include <fontconfig/fontconfig.h>
#endif


//static const char *too = "/usr/share/fonts/MSYH/msyh.ttf";

namespace WebCore {

FontPlatformData::FontPlatformData(const FontDescription& fontDescription, const AtomicString& familyName)
    : m_syntheticBold(false)
    , m_syntheticOblique(false)
    , m_isFixedPitch(false)
    , m_size(fontDescription.computedSize())
    , m_nativeFont(0)
{
    FontPlatformData::init();

#if USE(FONTCONFIG)
    CString familyNameString = familyName.string().utf8();
    const char* fcfamily = familyNameString.data();

    int fcslant = FC_SLANT_ROMAN;
    // FIXME: Map all FontWeight values to fontconfig weights.
    int fcweight = FC_WEIGHT_NORMAL;
    float fcsize = fontDescription.computedSize();
    if (fontDescription.italic())
        fcslant = FC_SLANT_ITALIC;
    if (fontDescription.weight() >= FontWeight600)
        fcweight = FC_WEIGHT_BOLD;

    int type = fontDescription.genericFamily();
    //printf("family = %s\n", fcfamily);

    FcConfig *config = FcConfigGetCurrent();
    FcPattern* pattern = FcPatternCreate();
    FcPattern* matchedPattern = NULL;
    
    if (!FcPatternAddString(pattern, FC_FAMILY, reinterpret_cast<const FcChar8*>(fcfamily)))
        goto freePattern;

    switch (type) {
        case FontDescription::SerifFamily:
            fcfamily = "serif";
            break;
        case FontDescription::SansSerifFamily:
            fcfamily = "sans-serif";
            break;
        case FontDescription::MonospaceFamily:
            fcfamily = "monospace";
            break;
        case FontDescription::NoFamily:
        case FontDescription::StandardFamily:
        default:
            fcfamily = "sans-serif";
    }

    if (!FcPatternAddString(pattern, FC_FAMILY, reinterpret_cast<const FcChar8*>(fcfamily)))
        goto freePattern;
    if (!FcPatternAddInteger(pattern, FC_SLANT, fcslant))
        goto freePattern;
    if (!FcPatternAddInteger(pattern, FC_WEIGHT, fcweight))
        goto freePattern;
    if (!FcPatternAddDouble(pattern, FC_PIXEL_SIZE, fcsize))
        goto freePattern;

    FcConfigSubstitute(config, pattern, FcMatchPattern);
    FcDefaultSubstitute(pattern);

    FcResult fcresult;
    matchedPattern = FcFontMatch(config, pattern, &fcresult);
    //FcPatternReference(matchedPattern);
    // FIXME: should we set some default font?
    if (!matchedPattern)
        goto freePattern;

    FcChar8 *fc_filename;
    char *filename;
    int id;
    id = 0;

    if (FcPatternGetString(matchedPattern, FC_FILE, 0, &fc_filename) != FcResultMatch) {
        LOG(FontEngine, "cannot retrieve font\n");
        goto freePattern;
    }

    filename = (char *) fc_filename; //use C cast as FcChar is a fontconfig type
    //printf("filename = %s\n", filename);

    if (FcPatternGetInteger(matchedPattern, FC_INDEX, 0, &id) != FcResultMatch) {
        LOG(FontEngine, "cannot retrieve font index\n");
        goto freePattern;
    }

    int spacing;
    if (FcPatternGetInteger(matchedPattern, FC_SPACING, 0, &spacing) == FcResultMatch)
        m_isFixedPitch = (spacing == FC_MONO);

    m_nativeFont = RefCountedFontFace::createFontFace(filename, id, static_cast<unsigned>(fontDescription.computedSize()));

freePattern:
    if (matchedPattern)
        FcPatternDestroy(matchedPattern);
    FcPatternDestroy(pattern);
    FcConfigDestroy(config);
#else
    char filename[1024];

    snprintf(filename, sizeof(filename), "%s/fonts/DroidSansFallback.ttf", getenv("VA_RUNTIME_ROOT"));
    m_nativeFont = RefCountedFontFace::createFontFace(filename, 0, static_cast<unsigned>(fontDescription.computedSize()));

#endif    
}

FontPlatformData::FontPlatformData(float size, bool bold, bool italic)
    : m_syntheticBold(bold)
    , m_syntheticOblique(italic)
    , m_isFixedPitch(false)
    , m_size(size)
    , m_nativeFont(0)
{
}

FontPlatformData::FontPlatformData(SharedBuffer* buffer, int id, int size)
    : m_syntheticBold(false)
    , m_syntheticOblique(false)
    , m_isFixedPitch(false)
    , m_size(size)
    , m_nativeFont(0)
{
    m_nativeFont = RefCountedFontFace::createFontFace(buffer, id, size);
}

FontPlatformData::FontPlatformData(const FontPlatformData& other)
{
    m_syntheticBold = other.m_syntheticBold;
    m_syntheticOblique = other.m_syntheticOblique;
    m_isFixedPitch = other.m_isFixedPitch;
    m_size = other.m_size;
    m_nativeFont = other.m_nativeFont;
}

FontPlatformData& FontPlatformData::operator=(const FontPlatformData& other)
{
    m_syntheticBold = other.m_syntheticBold;
    m_syntheticOblique = other.m_syntheticOblique;
    m_isFixedPitch = other.m_isFixedPitch;
    m_size = other.m_size;
    m_nativeFont = other.m_nativeFont;

    return *this;
}

bool FontPlatformData::init()
{
#if USE(FONTCONFIG)    
    static bool initialized = false;
    if (initialized)
        return true;

    if (!FcInit()) {
        fprintf(stderr, "Can't init font config library\n");
        return false;
    }

    initialized = true;
#endif    
    return true;
}

FontPlatformData::~FontPlatformData()
{
#if 0    
    if (m_nativeFont) {
        m_nativeFont->deref();
        m_nativeFont = NULL;
    }
#endif    
}

void FontPlatformData::setFont(BalFont* cr) const
{
    //ASSERT(m_scaledFont);

    BalNotImplemented();
}

bool FontPlatformData::operator==(const FontPlatformData& other) const
{
    if (m_nativeFont == other.m_nativeFont)
        return true;
    return false;
}

#ifndef NDEBUG
String FontPlatformData::description() const
{
    return String();
}
#endif

}
