
#include <wtf/HashMap.h>
#include <wtf/RefPtr.h>
#include <AtomicString.h>
#include <CString.h>

#include "RefCountedFontFace.h"

namespace WKAL {

struct RefCountedFontFaceCacheKey {
    RefCountedFontFaceCacheKey(const AtomicString& filename = AtomicString(), const SharedBuffer* buffer = NULL, unsigned id = 0, unsigned size = 0)
        : m_filename(filename)
        , m_buffer(buffer)
        , m_id(id)
        , m_size(size)
    {
    }
    RefCountedFontFaceCacheKey(WTF::HashTableDeletedValueType)
        : m_size(hashTableDeletedSize())
    {
    }
    bool isHashTableDeletedValue() const { return m_size == hashTableDeletedSize(); }

    bool operator==(const RefCountedFontFaceCacheKey& other) const
    {
        return (m_filename == other.m_filename
                && m_buffer == other.m_buffer
                && m_id == other.m_id
                && m_size == other.m_size);
    }
    AtomicString m_filename;
    const SharedBuffer *m_buffer;
    unsigned m_id;
    unsigned m_size;

private:
    static unsigned hashTableDeletedSize() { return 0xFFFFFFFF; }
};

struct RefCountedFontFaceCacheKeyHash {
    static unsigned hash(const RefCountedFontFaceCacheKey& nativeFont)
    {
        if (nativeFont.m_buffer) {
            uintptr_t hashCodes[1] = { reinterpret_cast<uintptr_t>(nativeFont.m_buffer) };
            return StringImpl::computeHash(reinterpret_cast<UChar*>(hashCodes), sizeof(hashCodes) / sizeof(UChar));
        }
        return nativeFont.m_filename.impl()->hash();
    }

    static bool equal(const RefCountedFontFaceCacheKey& a, const RefCountedFontFaceCacheKey& b)
    {
        return a == b;
    }
    static const bool safeToCompareToEmptyOrDeleted = true;
};

struct RefCountedFontFaceCacheKeyTraits : WTF::GenericHashTraits<RefCountedFontFaceCacheKey> {
    static const bool emptyValueIsZero = true;
    static const RefCountedFontFaceCacheKey& emptyValue()
    {
        const static RefCountedFontFaceCacheKey& key = RefCountedFontFaceCacheKey(nullAtom);
        return key;
    }
    static void constructDeletedValue(RefCountedFontFaceCacheKey& slot)
    {
        new (&slot) RefCountedFontFaceCacheKey(WTF::HashTableDeletedValue);
    }
    static bool isDeletedValue(const RefCountedFontFaceCacheKey& value)
    {
        return value.isHashTableDeletedValue();
    }
};

typedef HashMap<RefCountedFontFaceCacheKey, RefCountedFontFace*, RefCountedFontFaceCacheKeyHash, RefCountedFontFaceCacheKeyTraits> RefCountedFontFaceCache;

FT_Library RefCountedFontFace::m_library = 0;
static RefCountedFontFaceCache* m_fontFaceCache = 0;

bool RefCountedFontFace::init()
{
    static bool initialized = false;
    if (initialized)
        return true;

    FT_Error error = FT_Init_FreeType(&m_library);
    if (error) {
        fprintf(stderr, "Couldn't init FreeType engine\n");
        m_library = 0;
        return false;
    }

    if (!m_fontFaceCache)
        m_fontFaceCache = new RefCountedFontFaceCache();

    initialized = true;
    return true;
}

PassRefPtr<RefCountedFontFace> RefCountedFontFace::createFontFace(const char* filename, int id, int size)
{
    PassRefPtr<RefCountedFontFace> nativeFontFace;
    FT_Face face = NULL;
    BalNativeFontFace* nativeFont = NULL;

    init();

    RefCountedFontFaceCacheKey key(AtomicString(filename), NULL, id, size);
    RefCountedFontFaceCache::iterator it = m_fontFaceCache->find(key);
    if (it != m_fontFaceCache->end()) {
        ASSERT(it->second);
        printf("%s: reuse font face %s with id %d with size %d with counter %d\n", __FUNCTION__, filename, id, size, it->second->refCount());
        return adoptRef(it->second);
    }

    printf("%s: open font %s with id %d with size %d\n", __FUNCTION__, filename, id, size);

    FT_Error error;
    if (error = FT_New_Face(m_library, filename, id, &face)) {
        printf("fail to open fonti %s with index %d (error = 0x%x)\n", filename, id, error);
        goto exit;
    }

    FT_Set_Pixel_Sizes(face, 0, size);

#if PLATFORM(DFB_CAIRO)    
    static const cairo_font_options_t* defaultOptions = cairo_font_options_create();
    cairo_font_face_t *fontFace;
    cairo_matrix_t fontMatrix;
    cairo_matrix_t ctm;
    cairo_matrix_init_scale(&fontMatrix, size, size);
    cairo_matrix_init_identity(&ctm);
    fontFace = cairo_ft_font_face_create_for_ft_face(face, 0);
    nativeFont = cairo_scaled_font_create(fontFace, &fontMatrix, &ctm, defaultOptions);
    cairo_font_face_destroy(fontFace);
#endif

    nativeFontFace = adoptRef(new RefCountedFontFace(face, nativeFont));
    m_fontFaceCache->set(key, nativeFontFace.get());

exit:
    return nativeFontFace;
}

PassRefPtr<RefCountedFontFace> RefCountedFontFace::createFontFace(const SharedBuffer* buffer, int id, int size)
{
    ASSERT_ARG(buffer, buffer);

    PassRefPtr<RefCountedFontFace> nativeFontFace;
    int error;
    FT_Face face = NULL;
    BalNativeFontFace* nativeFont = NULL;

    init();

    RefCountedFontFaceCacheKey key(AtomicString(), buffer, id, size);
    RefCountedFontFaceCache::iterator it = m_fontFaceCache->find(key);
    if (it != m_fontFaceCache->end()) {
        ASSERT(it->second);
        printf("%s: reuse font face %p with id %d with size %d with counter %d\n", __FUNCTION__, buffer, id, size, it->second->refCount());
        return adoptRef(it->second);
    }

    printf("%s: open font %p with id %d with size %d\n", __FUNCTION__, buffer, id, size);

    error = FT_New_Memory_Face(m_library, reinterpret_cast<const FT_Byte*>(buffer->data()), buffer->size(), 0, &face);
    if (error)
        goto exit;

    FT_Set_Pixel_Sizes(face, 0, size);

#if PLATFORM(DFB_CAIRO)    
    static const cairo_font_options_t* defaultOptions = cairo_font_options_create();
    cairo_font_face_t *fontFace;
    cairo_matrix_t fontMatrix;
    cairo_matrix_t ctm;
    cairo_matrix_init_scale(&fontMatrix, size, size);
    cairo_matrix_init_identity(&ctm);
    fontFace = cairo_ft_font_face_create_for_ft_face(face, 0);
    nativeFont = cairo_scaled_font_create(fontFace, &fontMatrix, &ctm, defaultOptions);
    cairo_font_face_destroy(fontFace);
#endif

    nativeFontFace = adoptRef(new RefCountedFontFace(face, nativeFont));
    m_fontFaceCache->set(key, nativeFontFace.get());

exit:
    return nativeFontFace;
}

RefCountedFontFace::~RefCountedFontFace()
{
    printf("%s : %p, %p\n", __FUNCTION__, m_face, m_nativeFont);
    NotImplemented();
#if PLATFORM(DFB_CAIRO)
    
#elif PLATFORM(DFB)
#else
#endif
}

} //namespace WKAL

