/*
 * Copyright (C) 2008 Pleyo.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Pleyo nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PLEYO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL PLEYO OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "config.h"
#include "GraphicsContext.h"

#include "TransformationMatrix.h"
#include "FloatRect.h"
#include "Font.h"
#include "ImageBuffer.h"
#include "IntRect.h"
#include "NotImplemented.h"
#include "Path.h"
#include "Pattern.h"
#include "SimpleFontData.h"
#include <math.h>
#include <stdio.h>
#include <wtf/MathExtras.h>

#include "GraphicsContextPrivate.h"
#include "GraphicsContextPlatformPrivateDFB.h"
#include "DebugPrintf.h"

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

using std::min;
using std::max;

namespace WebCore {

GraphicsContext::GraphicsContext(PlatformGraphicsContext* cr)
    : m_common(createGraphicsContextPrivate())
    , m_data(new GraphicsContextPlatformPrivate(cr))
{
    setPaintingDisabled(!cr);
}

GraphicsContext::~GraphicsContext()
{
     destroyGraphicsContextPrivate(m_common);
     delete m_data;
#ifndef NDEBUG     
     m_common = NULL;
     m_data = NULL;
#endif
}

TransformationMatrix GraphicsContext::getCTM() const
{
    NotImplemented();
    TransformationMatrix a;
    return a;
}

PlatformGraphicsContext* GraphicsContext::platformContext() const
{
    return m_data->m_graphicsEngine.m_surface;
}

DFBGraphicsEngine* GraphicsContext::platformGraphicsEngine() const
{
    return &(m_data->m_graphicsEngine);
}

void GraphicsContext::savePlatformState()
{
    if (paintingDisabled())
        return;

    m_common->stack.append(m_common->state);
    m_data->m_graphicsEngine.save();
}

void GraphicsContext::restorePlatformState()
{
    if (paintingDisabled())
        return;

    if (m_common->stack.isEmpty()) {
        LOG(Graphics, "ERROR void BCGraphicsContext::restore() stack is empty");
        return;
    }

    m_common->state = m_common->stack.last();
    m_common->stack.removeLast();
    m_data->m_graphicsEngine.restore();
}

// Draws a filled rectangle with a stroked border.
void GraphicsContext::drawRect(const IntRect& rect)
{
    if (paintingDisabled())
        return;

    int x = rect.x() + origin().x();
    int y = rect.y() + origin().y();
    int w = rect.width();
    int h = rect.height();
    uint8_t alpha = fillColor().alpha();

    D_INFO("{%d, %d, %d, %d}\n", x, y, w, h);

    if (alpha)
    {
       if (!m_data->layers.isEmpty())
            alpha = static_cast<uint8_t>(alpha * m_data->layers.last());
       m_data->m_graphicsEngine.setColor(alpha, 
                fillColor().red(), 
                fillColor().green(), 
                fillColor().blue());
        m_data->m_graphicsEngine.fillRect(x, y, w, h);
    }

    if (strokeStyle() != NoStroke)
    {
       alpha = strokeColor().alpha();
        if (!m_data->layers.isEmpty())
            alpha = static_cast<uint8_t>(alpha * m_data->layers.last());
        m_data->m_graphicsEngine.setColor(alpha,
                strokeColor().red(),
                strokeColor().green(),
                strokeColor().blue());
        m_data->m_graphicsEngine.drawRect(x, y, w, h);
    }
}


// This is only used to draw borders.
void GraphicsContext::drawLine(const IntPoint& point1, const IntPoint& point2)
{
    if (paintingDisabled())
        return;

    StrokeStyle style = strokeStyle();
    if (style == NoStroke)
        return;
    
    float width = strokeThickness();
    if (width < 1)
        width = 1;

    IntPoint p1(point1 + IntSize(origin().x(), origin().y()));
    IntPoint p2(point2 + IntSize(origin().x(), origin().y()));
    Color color = strokeColor();

    D_INFO("{%d, %d} --> {%d, %d}\n", p1.x(), p1.y(), p2.x(), p2.y());

    int alpha;
    if (!m_data->layers.isEmpty())
        alpha = static_cast<int> (strokeColor().alpha() * m_data->layers.last());
    else
        alpha = strokeColor().alpha();

    m_data->m_graphicsEngine.setColor(alpha, color.red(), color.green(), color.blue());
    m_data->m_graphicsEngine.drawLine(p1.x(), p1.y(), p2.x(), p2.y());
}

// This method is only used to draw the little circles used in lists.
void GraphicsContext::drawEllipse(const IntRect& rect)
{
    if (paintingDisabled())
        return;

    float yRadius = .5 * rect.height();
    float xRadius = .5 * rect.width();
    // Mean that we will draw a circle
    if (xRadius == yRadius)
    {
        strokeArc(rect, 0, 360);
        return;
    }

    if (strokeStyle() != NoStroke)
    {
        unsigned width = static_cast<unsigned int>(strokeThickness());
        if (width == 0)
            width++;
    }
    Color color = strokeColor();

    D_INFO("{%d, %d} --> {%d, %d}\n",
            rect.x() + origin().x(),
            rect.y() + origin().y(), 
            rect.width(), rect.height());

    int alpha;
    if (!m_data->layers.isEmpty())
        alpha = static_cast<int> (strokeColor().alpha() * m_data->layers.last());
    else
        alpha = strokeColor().alpha();

    m_data->m_graphicsEngine.setColor(alpha, strokeColor().red(),
            strokeColor().green(), strokeColor().blue());
    m_data->m_graphicsEngine.drawEllipse(
            rect.x() + origin().x(),
            rect.y() + origin().y(),
            rect.width(),
            rect.height(), false);
}

// FIXME: This function needs to be adjusted to match the functionality on the Mac side.
void GraphicsContext::strokeArc(const IntRect& rect, int startAngle, int angleSpan)
{
    if (paintingDisabled())
        return;

     if (strokeStyle() == NoStroke)
        return;

     NotImplemented();
}

void GraphicsContext::drawConvexPolygon(size_t npoints, const FloatPoint* points, bool shouldAntialias)
{
    if (paintingDisabled())
        return;

    if (npoints <= 2)
        return;

    DFBPoint *dfbpoints = new DFBPoint[npoints];

    for(size_t i=0; i < npoints; i++)
    {
        dfbpoints[i].x = static_cast<int>(points[i].x() + origin().x());
        dfbpoints[i].y = static_cast<int>(points[i].y() + origin().y());
        D_INFO("{%d, %d}\n", dfbpoints[i].x, dfbpoints[i].y);
    }

    int alpha;
    if (!m_data->layers.isEmpty())
        alpha = static_cast<int> (strokeColor().alpha() * m_data->layers.last());
    else
        alpha = strokeColor().alpha();

    Color color = fillColor();
    m_data->m_graphicsEngine.setColor(alpha, color.red(), color.green(), color.blue());
    m_data->m_graphicsEngine.fillPolygon(dfbpoints, npoints);

    delete[] dfbpoints;
}

void GraphicsContext::fillRect(const FloatRect& rectangle, const Color& color)
{
    if (paintingDisabled())
        return;
    
    FloatRect rect(rectangle);
    rect.setLocation(rectangle.location() + IntSize(origin().x(), origin().y()));

    int x = static_cast<int>(rect.x());
    int y = static_cast<int>(rect.y());
    int w = static_cast<int>(rect.width());
    int h = static_cast<int>(rect.height());

    D_INFO("{%d, %d, %d, %d}\n", x, y, w, h);

    if (color.alpha())
    {
        int alpha;
        if (!m_data->layers.isEmpty())
            alpha = static_cast<int>(color.alpha() * m_data->layers.last());
        else
            alpha = color.alpha();

        m_data->m_graphicsEngine.setColor(alpha,
                color.red(), color.green(), color.blue());
        m_data->m_graphicsEngine.fillRect(x, y, w, h);
    }

}

void GraphicsContext::fillRect(const FloatRect& rect)
{
    fillRect(IntRect(rect), fillColor());
}

void GraphicsContext::clip(const FloatRect& rect)
{
    if (paintingDisabled())
        return;

    if (rect.isEmpty())
    {
        m_data->m_graphicsEngine.removeClip();
    } else {
        float x = rect.x() + origin().x();
        x = x > 0 ? x : 0;
        float y = rect.y() + origin().y();
        y = y > 0 ? y : 0;
        m_data->m_graphicsEngine.setClipRect(
                static_cast<int>(x),
                static_cast<int>(y),
                static_cast<int>(rect.width()),
                static_cast<int>(rect.height()));
    }
}

void GraphicsContext::clipPath(WindRule clipRule)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}


void GraphicsContext::drawFocusRing(const Color& color)
{
    if (paintingDisabled())
        return;

    int radius = (focusRingWidth() - 1) / 2;
    int offset = radius + focusRingOffset();

    const Vector<IntRect>& rects = focusRingRects();
    unsigned rectCount = rects.size();
    IntRect finalFocusRect;
    for (unsigned i = 0; i < rectCount; i++)
    {
        IntRect focusRect = rects[i];
        focusRect.setLocation(focusRect.location() + IntSize(origin().x(), origin().y()));
        focusRect.inflate(offset);
        finalFocusRect.unite(focusRect);
    }

    D_INFO("{%d, %d, %d, %d}\n", 
            finalFocusRect.x(), 
            finalFocusRect.y(), 
            finalFocusRect.width(), 
            finalFocusRect.height());
    // NOTE These rects are rounded on Mac
    // Force the alpha to 50%.  This matches what the Mac does with outline rings.
    //fillRect(finalFocusRect, Color::gray);
    //Color c = fillColor();
    //Color ringColor(color.red(), color.green(), color.blue(), 127);
    m_data->m_graphicsEngine.setColor(127, color.red(), color.green(), color.blue());
    m_data->m_graphicsEngine.drawRect(
                static_cast<int>(finalFocusRect.x()), 
                static_cast<int>(finalFocusRect.y()),
                static_cast<int>(finalFocusRect.width()), 
                static_cast<int>(finalFocusRect.height()));
}

void GraphicsContext::drawLineForText(const IntPoint& startPoint, int width, bool printing)
{
    if (paintingDisabled())
        return;

    //NotImplemented();
    // This is a workaround for http://bugs.webkit.org/show_bug.cgi?id=15659
    StrokeStyle savedStrokeStyle = strokeStyle();
    setStrokeStyle(SolidStroke);

    IntPoint point(startPoint);

    IntPoint endPoint = point + IntSize(width, 0);

    // NOTE we should adjust line to pixel boundaries
    drawLine(point, endPoint);

    setStrokeStyle(savedStrokeStyle);
}

void GraphicsContext::drawLineForMisspellingOrBadGrammar(const IntPoint& origin, int width, bool grammar)
{
    NotImplemented();
}

FloatRect GraphicsContext::roundToDevicePixels(const FloatRect& frect)
{
    return frect;
}

void GraphicsContext::translate(float x, float y)
{
    if (paintingDisabled())
        return;
    m_common->state.origin += IntSize(static_cast<int>(x), static_cast<int>(y));
}

IntPoint GraphicsContext::origin()
{
    return m_common->state.origin;
}


void GraphicsContext::setPlatformFillColor(const Color& col)
{
    // FIXME: this is probably a no-op but I'm not sure
    // notImplemented(); // commented-out because it's chatty and clutters output
	m_common->state.fillColor = col;
}

void GraphicsContext::setPlatformStrokeColor(const Color& col)
{
    // FIXME: this is probably a no-op but I'm not sure
    //notImplemented(); // commented-out because it's chatty and clutters output
	m_common->state.strokeColor = col;
}

void GraphicsContext::setPlatformStrokeThickness(float strokeThickness)
{
    m_common->state.strokeThickness = strokeThickness;
}

void GraphicsContext::setPlatformStrokeStyle(const StrokeStyle& strokeStyle)
{
	m_common->state.strokeStyle = strokeStyle;
}

void GraphicsContext::setURLForRect(const KURL& link, const IntRect& destRect)
{
    notImplemented();
}

void GraphicsContext::concatCTM(const TransformationMatrix& transform)
{
    NotImplemented();
}

void GraphicsContext::addInnerRoundedRectClip(const IntRect& rect, int thickness)
{
    NotImplemented();
}

void GraphicsContext::clipToImageBuffer(const FloatRect& rect, const ImageBuffer* imageBuffer)
{
    if (paintingDisabled())
        return;

    notImplemented();
}

void GraphicsContext::setPlatformShadow(IntSize const&, int, Color const&)
{
    notImplemented();
}

void GraphicsContext::clearPlatformShadow()
{
    notImplemented();
}

void GraphicsContext::beginTransparencyLayer(float opacity)
{
    if (paintingDisabled())
        return;
    m_data->layers.append(opacity);
    m_data->beginTransparencyLayer();
}

void GraphicsContext::endTransparencyLayer()
{
    if (paintingDisabled())
        return;

    m_data->layers.removeLast();
    m_data->endTransparencyLayer();
}

void GraphicsContext::clearRect(const FloatRect& rect)
{
    if (paintingDisabled())
        return;

    IntRect rectangle(rect);
    rectangle.setLocation(rectangle.location() + IntSize(origin().x(), origin().y()));
    //fillRect(rectangle, Color::white);
    m_data->m_graphicsEngine.save();
    m_data->m_graphicsEngine.setColor(0, 0, 0, 0);
    m_data->m_graphicsEngine.setPorterDuffRule(DSPD_SRC);
    m_data->m_graphicsEngine.fillRect(rectangle.x(), rectangle.y(), rectangle.width(), rectangle.height());
    m_data->m_graphicsEngine.restore();
}

void GraphicsContext::strokeRect(const FloatRect& rect, float width)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::setLineCap(LineCap lineCap)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::setLineDash(const DashArray&, float dashOffset)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::setLineJoin(LineJoin lineJoin)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::setMiterLimit(float miter)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::setAlpha(float)
{
    notImplemented();
}

static DFBSurfacePorterDuffRule CompositeToPorterDuffRule(CompositeOperator op)
{
    switch(op)
    {
        case CompositeClear:return DSPD_CLEAR;
        case CompositeCopy:return DSPD_SRC;
        case CompositeSourceOver:return DSPD_SRC_OVER;
        case CompositeSourceIn:return DSPD_SRC_IN;
        case CompositeSourceOut:return DSPD_SRC_OUT;
        case CompositeSourceAtop:return DSPD_SRC_ATOP;
        case CompositeDestinationOver:return DSPD_DST_OVER;
        case CompositeDestinationIn:return DSPD_DST_IN;
        case CompositeDestinationOut:return DSPD_DST_OUT;
        case CompositeDestinationAtop:return DSPD_DST_ATOP;
        case CompositeXOR:return DSPD_XOR;
        case CompositePlusDarker:return DSPD_ADD;
        case CompositeHighlight:return DSPD_ADD;
        case CompositePlusLighter:return DSPD_ADD;
        default: return DSPD_SRC_OVER;
    }
}

void GraphicsContext::setCompositeOperation(CompositeOperator op)
{
    if (paintingDisabled())
        return;
    D_INFO(" rule = %d\n", op);
    m_data->m_graphicsEngine.setPorterDuffRule(CompositeToPorterDuffRule(op));
}

void GraphicsContext::beginPath()
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::addPath(const Path& path)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::clip(const Path& path)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::clipOut(const Path& path)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::rotate(float radians)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::scale(const FloatSize& size)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::clipOut(const IntRect& r)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::clipOutEllipseInRect(const IntRect& r)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::fillRoundedRect(const IntRect& r, const IntSize& topLeft, const IntSize& topRight, const IntSize& bottomLeft, const IntSize& bottomRight, const Color& color)
{
    if (paintingDisabled())
        return;

    NotImplemented();
}

void GraphicsContext::setBalExposeEvent(BalEventExpose* expose)
{
    m_expose = expose;
}

BalEventExpose* GraphicsContext::balExposeEvent() const
{
    return m_expose;
}

BalDrawable* GraphicsContext::balDrawable() const
{
    //if (!m_data->expose)
    //    return 0;

    return 0;//GDK_DRAWABLE(m_data->expose->window);
}

float GraphicsContext::transparencyLayer()
{
    if (!m_data->layers.isEmpty())
        return m_data->layers.last();
    else
        return 1.0;
}

IntPoint GraphicsContext::translatePoint(const IntPoint& point) const
{
    NotImplemented();
    return point;
}

void GraphicsContext::strokePath() 
{
    NotImplemented();
}

void GraphicsContext::drawPath()
{
    NotImplemented();
}

void GraphicsContext::fillPath()
{
    NotImplemented();
}

void GraphicsContext::setPlatformFillPattern(Pattern* pattern)
{
    NotImplemented();
}

void GraphicsContext::setPlatformStrokePattern(Pattern* pattern)
{
    NotImplemented();
}

void GraphicsContext::setPlatformFillGradient(Gradient* gradient)
{
    NotImplemented();
}

void GraphicsContext::setPlatformStrokeGradient(Gradient* gradient)
{
    NotImplemented();
}

void GraphicsContext::setPlatformShouldAntialias(bool enable)
{
    if (paintingDisabled())
        return;

    NotImplemented();

    // When true, use the default Cairo backend antialias mode (usually this
    // enables standard 'grayscale' antialiasing); false to explicitly disable
    // antialiasing. This is the same strategy as used in drawConvexPolygon().
    //cairo_set_antialias(m_data->cr, enable ? CAIRO_ANTIALIAS_DEFAULT : CAIRO_ANTIALIAS_NONE);
}

} // namespace WebCore

