/*
 * Copyright (C) 2008 Pleyo.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Pleyo nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PLEYO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL PLEYO OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include "config.h"
#include "owb-config.h"
#include "BitmapImage.h"

#include "TransformationMatrix.h"
#include "FloatRect.h"
#include "GraphicsContext.h"
#include "ImageObserver.h"
#include "PlatformString.h"
#include "SharedBuffer.h"
#include "DebugPrintf.h"

#include <math.h>

// This function loads resources from WebKit
Vector<char> loadResourceIntoArray(const char* name)
{
    WebCore::String fullPath = RESOURCE_PATH;
    WebCore::String extension = ".png";
    fullPath += name;
    fullPath += extension;
    
    RefPtr<WebCore::SharedBuffer> buffer = WebCore::SharedBuffer::createWithContentsOfFile(fullPath);
    Vector<char> array;
    if (buffer)
        array = buffer.get()->buffer();
    return array;
}

namespace WKAL {

bool FrameData::clear(bool clearMetadata)
{
    if (clearMetadata)
        m_haveMetadata = false;

    if (m_frame)
    {
        LOG(Graphics, "release surface %p \n", m_frame);
        OWB_SAFE_RELEASE(m_frame);
        return true;
    }

    return false;
}

BitmapImage::BitmapImage(BalSurface* surface, ImageObserver* observer)
    : Image(observer)
    , m_currentFrame(0)
    , m_frames(0)
    , m_frameTimer(0)
    , m_repetitionCount(cAnimationNone)
    , m_repetitionCountStatus(Unknown)
    , m_repetitionsComplete(0)
    , m_isSolidColor(false)
    , m_animationFinished(true)
    , m_allDataReceived(true)
    , m_haveSize(true)
    , m_sizeAvailable(true)
    , m_decodedSize(0)
    , m_haveFrameCount(true)
    , m_frameCount(1)
{
    initPlatformData();

    // TODO: check to be sure this is an image surface
    int width;
    int height;
    OWB_DFBCHECK(surface->AddRef(surface));
    OWB_DFBCHECK(surface->GetSize(surface, &width, &height)); 
    m_decodedSize = width * height * 4;
    m_size = IntSize(width, height);

    m_frames.grow(1);
    m_frames[0].m_frame = surface;
    checkForSolidColor();
}


void BitmapImage::draw(GraphicsContext* context, const FloatRect& dst, const FloatRect& src, CompositeOperator op)
{
    IDirectFBSurface* image = frameAtIndex(m_currentFrame);
    int image_width;
    int image_height;

    if (!image)
        return;
    OWB_DFBCHECK(image->GetSize(image, &image_width, &image_height));
    
    if (mayFillWithSolidColor())
    {
        fillWithSolidColor(context, dst, solidColor(), op);
        return;
    }

    context->save();
    //context->clip(IntRect(dst)); // don't draw outside this
    // Set the compositing operation.
    if (op == CompositeSourceOver && !frameHasAlphaAtIndex(m_currentFrame))
        context->setCompositeOperation(CompositeCopy);
    else
        context->setCompositeOperation(op);

    DFBGraphicsEngine *cr = context->platformGraphicsEngine();
    IntRect sourceRect(src);
    IntRect destRect(dst);

    // Draw the image.
    if (0 == sourceRect.width())
        sourceRect.setWidth(image_width);
    if (0 == sourceRect.height())
        sourceRect.setHeight(image_height);

    D_INFO("{%d, %d, %d, %d} --> {%d, %d}\n",
            sourceRect.x(),
            sourceRect.y(),
            sourceRect.width(),
            sourceRect.height(),
            destRect.x() + context->origin().x(),
            destRect.y() + context->origin().y());


    cr->stretchBlit(
        destRect.x() + context->origin().x(),
        destRect.y() + context->origin().y(),
        destRect.x() + destRect.width() + context->origin().x() - 1,
        destRect.y() + destRect.height() + context->origin().y() - 1,
        image,
        sourceRect.x(),
        sourceRect.y(),
        sourceRect.x() + sourceRect.width() - 1,
        sourceRect.y() + sourceRect.height() - 1);

    startAnimation();
    context->restore();

    if (imageObserver())
        imageObserver()->didDraw(this);
}

void Image::drawPattern(GraphicsContext* context, const FloatRect& tileRect, const TransformationMatrix& patternTransform,
                        const FloatPoint& phase, CompositeOperator op, const FloatRect& destRect)
{
    if (destRect.isEmpty())
        return;

    IDirectFBSurface* image = nativeImageForCurrentFrame();
    int image_width, image_height;
    if (!image) // If it's too early we won't have an image yet.
        return;
    OWB_DFBCHECK(image->GetSize(image, &image_width, &image_height));

    D_INFO(" tileRect : {%f, %f, %f, %f}, patternTransform = {}, phase = {%f, %f}, destRect = {%f, %f, %f, %f} image = {%d, %d}\n",
            tileRect.x(), tileRect.y(), tileRect.width(), tileRect.height(),
            phase.x(), phase.y(),
            destRect.x(), destRect.y(), destRect.width(), destRect.height(),
            image_width, image_height
            );

    DFBGraphicsEngine *cr = context->platformGraphicsEngine();
    context->save();
    context->setCompositeOperation(op);

    // Check and see if a single draw of the image can cover the entire area we are supposed to tile.
    // save context info
    context->clip(IntRect(destRect)); // don't draw outside this

    IntRect dest(0, 0, image_width, image_height);
    IntRect src(tileRect);

    int xMax = static_cast<int>(context->origin().x() + destRect.x() + destRect.width());
    int yMax = static_cast<int>(context->origin().y() + destRect.y() + destRect.height());

    if (0 == src.width())
        src.setWidth(image_width);
    if (0 == src.height())
        src.setHeight(image_height);

    for (int x = static_cast<int>(destRect.x()); x < xMax; x += image_width)
    {
        for (int y = static_cast<int>(destRect.y()); y < yMax; y += image_height)
        {
            dest.setLocation(IntPoint(x, y) + IntSize(context->origin().x(), context->origin().y()));
            cr->stretchBlit(
                dest.x(),
                dest.y(),
                dest.x() + dest.width() - 1,
                dest.y() + dest.height() - 1,
                image,
                src.x(),
                src.y(),
                src.x() + src.width() - 1,
                src.y() + src.height() - 1);
        }
    }

    context->restore();

    if (imageObserver())
        imageObserver()->didDraw(this);
}

void BitmapImage::initPlatformData()
{
}

void BitmapImage::invalidatePlatformData()
{
}

PassRefPtr<Image> Image::loadPlatformResource(const char *name)
{
    Vector<char> arr = loadResourceIntoArray(name);
    RefPtr<BitmapImage> img = BitmapImage::create();
    RefPtr<SharedBuffer> buffer = SharedBuffer::create(arr.data(), arr.size());
    img->setData(buffer, true);
    return img.release();
}

void BitmapImage::checkForSolidColor()
{
    // FIXME: It's easy to implement this optimization. Just need to check the RGBA32 buffer to see if it is 1x1.
    m_isSolidColor = false;
}

}

