
#include <math.h>

#include "wtf/Vector.h"
#include "DFBGraphicsEngine.h"


// some internal function of DirectFB
extern "C" DFBBoolean dfb_clip_rectangle(const DFBRegion* clip, DFBRectangle* rect);
extern "C" DFBBoolean dfb_clip_line(const DFBRegion* clip, DFBRegion *line);

//
namespace WKAL {

#define CLAMP(x, low, high) (((x) > (high)) ? (high) : (((x) < (low)) ? (low) : (x)))
#define MAX(x, y) ((x) > (y) ? (x) : (y))
#define MIN(x, y) ((x) < (y) ? (x) : (y))

#if 0
#define DFBARC          0x0001
#define DFBOUTLINE      0x0002
#define DFBARCOUTLINE   0x0003
#define DFBPIE          0x0004
#define DFBELLIPSE      0x0008
#define DFBELLIPSEFILL  0x0010
#endif

#define qcos QCOS
#define qsin QSIN

static inline double QCOS(double a)
{
    return cos (a * M_PI / 180.);
}
static inline double QSIN(double a)
{
    return sin (a * M_PI / 180.);
}

static inline bool dfb_clip_rectangle_indirect(const DFBRegion *clip, int *x, int *y, int *w, int *h)
{
    DFBRectangle rect = {*x, *y, *w, *h};

    if (DFB_TRUE == dfb_clip_rectangle(clip, &rect) && rect.w > 0 && rect.h > 0)
    {
        *x = rect.x;
        *y = rect.y;
        *w = rect.w;
        *h = rect.h;
        return true;
    }
    return false;
}

DFBGraphicsEngine::DFBGraphicsEngine(PlatformGraphicsContext* cr)
        : m_surface(cr)
        , m_surfaceDataBuffer(0)
        , m_surfacePitch(0)
        , extraAlpha(255)
        , m_surfacePorterDuffRule(DSPD_SRC_OVER)
        , m_surfaceDrawingFlags(DSDRAW_BLEND)
        , m_surfaceBlittingFlags(DSBLIT_BLEND_ALPHACHANNEL)
{
    ASSERT(m_surface);
 
    OWB_DFBCHECK(m_surface->AddRef(m_surface));
    OWB_DFBCHECK(m_surface->GetSize(m_surface, &m_surfaceWidth, &m_surfaceHeight));
    OWB_DFBCHECK(m_surface->GetPixelFormat(m_surface, &m_surfaceFormat));
    
    m_surfaceColor.a = m_surfaceColor.r = m_surfaceColor.g = m_surfaceColor.g = 255;
    m_surfaceClip.x1 = 0;
    m_surfaceClip.y1 = 0;
    m_surfaceClip.x2 = m_surfaceWidth - 1;
    m_surfaceClip.y2 = m_surfaceHeight - 1;
    m_clipped = false;

    OWB_DFBCHECK(m_surface->SetPorterDuff(m_surface, m_surfacePorterDuffRule));
    OWB_DFBCHECK(m_surface->SetDrawingFlags(m_surface, m_surfaceDrawingFlags));
    OWB_DFBCHECK(m_surface->SetBlittingFlags(m_surface, m_surfaceBlittingFlags));
    OWB_DFBCHECK(m_surface->SetColor(m_surface, m_surfaceColor.r, m_surfaceColor.g, m_surfaceColor.b, m_surfaceColor.a));
}

DFBGraphicsEngine::~DFBGraphicsEngine()
{
    OWB_SAFE_RELEASE(m_surface);
}

void DFBGraphicsEngine::setClipRect(int x, int y, int w, int h)
{
    if ( x == m_surfaceClip.x1 
      && y == m_surfaceClip.y1
      && ((x + w - 1) == m_surfaceClip.x2)
      && ((y + h - 1) == m_surfaceClip.y2)
      )
        return;

    //D_INFO(" {%d, %d, %d, %d}\n", x, y, w, h);

    //Catch a completely clipped rectangle here so that we no longer attempt any drawing
    //operations until the clip area is set again
    //Complete clipping is defined when either the width or height are <= zero, or the clipping
    //area is outside the screen area
    //In these cases, directfb cannot handle the clip so we handle it here in the native porting layer.
	if (x >= m_surfaceWidth || y >= m_surfaceHeight ||
        w <= 0 || h <= 0 ||
        ((x + w) < 0) || ((y + h) < 0) 
    )
	{
        //set the clipped status
        m_clipped = true;

        return;
    }
 
    m_clipped = false;
    
	if (x < 0)
	{
		w += x;
		x = 0;
	}

	if (y < 0)
	{
		h += y;
		y = 0;
	}

	if (x + w > m_surfaceWidth)
		w = m_surfaceWidth - x;

	if (y + h > m_surfaceHeight)
		h = m_surfaceHeight - y;

	m_surfaceClip.x1 = x;
	m_surfaceClip.y1 = y;
	m_surfaceClip.x2 = x + w - 1;
	m_surfaceClip.y2 = y + h - 1;

    OWB_DFBCHECK(m_surface->SetClip(m_surface, &m_surfaceClip));
}

void DFBGraphicsEngine::removeClip()
{
    setClipRect(0, 0, m_surfaceWidth, m_surfaceHeight);
}

void DFBGraphicsEngine::save()
{
    DFBGraphicsEngineState state(m_surfaceColor, extraAlpha,
            m_surfacePorterDuffRule,
            m_surfaceDrawingFlags,
            m_surfaceBlittingFlags,
            m_surfaceClip);

    m_stack.append(state);
}

void DFBGraphicsEngine::restore()
{
    if (m_stack.isEmpty())
    {
        LOG(Graphics, "ERROR DFBGraphicsEngine::restore stack is empty");
        return;
    }

    DFBGraphicsEngineState state = m_stack.last();
    m_stack.removeLast();

    setColor(state.color);
    extraAlpha = state.extraAlpha;
    setPorterDuffRule(state.duffRule);
    setDrawingFlags(state.drawingFlags);
    setBlittingFlags(state.blittingFlags);
    setClipRect(state.clipRegion.x1, state.clipRegion.y1,
            state.clipRegion.x2 - state.clipRegion.x1 + 1,
            state.clipRegion.y2 - state.clipRegion.y1 + 1);
}

DFBColor DFBGraphicsEngine::unpackPixel(DFBSurfacePixelFormat format, unsigned int pixel)
{
    DFBColor color = {0, 0, 0, 0};
    NotImplemented();
    return color;
}

unsigned int DFBGraphicsEngine::packPixel(DFBSurfacePixelFormat format, DFBColor color)
{
    NotImplemented();
    return 0;
}

bool DFBGraphicsEngine::surfaceLock()
{
    if (DFB_OK == m_surface->Lock(m_surface, DSLF_WRITE, (void**)&m_surfaceDataBuffer, &m_surfacePitch))
    {
        return true;
    } else {
        printf (" failed to lock surface \n");
    }
    return false;
}

void DFBGraphicsEngine::surfaceUnlock()
{
    if (DFB_OK == m_surface->Unlock(m_surface))
    {
        m_surfaceDataBuffer = NULL;
        m_surfacePitch = 0;
    } else {
        printf (" failed to unlock surface \n");
    }
}

void DFBGraphicsEngine::drawPixel(int x, int y)
{
    //D_INFO(" {%d, %d}\n", x, y);
    if (m_clipped 
        || !m_surface 
        || x < m_surfaceClip.x1
        || x > m_surfaceClip.x2
        || y < m_surfaceClip.y1
        || y > m_surfaceClip.y2)
        return;

    OWB_DFBCHECK(m_surface->FillRectangle(m_surface, x, y, 1, 1));
}

DFBColor DFBGraphicsEngine::readPixel(int x, int y)
{
    DFBColor color;
    //D_INFO(" {%d, %d}\n", x, y);

	if (x < 0) x = 0;
	if (y < 0) y = 0;
	if (x >= m_surfaceWidth) x = m_surfaceWidth - 1;
	if (y >= m_surfaceHeight) y = m_surfaceHeight - 1;
 
    if (surfaceLock())
    {
        unsigned int *dst32;

        dst32 = (unsigned int*)(m_surfaceDataBuffer + y * m_surfacePitch);
        color = unpackPixel(m_surfaceFormat, dst32[x]);

        surfaceUnlock();
    }

    return color;
}

void DFBGraphicsEngine::drawRect(int x, int y, int w, int h)
{
    /* Draw the Rectangle if needToDraw is set to TRUE */
    if (m_clipped || !m_surface)
        return;

    if (dfb_clip_rectangle_indirect(&m_surfaceClip, &x, &y, &w, &h) == false)
        return;
        
    //D_INFO(" {%d, %d, %d, %d}\n", x, y, w, h);
    OWB_DFBCHECK(m_surface->DrawRectangle(m_surface, x, y, w, h));
}

void DFBGraphicsEngine::fillRect(int x, int y, int w, int h)
{
    if (m_clipped || !m_surface)
        return;

    if (dfb_clip_rectangle_indirect(&m_surfaceClip, &x, &y, &w, &h) == false)
        return;

    //D_INFO(" {%d, %d, %d, %d}\n", x, y, w, h);
    OWB_DFBCHECK(m_surface->FillRectangle(m_surface, x, y, w, h));
}


void DFBGraphicsEngine::drawLine(int x1, int y1, int x2, int y2)
{
    //D_INFO(" {%d, %d, %d, %d}\n", x1, y1, x2, y2);
    if (m_clipped || !m_surface)
        return;

    /* Coordinates can't go beyond 16 bits */
    x1 = CLAMP(x1, -32768, 32767);
    y1 = CLAMP(y1, -32768, 32767);
    x2 = CLAMP(x2, -32768, 32767);
    y2 = CLAMP(y2, -32768, 32767);

    OWB_DFBCHECK(m_surface->DrawLine(m_surface, x1, y1, x2, y2));
}

void DFBGraphicsEngine::drawPolygon(DFBPoint* points, int n)
{
    //D_INFO("\n");
    if (m_clipped || !m_surface)
        return;

    for(int i = 0; i < n - 1; i++)
    {
        drawLine(points[i].x, points[i].y, points[i+1].x, points[i+1].y);
    }
}

/////////////// Fill Polygon ////

typedef struct
{
    int x1, y1, x2, y2;
    double x, m;
} edge_t;

static int
edge_cmp (const void *lvp, const void *rvp)
{
    /* convert from void pointers to structure pointers */
    const edge_t *lp = (const edge_t *) lvp;
    const edge_t *rp = (const edge_t *) rvp;
                                                                                
    /* if the minimum y values are different, sort on minimum y */
    if (lp->y1 != rp->y1)
        return lp->y1 - rp->y1;
                                                                                
    /* if the current x values are different, sort on current x */
    if (lp->x < rp->x)
        return -1;
    else if (lp->x > rp->x)
        return +1;
                                                                                
    /* otherwise they are equal */
    return 0;
}

#define swap(a,b) do { a ^= b; b ^= a; a ^= b; } while (0)

void DFBGraphicsEngine::fillPolygon(DFBPoint* pointtable, int count)
{
    edge_t *get;
    int nge = 0;
    int cge = 0;

    edge_t *aet;
    int nae = 0;
    int i, y, newCount;
    DFBPoint *newpointtable = NULL;

    for(int i = 0; i < count; i++)
    {
        //D_INFO("{%d, %d} \n", pointtable[i].x, pointtable[i].y);
    }

    if (m_clipped)
        return; /* No need to draw */

    if (count < 3)
        return; /* Has to be a triangle atleast */

    if(isPolyVisible(pointtable, count) != true)
    {
        /* All points are not visible, try to clip! */
        newpointtable = (DFBPoint*)calloc(count * 2, sizeof(DFBPoint));

        if (newpointtable != NULL)
        {
            if (clipPoly(pointtable, count, newpointtable, &newCount) == false)
            {
                /* Not visible at all, quit! */
                free(newpointtable);
                return;
            }
            pointtable = newpointtable;
            count = newCount;
        }
    }

    /*
     ** at this stage all points in the poly have the potential to be visible
     */
    get = (edge_t *) calloc (count, sizeof (edge_t));
    aet = (edge_t *) calloc (count, sizeof (edge_t));

    if ((get == 0) || (aet == 0))
    {
        /* error, couldn't allocate one or both of the needed tables */
        if (get)
            free (get);
        if (aet)
            free (aet);
        if (newpointtable)
            free (newpointtable);
        return;
    }
    /* setup the global edge table */
    for (i = 0; i < count; ++i)
    {
        get[nge].x1 = pointtable[i].x;
        get[nge].y1 = pointtable[i].y;
        get[nge].x2 = pointtable[(i + 1) % count].x;
        get[nge].y2 = pointtable[(i + 1) % count].y;
        if (get[nge].y1 != get[nge].y2)
        {
            if (get[nge].y1 > get[nge].y2)
            {
                swap(get[nge].x1, get[nge].x2);
                swap(get[nge].y1, get[nge].y2);
            }
            get[nge].x = get[nge].x1;
            get[nge].m = get[nge].x2 - get[nge].x1;
            get[nge].m /= get[nge].y2 - get[nge].y1;
            ++nge;
        }
    }

    qsort (get, nge, sizeof (get[0]), edge_cmp);

    /* start with the lowest y in the table */
    y = get[0].y1;

    /* 
     * Optimization steps:
     * - Lock the surface 
     * - Call DirectFB DrawLine API directly instead of DFBDrawLine.
     */
     do
     {
         /* add edges to the active table from the global table */
         while((nge > 0) && (get[cge].y1 == y))
         {
             aet[nae] = get[cge++];
             --nge;
             aet[nae++].y1 = 0;
         }

         qsort (aet, nae, sizeof (aet[0]), edge_cmp);

         /* using odd parity, render alternating line segments */
         for(i = 1; i < nae; i += 2)
         {
             int l = (int) aet[i - 1].x;
             int r = (int) aet[i].x;
             if (r > l)
             {
                 OWB_DFBCHECK(m_surface->DrawLine(m_surface, l, y, r - 1, y));
             }
         }

         /* prepare for the next scan line */
         ++y;

         /* remove inactive edges from the active edge table */
         /* or update the current x position of active edges */
         for(i = 0; i < nae; ++i)
         {
             if (aet[i].y2 == y)
                 aet[i--] = aet[--nae];
             else
                 aet[i].x += aet[i].m;
         }

         /* keep doing this while there are any edges left */
     } while((nae > 0) || (nge > 0));

    /* all done, free the edge tables */
    free (get);
    free (aet);

    if (newpointtable)
        free (newpointtable);

    return;
}



////
bool DFBGraphicsEngine::isLineVisible(int x1, int y1, int x2, int y2)
{
    return true;
}

typedef struct 
{
    float fX;
    float fY;
} fPOINT;

bool DFBGraphicsEngine::clipPoly(DFBPoint* pointtable, int count, DFBPoint* newPointtable, int *newCount)
{
    float Ax, Bx, Ay, By;
    int sIns, pIns;
    int i, j, n;
    int clipX1 = m_surfaceClip.x1;
    int clipY1 = m_surfaceClip.y1;
    int clipX2 = m_surfaceClip.x2;
    int clipY2 = m_surfaceClip.y2;
    fPOINT *temppointtable1 = NULL;
    fPOINT *temppointtable2 = NULL;

    temppointtable1 = (fPOINT*)alloca(count * 2 * sizeof (fPOINT));
    temppointtable2 = (fPOINT*)alloca(count * 2 * sizeof (fPOINT));

    if (!(temppointtable1 && temppointtable2))
    {
        printf(" Out Of Memory \n");
        return false;  /* can't check it */
    }

    /* Clip against x = clipX2 */
    n = count;

    if (n == 0)
        return false;

    Bx = pointtable[n - 1].x;
    By = pointtable[n - 1].y;

    for (i = 0, j = 0; i < n; i++)
    {
        Ax = Bx;
        Ay = By;
        Bx = pointtable[i].x;
        By = pointtable[i].y;

        sIns = Ax <= (float)clipX2;
        pIns = Bx <= (float)clipX2;
        if (sIns != pIns)
        {
            temppointtable1[j].fX = (float)clipX2;
            temppointtable1[j++].fY = Ay + (By - Ay) * ((float)clipX2 - Ax) / (Bx - Ax);
        }

        if (pIns)
        {
            temppointtable1[j].fX = Bx;
            temppointtable1[j++].fY = By;
        }
    }

    /* Clip against x = clipX1 */
    n = j;
    if (n == 0)
        return false;

    Bx = temppointtable1[n - 1].fX;
    By = temppointtable1[n - 1].fY;

    for (i = 0, j = 0; i < n; i++)
    {
        Ax = Bx;
        Ay = By;
        Bx = temppointtable1[i].fX;
        By = temppointtable1[i].fY;

        sIns = Ax >= (float)clipX1;
        pIns = Bx >= (float)clipX1;

        if (sIns != pIns)
        {
            temppointtable2[j].fX = (float)clipX1;
            temppointtable2[j++].fY = Ay + (By - Ay) * ((float)clipX1 - Ax) / (Bx - Ax);
        }
        if (pIns)
        {
            temppointtable2[j].fX = Bx;
            temppointtable2[j++].fY = By;
        }
    }

    /* Clip against y = clipY2 */
    n = j;
    if (n == 0)
        return false;

    Bx = temppointtable2[n - 1].fX;
    By = temppointtable2[n - 1].fY;

    for (i = 0, j = 0; i < n; i++)
    {
        Ax = Bx;
        Ay = By;
        Bx = temppointtable2[i].fX;
        By = temppointtable2[i].fY;

        sIns = Ay <= (float)clipY2;
        pIns = By <= (float)clipY2;
        if (sIns != pIns)
        {
            temppointtable1[j].fX = Ax + (Bx - Ax) * ((float)clipY2 - Ay) / (By - Ay);
            temppointtable1[j++].fY = (float)clipY2;
        }
        if (pIns)
        {
            temppointtable1[j].fX = Bx;
            temppointtable1[j++].fY = By;
        }
    }

    /* Clip against y = clipY1 */
    n = j;
    if (n == 0)
        return false;

    Bx = temppointtable1[n - 1].fX;
    By = temppointtable1[n - 1].fY;

    for (i = 0, j = 0; i < n; i++)
    {
        Ax = Bx;
        Ay = By;
        Bx = temppointtable1[i].fX;
        By = temppointtable1[i].fY;

        sIns = Ay >= (float)clipY1;
        pIns = By >= (float)clipY1;
        if (sIns != pIns)
        {
            newPointtable[j].x = Ax + (Bx - Ax) * ((float)clipY1 - Ay) / (By - Ay);
            newPointtable[j++].y = clipY1;
        }
        if (pIns)
        {
            newPointtable[j].x = Bx;
            newPointtable[j++].y = By;
        }
    }

    if (j == 0)
        return false;

    newCount[0] = j;

    return true;
}

bool DFBGraphicsEngine::isPolyVisible(DFBPoint *pointtable, int count)
{
    if (m_clipped)
        return false;

	for (int i = 0; i < count; i++)
	{
		if (!(pointtable[i].x >= m_surfaceClip.x1
              && pointtable[i].x <= m_surfaceClip.x2 
              && pointtable[i].y >= m_surfaceClip.y1
              && pointtable[i].y <= m_surfaceClip.y2))
			return false;
	}
	return true;
}

/// arc

/* argument holder for pie, arc and ellipse functions*/

#define ABS(a) (((a)<0) ? -(a) : (a))

static bool clipCodeRA(int iRx1, int iRy1, int iRx2, int iRy2, int Cx, int Cy, int iRad)
{
    float fRx1, fRy1, fRx2, fRy2, fRad2;
    float fRad = (double) iRad;

    fRad2 = fRad * fRad;

    /* Translate coordinates, placing C at the origin. */
    fRx1 = iRx1 - Cx;
    fRy1 = iRy1 - Cy;
    fRx2 = iRx2 - Cx;
    fRy2 = iRy2 - Cy;

    if (fRx2 < 0)               /* R to left of circle center */
	{
		if (fRy2 < 0)           /* R in lower left corner */
		{
			if ((fRx2 * fRx2 + fRy2 * fRy2) < fRad2) 
				return true;
			else 
				return false;
		}
        else if (fRy1 > 0)      /* R in upper left corner */
		{
            if ((fRx2 * fRx2 + fRy1 * fRy1) < fRad2)
				return true;
			else
				return false;
		}
        else                    /* R due West of circle */
		{
            if (ABS (fRx2) < fRad)
				return true;
			else
				return false;
		}
	}
    else if (fRx1 > 0)          /* R to right of circle center */
	{
        if (fRy2 < 0)           /* R in lower right corner */
		{
			if ((fRx1 * fRx1) < fRad2)
				return true;
			else
				return false;
		}
        else if (fRy1 > 0)      /* R in upper right corner */
		{
            if ((fRx1 * fRx1 + fRy1 * fRy1) < fRad2)
				return true;
			else
				return false;
		}
        else                    /* R due East of circle */
		{
            if (fRx1 < fRad)
				return true;
			else
				return false;
		}
	}
    else if (fRy2 < 0)    /* R due South of circle */
	{
       	if (ABS (fRy2) < fRad) 
			return true; 
		else
		    return false;
	}
    else if (fRy1 > 0)          /* R due North of circle */
	{
       	if (fRy1 < fRad)
			return true; 
		else 
			return false;
	}
    else                        /* R contains circle centerpoint */
	{
        return true;
	}
}

static inline bool isArcRectVisible(DFBRegion *clip, int arcX, int arcY, int radiusX, int radiusY)
{
	int clipX1 = clip->x1, clipY1 = clip->y1, clipX2 = clip->x2, clipY2 = clip->y2;

	if (clipX2 < clipX1 || clipY2 < clipY1)
		return false;

	//If the whole arc is within the visible area then return TRUE, else more
	//checks are necessary to determine whether the arc is partially visible or
	//totally invisible.
	if(!((arcX - radiusX >= clipX1) && (arcY - radiusY >= clipY1) &&
		(arcX + radiusX < clipX2) && (arcY + radiusY < clipY2)))
	{
		//is any rectangle enclosing the arc visible?
		if (!((arcX + radiusX < clipX1) || (arcY + radiusY < clipY1) || 
			 ((arcX - radiusX) >= clipX2) || ((arcY - radiusY) >= clipY2)))
		{
			//the arc may still be totally visible, but more checks are needed.
			//is any of the circle with larger radius visible?
			int maxR = (radiusX > radiusY) ? radiusX : radiusY;

			if (clipCodeRA(clipX1, clipY1, clipX2, clipY2, arcX, arcY, maxR))
			{
				return true; //partially visible
			}
		}
		return false;
	}

	//the arc is visible
    return true;
}

static inline bool doesRectIntersectClip(DFBRegion *clip, int x1, int y1, int x2, int y2)
{
	if (x1 >= clip->x2 || y1 >= clip->y2 || x2 < clip->x1 || y2 < clip->y1)
		return false;

	return true;
}

/*
 * Clip a line segment for arc or pie drawing.
 * Returns 0 if line is clipped or on acceptable side, 1 if it's vertically
 * on other side, otherwise 3.
 */
static int clip_line (int xe, int ye, int dir, int y, int * x0, int * x1)
{
    /* hline on the same vertical side with the given edge? */
    if ((y >= 0 && ye >= 0) || (y < 0 && ye < 0))
    {
        int x;
                                                                                
        if (ye == 0)
            x = xe;
        else
            x = (int) (long) xe *y / ye;
			
		if (x >= *x0 && x <= *x1)
        {
            if (dir > 0)
                *x0 = x;
            else
                *x1 = x;
            return 0;
        }
        else
        {
            if (dir > 0)
            {
                if (x <= *x0)
                    return 0;
            }
            else
            {
                if (x >= *x1)
                    return 0;
            }
        }
        return 3;
	}
	return 1;
}

/* relative offsets, direction from left to right. */
void DFBGraphicsEngine::draw_line(SLICE * slice, int x0, int y, int x1)
{
    int dbl = (slice->adir > 0 && slice->bdir < 0);
    int discard, ret;
    int x2 = x0, x3 = x1;
         
    if (y == 0)
    {
		if (slice->fill == false)
        {
        	ret = clip_line (slice->ax, slice->ay, slice->adir, y, &x0, &x1);
        	switch (ret)
            {
			case 0:
            	    if ((slice->bdir < 0) || 
				    (slice->adir > 0 && slice->bdir > 0 && 
				     x0 > 0 ))
	            			drawPixel(slice->x0 + x1, slice->y0 +y);
    	        		break;
                                                                                
            		case 1:
            			if ((slice->bdir < 0 && x0 < 0) || 
				    (slice->bdir > 0 && y > 0) ||
                  		    (slice->adir > 0 && slice->bdir > 0 && 
				     x0 > 0 ) || (slice->adir > 0 && 
				     slice->bdir < 0 && x0 > 0 ))
            				drawPixel(slice->x0 + x1, slice->y0 +y);
	            		break;
                                                                                
            		case 2:
				            drawPixel(slice->x0 + x1, slice->y0 +y);
    	        		break;
                                                                                
            		case 3:
            			if (slice->bdir < 0 && x0 < 0)
                			drawPixel(slice->x0 + x1, slice->y0 +y);
        	    		break;

            		default:
              			break;
            		}
                                                                                
          	return;
        }
        /* edges on different sides */
        if ((slice->ay <= 0 && slice->by >= 0) || 
	    (slice->ay >= 0 && slice->by <= 0))
        {
	    if (slice->adir < 0)
            {
            	if (x1 > 0)
                	x1 = 0;
            }
            if (slice->bdir > 0)
            {
                if (x0 < 0)
                    x0 = 0;
            }
        }
        else
        {
            if (!dbl)
            {
                /* FIXME leaving in draws dot in center */
            	drawPixel(slice->x0, slice->y0);
                return;
            }
	}
      	drawLine(slice->x0 + x0, slice->y0, slice->x0 + x1, slice->y0);
        return;
    }
    /* clip left edge / line */
    ret = clip_line (slice->ax, slice->ay, slice->adir, y, &x0, &x1);
                                                                                
    if (dbl)
    {
    	if (!ret)
        {
        	/* edges separate line to two parts */
         	drawLine(slice->x0 + x0, slice->y0 + y, slice->x0 + x1, slice->y0 + y);
            x0 = x2;
            x1 = x3;
        }
    }
    else
	{
        if (ret > 1)
        {
            return;
        }
    }
                                                                                
    discard = ret;
    ret = clip_line (slice->bx, slice->by, slice->bdir, y, &x0, &x1);
                                                                                
    discard += ret;
    if (discard > 2 && !(dbl && ret == 0 && discard == 3))
    {
        return;
    }
    if (discard == 2)
    {
        /* line on other side than slice */
        if (slice->adir < 0 || slice->bdir > 0)
        {
            return;
	}
    }
    drawLine(slice->x0 + x0, slice->y0 + y, slice->x0 + x1, slice->y0 + y);
}

void DFBGraphicsEngine::GdArc (int x0, int y0, int rx, int ry,
       int ax, int ay, int bx, int by, bool fill) 
{
    int adir, bdir;
    SLICE slice;
    double xSquared, ySquared;
    int i, x, y;
    double dlenX;
    int cenX, cenY, ilenX;
    int NewX;

    if (rx <= 0 || ry <= 0)
        return;

    /*
     * If the arc is outside the visible area then return
     */
    if (isArcRectVisible(&m_surfaceClip, x0, y0, rx, ry) == false)
		return;

    if (doesRectIntersectClip(&m_surfaceClip, x0-rx, y0-ry, x0+rx, y0+ry) == false)
		return;	

    /*
     * Calculate right/left side clipping, based on quadrant.
     * dir is positive when right side is filled and negative when
     * left side is to be filled.
     *
     * >= 0 is bottom half
     */
    if (ay >= 0)
        adir = 1;
    else
        adir = -1;
                                                                                
    if (by >= 0)
        bdir = -1;
    else
        bdir = 1;
                                                                                
    /*
     * This fix is required to make the clip_line algorithm work. Doing this 
     * may cause some half circles to be outlined with a slightly bent line on 
     * the x axis.
     */
    if (ay == 0)
        ++ay;
    if (by == 0)
        ++by;
                                                                                
    /* swap rightmost edge first */
    if (bx > ax)
    {
        swap(ax, bx);
        swap(ay, by);
        swap(adir, bdir);
    }
                                                                                
    slice.x0 = x0;
    slice.y0 = y0;
    slice.rx = rx;
    slice.ry = ry;
    slice.ax = ax;
    slice.ay = ay;
    slice.bx = bx;
    slice.by = by;
    slice.adir = adir;
    slice.bdir = bdir;
    slice.fill = fill;
                                                                                
    x = 0; // x is relative to enclosing rect;
    y = 0; // y is relative to enclosing rect;
                                                                                
    if (fill) {
    	NewX = x-1;
	    cenX = x;
	}
    else {
    	NewX = x;
    	cenX = x+1;
    }
                                                                                
    ilenX = rx;
    cenY = y;
    xSquared = rx*rx;
    ySquared = ry*ry;
	for(i=0;i<=ry;i++) {
    	if (ry == i) {
        	dlenX = ilenX/2;
      	}
	    else {
        	dlenX=sqrt(xSquared*(ySquared-(i*i)));
	        dlenX=dlenX/ry;
    	}
	    if (fill) {
        	/*
	        ** fillArc
    	    	*/
        	draw_line(&slice, cenX-dlenX, cenY-i, ilenX+NewX);     /* N */
	        if (i > 1) {
    	    	draw_line(&slice, cenX-dlenX, cenY+i-1, ilenX+NewX);   /* S */
	        }
    	}
      	else {
		/*
	        ** drawArc
    	    	*/
                                                                                
        	/* y max */
	        if (i == ry) {
    	    	draw_line(&slice, -ilenX+NewX, cenY-i, ilenX+NewX); /* N */
	            draw_line(&slice, -ilenX+NewX, cenY+i, ilenX+NewX); /* S */
	        }
    	    else {
        		draw_line(&slice, -ilenX+NewX, cenY-i, cenX-dlenX-1);   /* NW */
		        draw_line(&slice, cenX+dlenX-1, cenY-i, ilenX+NewX);  /* NE */
                                                                                
		        if (i != 0) {
		            draw_line(&slice, -ilenX+NewX, cenY+i, cenX-dlenX-1);  /* SW */
     		        draw_line(&slice, cenX+dlenX-1, cenY+i, ilenX+NewX); /* SE */
	        	}
	        }
      	}
      	ilenX = dlenX;
	}
}

void DFBGraphicsEngine::drawArc(int x, int y, int width, int height, int startAngle, int endAngle, bool fill)
{
    int ax,ay,bx,by;
    float a,b;

    //D_INFO("{%d, %d, %d, %d, %d, %d, %d}\n", x, y, width, height, startAngle, endAngle, fill);
    if (m_clipped || !m_surface)
        return; /* No need to draw */

    if ((endAngle - 23040) >= startAngle)
    {
        drawEllipse(x, y, width, height, fill);
        return;
    }

    /* calculate pie edge offsets from center to the ellipse rim */
    ax = qcos (startAngle / 64.) * width;
    bx = qcos (endAngle / 64.) * width;

    a = -qsin (startAngle / 64.);
    b = -qsin (endAngle / 64.);
    ay = a * height;
    by = b * height;
    /* call integer routine */
    GdArc (x, y, width, height, ax, ay, bx, by, fill);
}

void DFBGraphicsEngine::drawEllipse(int x, int y, int width, int height, bool fill)
{
    double xSquared, ySquared;
    int i;
    double dlenX;
    int cenX, cenY, ilenX;
    int NewX;

    //D_INFO("{%d, %d, %d, %d, %d}\n", x, y, width, height, fill);
	if (m_clipped || !m_surface)
		return; /* No need to draw */
                                                                                
	if (width < 0 || height < 0)
		return;

	if (isArcRectVisible(&m_surfaceClip, x, y, width, height) == false)
		return;

	if (doesRectIntersectClip(&m_surfaceClip, x-width, y-height, x+width, y+height) == false)
		return;

    if (fill) 
    {
    	NewX = x-1;
        cenX = x;
    } else {
        NewX = x;
        cenX = x+1;
    }
                                                                                
    ilenX = width;
    cenY = y;
    xSquared = width*width;
    ySquared = height*height;
                                                                                
    for(i = 0; i <= height; i++)
    {
	    if (height == i)
        {
    		dlenX = ilenX/2;
	    }
    	else {
    		dlenX=sqrt(xSquared*(ySquared-(i*i)));
	        dlenX=dlenX/height;
    	}

	    if (fill)
        {
    		/*
	    	 ** fillOval
	    	 */
    	    drawLine(cenX-dlenX, cenY-i, ilenX+NewX, cenY-i);       /* N */
        	if (i > 1)
            {
	        	drawLine(cenX-dlenX, cenY+i-1, ilenX+NewX, cenY+i-1); /* S */
    	    }
	    }
    	else {
	    	/*
		     ** drawOval
    		 */
                                                                                
	        /* y max */
    	    if (i == height) {
        		drawLine(-ilenX+NewX, cenY-i, ilenX+NewX, cenY-i); /* N */
            	drawLine(-ilenX+NewX, cenY+i, ilenX+NewX, cenY+i); /*
S */
	        }
		    else {
    			drawLine(-ilenX+NewX, cenY-i, cenX-dlenX, cenY-i);   /* NW */
            	drawLine(cenX+dlenX-1, cenY-i, ilenX+NewX, cenY-i);  /* NE */
                                                                                
	           	if (i != 0) {
    	    		drawLine(-ilenX+NewX, cenY+i, cenX-dlenX, cenY+i);  /* SW */
	    	        drawLine(cenX+dlenX-1, cenY+i, ilenX+NewX, cenY+i); /* SE */
    	    	}
	        } /* end if-else(i==Height) */
    	} /* end of if-else (fill)*/
	    ilenX = dlenX;
    } /* end of (for) */
}

// bit blt

void DFBGraphicsEngine::bitBlit(int dx, int dy, int width, int height, 
        IDirectFBSurface *source, int sx, int sy)
{
    int source_width;
    int source_height;

	if (m_clipped || !m_surface || !source)
		return;
	
    OWB_DFBCHECK(source->GetSize(source, &source_width, &source_height));
	if (sx >= source_width || sy >= source_height)
		return;
	if (dx >= m_surfaceWidth || dy >= m_surfaceHeight)
		return;

	/* 
	 * Clip blit area to source screen size. Don't really have any 
	 * clip area set up for source.
	 */
	if (sx < 0)
	{
		if (sx + width < 0)
			return;
		width += sx;
		dx -= sx;
		sx = 0;
	}

	if (sy < 0)
	{
		if (sy + height < 0)
			return;
		height += sy;
		dy -= sy;
		sy = 0;
	}

	if (sx + width > source_width)
		width = source_width - sx;

	if (sy + height > source_height)
		height = source_height - sy;

	if (dx < m_surfaceClip.x1)
	{
		if (dx + width < m_surfaceClip.x1)
			return;
		width -= m_surfaceClip.x1 - dx;
		sx += m_surfaceClip.x1 - dx;
		dx = m_surfaceClip.x1;
	}
	
	if (dy < m_surfaceClip.y1)
	{
		if (dy + height < m_surfaceClip.y1)
			return;
		height -= m_surfaceClip.y1 - dy;
		sy += m_surfaceClip.y1 - dy;
		dy = m_surfaceClip.y1;	
	}

	if (dx + width - 1 > m_surfaceClip.x2)
		width -= (dx + width - 1) - m_surfaceClip.x2;
	if (dy + height - 1 > m_surfaceClip.y2)
		height -= (dy + height - 1) - m_surfaceClip.y2;

	if (width <= 0 || height <= 0)
		return;

	if (sx >= source_width || sy >= source_height)
		return;

	if (dx >= m_surfaceWidth || dy >= m_surfaceHeight)
		return;
	
    /* In all of these conditions it would be Blit with SRC */
    DFBRectangle srcRect;

    srcRect.x = sx;
    srcRect.y = sy;
    srcRect.w = width;
    srcRect.h = height;

    //D_INFO("{%d, %d, %d, %d, %p, %d, %d}\n", dx, dy, width, height, source, sx, sy);
    OWB_DFBCHECK(m_surface->Blit(m_surface, source, &srcRect, dx, dy));
}

void DFBGraphicsEngine::stretchBlit(int dx1, int dy1, 
		   int dx2, int dy2, IDirectFBSurface *source, 
		   int sx1, int sy1, int sx2, int sy2)
{
	int temp;
	int srcWidth, srcHeight, dstWidth, dstHeight;
	int clippedDstX1, clippedDstY1, clippedDstX2, clippedDstY2;
	bool flipX, flipY;
		
    //D_INFO("{%d, %d, %d, %d, %p, %d, %d, %d, %d}\n", dx1, dy1, dx2, dy2, source, sx1, sy1, sx2, sy2);

    if (m_clipped || !m_surface)
        return;
	
	dstWidth = dx2 - dx1;

    if (dstWidth < 0)
    {
        dstWidth--;
        swap(dx1, dx2);
    }
    else
        dstWidth++;

	dstHeight = dy2 - dy1;

    if (dstHeight < 0)
    {
        dstHeight--;
        swap(dy1, dy2);
    }
    else
        dstHeight++;

    srcWidth = sx2 - sx1;

    if (srcWidth < 0)
    {
        srcWidth--;
        swap(sx1, sx2);
    }
    else
        srcWidth++;

    srcHeight = sy2 - sy1;

    if (srcHeight < 0)
    {
        srcHeight--;
        swap(sy1, sy2);
    }
    else
        srcHeight++;
	
	if (dx1 > dx2 || dy1 > dy2 || sx1 > sx2 || sy1 > sy2)
		return;

	/* Use Blit if we are not Stretching or Flipping */
	if (srcWidth == dstWidth && srcHeight == dstHeight)
    {
		bitBlit(dx1, dy1, abs(srcWidth), abs(srcHeight), source, sx1, sy1);
		return;
	}
 
	/* Clip destination area to clip area */
	clippedDstX1 = MAX(m_surfaceClip.x1, dx1);
	clippedDstY1 = MAX(m_surfaceClip.y1, dy1);
	clippedDstX2 = MIN(m_surfaceClip.x2, dx2);
	clippedDstY2 = MIN(m_surfaceClip.y2, dy2);

	if (clippedDstX2 < clippedDstX1 || clippedDstY2 < clippedDstY1)
		return; /* Do nothing! */
    
	/* Check if Flip be needed */
	flipX = (srcWidth < 0 && dstWidth > 0) || 
		(srcWidth > 0 && dstWidth < 0);
	flipY = (srcHeight < 0 && dstHeight > 0) ||
		(srcHeight > 0 && dstHeight < 0);
    if (flipX || flipY)
    {
        NotImplemented();
        return;
    }

    /* Go DirectFB API */
    DFBRectangle srcRect, dstRect;

    if (clippedDstX1 != dx1)
    {
        sx1 += (int)((clippedDstX1 - dx1) * (srcWidth / (float)dstWidth));
        dx1 = clippedDstX1;
    }
    if (clippedDstX2 != dx2)
    {
        sx2 += (int)((clippedDstX2 - dx2) * (srcWidth / (float)dstWidth));
        dx2 = clippedDstX2;
    }
    if (clippedDstY1 != dy1)
    {
        sy1 += (int)((clippedDstY1 - dy1) * (srcHeight / (float)dstHeight));
        dy1 = clippedDstY1;
    }
    if (clippedDstY2 != dy2)
    {
        sy2 += (int)((clippedDstY2 - dy2) * (srcHeight / (float)dstHeight));
        dy2 = clippedDstY2;
    }

    srcRect.x = sx1;
    srcRect.y = sy1;
    srcRect.w = sx2 - sx1 + 1;
    srcRect.h = sy2 - sy1 + 1;

    dstRect.x = dx1;
    dstRect.y = dy1;
    dstRect.w = dx2 - dx1 + 1;
    dstRect.h = dy2 - dy1 + 1;

    OWB_DFBCHECK(m_surface->StretchBlit(m_surface, source, &srcRect, &dstRect));
}

IDirectFB* DFBGlobals::getDirectFB()
{
    static IDirectFB* instance = NULL;
    if (instance == NULL)
    {
        DirectFBCreate(&instance);
    }
    return instance;
}

} // namespace WKAL

