#ifndef _DIRECTFB_GRAPHICS_ENGINE_H_
#define _DIRECTFB_GRAPHICS_ENGINE_H_

#include "BALBase.h"
#include <stdlib.h>
#include <stdio.h>

namespace WKAL {

    class DFBGraphicsEngine : public WKALBase
    {
        public:
            DFBGraphicsEngine(PlatformGraphicsContext* cr);
            
            virtual ~DFBGraphicsEngine();

			void drawLine(int x1, int y1, int x2, int y2);
			
			void drawPolygon(DFBPoint* points, int n);
			
			void fillPolygon(DFBPoint* points, int n);
			
			void drawArc(int x, int y, int w, int h, int startAngle, int endAngle, bool fill);
			
			void drawEllipse(int x, int y, int w, int h, bool fill);
			
			void drawString(int font, int x, int y, const char *);
			
			void drawRect(int x, int y, int w, int h);
			
			void fillRect(int x, int y, int w, int h);
			
			DFBColor readPixel(int x, int y);
			
			void drawPixel(int x, int y);
			
			DFBColor unpackPixel(DFBSurfacePixelFormat format, unsigned int pixel);
			unsigned int packPixel(DFBSurfacePixelFormat format, DFBColor color);
			
            void bitBlit(int dx, int dy, int width, int height, IDirectFBSurface *source, int sx, int sy);

            void stretchBlit(int dx1, int dy1, int dx2, int dy2, IDirectFBSurface *source, int sx1, int sy1, int sx2, int sy2);

			void setClipRect(int x, int y, int w, int h);
			
			void removeClip();			

            void save();

            void restore();

        private:
        
            // poly
        	bool isLineVisible(int x1, int y1, int x2, int y2);

            bool isPolyVisible(DFBPoint* points, int count);

            bool clipPoly(DFBPoint* pointtable, int count, DFBPoint* newPointtable, int* newcount);
        	
            // arc
            typedef struct
            {
                int x0, y0;
                int rx, ry;
                int ax, ay;
                int bx, by;
                int adir, bdir;
                bool fill;
            }
            SLICE;

            void draw_line(SLICE * slice, int x0, int y, int x1);

            void GdArc(int x0, int y0, int rx, int ry, int ax, int ay, int bx, int by, bool fill);

			bool surfaceLock();
			
			void surfaceUnlock();

        public: //FIXME GraphicsContext need to access m_surface
            IDirectFBSurface *m_surface;

        private:
            /*
             * current settings on this surface.
             */
            unsigned char *m_surfaceDataBuffer;
            int  m_surfacePitch;
            int m_surfaceWidth;
            int m_surfaceHeight;
            DFBSurfacePixelFormat m_surfaceFormat;
            DFBColor m_surfaceColor;
            unsigned char extraAlpha;
            DFBSurfacePorterDuffRule m_surfacePorterDuffRule;
            DFBSurfaceDrawingFlags m_surfaceDrawingFlags;
            DFBSurfaceBlittingFlags m_surfaceBlittingFlags;
            DFBRegion m_surfaceClip;
            bool m_clipped;

            struct DFBGraphicsEngineState
            {
                DFBGraphicsEngineState(DFBColor& _color, unsigned char _extraAlpha, DFBSurfacePorterDuffRule _duffRule, DFBSurfaceDrawingFlags _drawingFlags, DFBSurfaceBlittingFlags _blittingFlags, DFBRegion& _clipRegion)
                    : color(_color)
                    , extraAlpha(_extraAlpha)
                    , duffRule(_duffRule)
                    , drawingFlags(_drawingFlags)
                    , blittingFlags(_blittingFlags)
                    , clipRegion(_clipRegion)
                {}

                DFBColor color;
                unsigned char extraAlpha;
                DFBSurfacePorterDuffRule duffRule;
                DFBSurfaceDrawingFlags drawingFlags;
                DFBSurfaceBlittingFlags blittingFlags;
                DFBRegion clipRegion;
            };

            Vector<DFBGraphicsEngineState> m_stack;

       public:
            inline int getWidth(){return m_surfaceWidth;}
            inline int getHeight(){return m_surfaceHeight;}
            inline DFBColor getColor(){ return m_surfaceColor;}
            inline void setColor(uint8_t a, uint8_t r, uint8_t g, uint8_t b)
            {
                if ( m_surfaceColor.a != a 
                  || m_surfaceColor.r != r
                  || m_surfaceColor.g != g
                  || m_surfaceColor.b != b
                  )
                {
                    //D_INFO("{%d, %d, %d, %d} \n", a, r, g, b);
                    m_surfaceColor.a = a;
                    m_surfaceColor.r = r;
                    m_surfaceColor.g = g;
                    m_surfaceColor.b = b;
                    OWB_DFBCHECK(m_surface->SetColor(m_surface, r, g, b, a));
                }
            }
            inline void setColor(DFBColor& color)
            {
                if (color.a != m_surfaceColor.a
                        || color.r != m_surfaceColor.r
                        || color.g != m_surfaceColor.g
                        || color.b != m_surfaceColor.b
                   )
                {
                    //D_INFO("{%d, %d, %d, %d} \n", color.a, color.r, color.g, color.b);
                    m_surfaceColor = color;
                    OWB_DFBCHECK(m_surface->SetColor(m_surface, color.r, color.g, color.b, color.a));
                }
            }
            inline DFBSurfacePorterDuffRule getPorterDuffRule(){return m_surfacePorterDuffRule;}
            inline void setPorterDuffRule(DFBSurfacePorterDuffRule rule)
            {
                if (rule != m_surfacePorterDuffRule)
                {
                    //D_INFO(" rule = %d \n", rule);
                    m_surfacePorterDuffRule = rule;
                    OWB_DFBCHECK(m_surface->SetPorterDuff(m_surface, rule));
                }
            }
            inline DFBSurfaceDrawingFlags getDrawingFlags(){return m_surfaceDrawingFlags;}
            inline void setDrawingFlags(DFBSurfaceDrawingFlags flags)
            {
                if (m_surfaceDrawingFlags != flags)
                {
                    //D_INFO("drawingFlags = %d \n", flags);
                    m_surfaceDrawingFlags = flags;
                    OWB_DFBCHECK(m_surface->SetDrawingFlags(m_surface, flags));
                }
            }
            inline DFBSurfaceBlittingFlags getBlittingFlags(){return m_surfaceBlittingFlags;}
            inline void setBlittingFlags(DFBSurfaceBlittingFlags flags)
            {
                if (m_surfaceBlittingFlags != flags)
                {
                    //D_INFO("blittingFlags = %d \n", flags);
                    m_surfaceBlittingFlags = flags;
                    OWB_DFBCHECK(m_surface->SetBlittingFlags(m_surface, flags));
                }
            }
    };

    class DFBGlobals
    {
    public:
        static IDirectFB* getDirectFB();
    };

} //namespace WKAL

#endif //_DIRECTFB_GRAPHICS_ENGINE_H_

