/*
 * Copyright (C) 2008, 2009 Julien Chaffraix <julien.chaffraix@gmail.com>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE COMPUTER, INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "CookieManager.h"

#include "BALBase.h"
#include "Cookie.h"
#include "CookieParser.h"
#include "CString.h"
#include "CurrentTime.h"
#include "FileSystem.h"
#include "FileIO.h"
#include "Logging.h"

namespace WebCore {

CookieManager& cookieManager()
{
    static CookieManager cookieManager;
    return cookieManager;
} 

CookieManager::CookieManager()
    : m_count(0)
{
#ifdef VA_PLATFORM_BDH    
    m_cookieJarFileName = pathByAppendingComponent("/root/ada", "cookieCollection.db");
#else
    m_cookieJarFileName = pathByAppendingComponent(OWB_DATA, "cookieCollection.db");
#endif
    loadAllCookies();
}

CookieManager::~CookieManager()
{
    //saveAllCookies();
    removeAllCookies();
}

void CookieManager::setCookies(const KURL& url, const String& value)
{
    CookieParser parser(url);
    Vector<Cookie*> cookies = parser.parse(value);

    for (size_t i = 0; i < cookies.size(); ++i) {
        Cookie* cookie = cookies[i];
        if (!shouldRejectForSecurityReason(cookie, url))
            checkAndTreatCookie(cookie);
        else
            delete cookie;
    }
}

bool CookieManager::shouldRejectForSecurityReason(const Cookie* cookie, const KURL& url)
{
    // Check if path attribute is a prefix of the request URI.
    if (!url.path().startsWith(cookie->path())) {
        LOG_ERROR("Cookie %s is rejected because its path does not math the URL %s\n", cookie->toString().utf8().data(), url.string().utf8().data());
        return true;
    }

    // Check if the domain starts with a dot.
    // FIXME: This check is in fact disabled because the CookieParser prefix the domain if needed and should be removed.
#if 0
    if (cookie->domain()[0] != '.') {
        LOG_ERROR("Cookie %s is rejected because its domain does not start with a dot.\n", cookie->toString().utf8().data());
        return true;
    }
#endif

    // Check if the domain contains an embedded dot.
    const String& cookieDomain = cookie->domain();
    if (!cookieDomain.isEmpty() && cookieDomain[0] == '.') {
        int dotPosition = cookieDomain.find(".", 1);
        if (dotPosition == -1 || static_cast<unsigned int>(dotPosition) == cookieDomain.length()) {
            LOG_ERROR("Cookie %s is rejected because its domain does not contain an embedded dot.\n", cookie->toString().utf8().data());
            return true;
        }
    }

    // The request host should domain match the Domain attribute.
    bool localFile = url.isLocalFile();
    String urlDomain = localFile ? Cookie::FileCookieDomain : url.host();
    if (!urlDomain.endsWith(cookieDomain) && (("." + urlDomain) != cookieDomain)) {
        LOG_ERROR("Cookie %s is rejected because its domain does not domain match the URL %s\n", cookie->toString().utf8().data(), url.string().utf8().data());
        return true;
    }
    // We should check for an embedded dot in the portion of string in the host not in the domain
    // but to match firefox behaviour we do not.

    return false;
}

String CookieManager::getCookie(const KURL& url, HttpOnlyCookieFiltering filter)
{
    bool isConnectionSecure = false;
    static String httpsPrefix("https:");
    if (url.string().startsWith(httpsPrefix, false))
        isConnectionSecure = true;

    // The max size is the number of cookie per host multiplied by the maximum length of a cookie. We add 1 for the final '\0'.
    static const size_t cookiesMaxLength = s_maxCookieLength * s_maxCookieCountPerHost + 1;
    Vector<UChar> cookiePairs;
    cookiePairs.reserveInitialCapacity(cookiesMaxLength);
    String cookieDomain = url.isLocalFile() ? Cookie::FileCookieDomain : url.host();
    HashMap<String, Cookie*> allResultCookies;
    for (HashMap<String, CookieMap*>::iterator it = m_managerMap.begin(); it != m_managerMap.end(); ++it) {

        // Handle sub-domain by only looking at the end of the host.
        if (/*it->first.endsWith(cookieDomain)*/equalIgnoringCase(it->first, cookieDomain)
                || (it->first.startsWith(".", false) && ("." + cookieDomain).endsWith(it->first, false))) {
            // Get CookieMap to check for expired cookies.
            Vector<Cookie*> cookies = it->second->getCookies();
            for (size_t i = 0; i < cookies.size(); ++i) {
                Cookie* cookie = cookies[i];
                // Get the cookies filtering out the secure cookies on an unsecure connection and HttpOnly cookies if requested.
                if (url.path().startsWith(cookie->path(), false) && (isConnectionSecure || !cookie->isSecure()) && (filter == WithHttpOnlyCookies || !cookie->isHttpOnly()) && !cookie->hasExpired()) {
                    if (allResultCookies.contains(cookie->name())) {
                        Cookie* oldCookie = allResultCookies.get(cookie->name());
                        if (oldCookie->domain() == cookie->domain())
                            if (cookie->path().startsWith(oldCookie->path())) //check longest path
                                allResultCookies.set(cookie->name(), cookie);
                        else if (cookie->domain().endsWith(oldCookie->domain())) //check longest domain
                                allResultCookies.set(cookie->name(), cookie);                       
                    } else
                        allResultCookies.add(cookie->name(), cookie);
                }
            }
        }
    }

    bool first = true;
    for(HashMap<String, Cookie*>::iterator it = allResultCookies.begin(); it != allResultCookies.end(); ++it) {
        Cookie* cookie = it->second;
        static const String space(" ");
        static const String equal("=");
        static const String semiColon(";");
        if (!first) {
            append(cookiePairs, semiColon);
            append(cookiePairs, space);
        } else
            first = false;
        append(cookiePairs, cookie->name());
        append(cookiePairs, equal);
        append(cookiePairs, cookie->value());
    }

    // Per construction of our cookies, we should not grow our vector.
    ASSERT(cookiePairs.capacity() == cookiesMaxLength);

    // Append the final '\0'.
    static const String nullTerminator("\0");
    append(cookiePairs, nullTerminator);
    return String::adopt(cookiePairs);
}

void CookieManager::removeAllCookies()
{
    HashMap<String, CookieMap*>::iterator first = m_managerMap.begin();
    HashMap<String, CookieMap*>::iterator end = m_managerMap.end();
    for (HashMap<String, CookieMap*>::iterator it = first; it != end; ++it)
        delete it->second;

    m_count = 0;
}

void CookieManager::setCookieJar(const char* fileName)
{
    m_cookieJarFileName = String(fileName);
    loadAllCookies();
}

void CookieManager::checkAndTreatCookie(Cookie* cookie)
{
    ASSERT(cookie->domain().length());

    CookieMap* curMap = m_managerMap.get(cookie->domain());

    // Check for cookie to remove in case it is not a session cookie and it has expired.
    if (cookie->hasExpired()) {
        // The cookie has expired so check we have a valid HashMap so try delete it.
        if (curMap) {
            // Check if we have a cookie to remove and update information accordingly.
            Cookie* prevCookie = curMap->takePrevious(cookie);
            if (prevCookie) {
                removedCookie();
                delete prevCookie;
            }
        }
        
        delete cookie;
    } else {
        if (!curMap) {
            curMap = new CookieMap();
            m_managerMap.add(cookie->domain(), curMap);
            addCookieToMap(curMap, cookie);
        } else {
            // Check if there is a previous cookie.
            Cookie* prevCookie = curMap->takePrevious(cookie);

            if (prevCookie) {
                update(curMap, prevCookie, cookie);
                delete prevCookie;
            } else
                addCookieToMap(curMap, cookie);
        }
    }
}

void CookieManager::addCookieToMap(CookieMap* map, Cookie* cookie)
{
    // Check if we do not have reached the cookie's threshold.
    // FIXME : should split the case and remove one cookie among all the other if m_count >= max_count
    if (map->count() > s_maxCookieCountPerHost || m_count >= s_globalMaxCookieCount) {
        Cookie* rmCookie = map->removeOldestCookie();
        removedCookie();
        delete rmCookie;
    }

    map->add(cookie);

    m_count++;
}

void CookieManager::update(CookieMap* map, Cookie* prevCookie, Cookie* newCookie)
{
    ASSERT(!map->takePrevious(prevCookie));
    map->add(newCookie);
}

void CookieManager::saveAllCookies()
{
    File file(m_cookieJarFileName);

    if (file.open('w') == -1)
        return;

    HashMap<String, CookieMap*>::iterator first = m_managerMap.begin();
    HashMap<String, CookieMap*>::iterator end = m_managerMap.end();
    for (HashMap<String, CookieMap*>::iterator it = first; it != end; ++it) {
        CookieMap* cookieMap = it->second;
        Vector<Cookie*> cookies = cookieMap->getCookies();

        for(size_t i = 0; i < cookies.size(); i++) {
            Cookie* cookie = cookies[i];

            if (!cookie->hasExpired() && !cookie->isSession()) {
                file.write(cookie->toString());
                file.write("\n");
            }
        }
    }
}

void CookieManager::loadAllCookies()
{
    File file(m_cookieJarFileName);

    if (file.open('r') == -1)
        return;
    
    int size = file.getSize();
    char* data = file.read(size);
    String value(data, size);
    CookieParser parser;
    Vector<Cookie*> cookies = parser.parse(value);

    for (size_t i = 0; i < cookies.size(); ++i) {
        Cookie* newCookie = cookies[i];
        checkAndTreatCookie(newCookie);
    }
    delete data;
}

} // namespace WebCore
