#include "config.h"
#include "FramelessScrollView.h"
#include "FrameView.h"
#include "IntRect.h"
#include "GraphicsContext.h"
#include "PlatformKeyboardEvent.h"
#include "PlatformMouseEvent.h"
#include "PlatformWheelEvent.h"

#include "FL_HostWindowImpl.h"
#include <FL/Fl.H>
#include <FL/x.H>
#include <stdio.h>

namespace WebCore {

FL_HostWindowImpl::FL_HostWindowImpl(IntRect& widgetRect, FramelessScrollView* widget)
    : Fl_Window(widgetRect.x(), widgetRect.y(), widgetRect.width(), widgetRect.height(), "FL_HostWindowImpl")
    , m_widget(widget)
    , m_cairoSurface(0)
    , m_context(0)
    , m_surface(0)
{
    clear_border();
    modal();
    
    if (m_widget)
        m_widget->setClient(this);
    //printf("{%d, %d, %d, %d}\n", widgetRect.x(), widgetRect.y(), widgetRect.width(), widgetRect.height());    
}

FL_HostWindowImpl::~FL_HostWindowImpl()
{
    if (m_widget)
        m_widget->setClient(0);
        
    if (m_context)
        cairo_destroy(m_context);
    if (m_cairoSurface)
        cairo_surface_destroy(m_cairoSurface);        
}

// Fl_Window methods
void FL_HostWindowImpl::draw()
{
    IDirectFBSurface* surface = reinterpret_cast<IDirectFBSurface*>(getClientSurface());
    if (surface != m_surface) {
        if (m_context)
            cairo_destroy(m_context);
        if (m_cairoSurface)
            cairo_surface_destroy(m_cairoSurface);

        m_surface = surface;                        
        m_context = 0;
        m_cairoSurface = 0;
        if (m_surface) {
            m_cairoSurface = cairo_directfb_surface_create(fl_dfb, surface);
            if (!m_cairoSurface)
                fprintf(stderr, "failed to create cairo surface\n");
            m_context = cairo_create(m_cairoSurface);
            if (!m_context)
                fprintf(stderr, "failed to create cairo context\n");
        }
    }
    if (m_context && m_widget) {
        GraphicsContext gc(m_context);
        IntRect r(0, 0, w(), h());
        
        m_widget->paint(&gc, r);//TODO just repaint dirty rect.
    }
}

int FL_HostWindowImpl::handle(int event)
{
    if (!m_widget || !fl_dfb_event)
        return Fl_Window::handle(event);

    //printf("%s : event = %d\n", __FUNCTION__, event);

    DFBEvent dfbevent;    
    
    switch(event) {
    	case FL_FOCUS:
        case FL_UNFOCUS:
            return 1;
        case FL_PUSH:
            if (fl_FLEventToDFBEvent(event, dfbevent))
                m_widget->handleMouseDownEvent(PlatformMouseEvent(&dfbevent));
            return 1;
        case FL_RELEASE:
            if (fl_FLEventToDFBEvent(event, dfbevent))
                m_widget->handleMouseReleaseEvent(PlatformMouseEvent(&dfbevent));
            return 1;
        case FL_MOVE:
        case FL_DRAG:
            if (fl_FLEventToDFBEvent(event, dfbevent))
                m_widget->handleMouseMoveEvent(PlatformMouseEvent(&dfbevent));
            return 1;

        case FL_SHORTCUT:
        case FL_KEYDOWN:
        case FL_KEYUP:
            if (fl_FLEventToDFBEvent(event, dfbevent))
                m_widget->handleKeyEvent(PlatformKeyboardEvent(&dfbevent));
            return 1;

        default:
            return Fl_Window::handle(event);
    }    
}

void FL_HostWindowImpl::show()
{
    Fl_Window::show();
    Fl::focus(this);   
}

void FL_HostWindowImpl::hide()
{
    //if (Fl::grab() == this)
    //    Fl::grab(0);
    Fl_Window::hide();
}
    
//-----------------------------------------------------------------------------
// WebCore::HostWindow

void FL_HostWindowImpl::repaint(const IntRect& paintRect,
                               bool contentChanged,
                               bool immediate,
                               bool repaintContentOnly)
{
    // Ignore spurious calls.
    if (!contentChanged || paintRect.isEmpty())
        return;
    damage(FL_DAMAGE_CHILD, paintRect.x(), paintRect.y(), paintRect.width(), paintRect.height());        
}

void FL_HostWindowImpl::scroll(const IntSize& scrollDelta,
                              const IntRect& scrollRect,
                              const IntRect& clipRect)
{
    redraw();//TODO
}

IntPoint FL_HostWindowImpl::screenToWindow(const IntPoint& point) const
{
    notImplemented();
    return IntPoint();
}

IntRect FL_HostWindowImpl::windowToScreen(const IntRect& rect) const
{
    notImplemented();
    return IntRect();
}

void FL_HostWindowImpl::scrollRectIntoView(const IntRect&, const ScrollView*) const
{
    // Nothing to be done here since we do not have the concept of a container
    // that implements its own scrolling.
}

void FL_HostWindowImpl::scrollbarsModeDidChange() const
{
    // Nothing to be done since we have no concept of different scrollbar modes.
}

}
