/*
 * Copyright (C) 2009 University of Szeged
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY UNIVERSITY OF SZEGED ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MIPSAssembler_h
#define MIPSAssembler_h

#include <wtf/Platform.h>

#if ENABLE(ASSEMBLER) && PLATFORM(MIPS)

#include "AssemblerBuffer.h"
#include <wtf/Assertions.h>
#include <wtf/Vector.h>
#include <stdint.h>

namespace JSC {

namespace MIPS {
    typedef enum {
        r0 = 0, r1, r2, r3, r4, r5, r6, r7,
        r8, r9, r10, r11, r12, r13, r14, r15,
        r16, r17, r18, r19, r20, r21, r22, r23,
        r24, r25, r26, r27, r28, r29, r30, r31,

        zero = r0,
        
        at = r1,
        
        v0 = r2,
        v1 = r3,
        
        a0 = r4,
        a1 = r5,
        a2 = r6,
        a3 = r7,
        
        t0 = r8,
        t1 = r9,
        t2 = r10,
        t3 = r11,
        t4 = r12,
        t5 = r13,
        t6 = r14,
        t7 = r15,
        t8 = r24,
        t9 = r25,

        s0 = r16,
        s1 = r17,
        s2 = r18,
        s3 = r19,
        s4 = r20,
        s5 = r21,
        s6 = r22,
        s7 = r23,
        
        k0 = r26,
        k1 = r27,
        
        gp = r28,
        
        sp = r29,
        
        fp = r30,
        
        ra = r31
    } RegisterID;

    typedef enum {
        f0, f1, f2, f3, f4, f5, f6, f7,
        f8, f9, f10, f11, f12, f13, f14, f15,
        f16, f17, f18, f19, f20, f21, f22, f23,
        f24, f25, f26, f27, f28, f29, f30, f31,

        fp0 //FIXME
    } FPRegisterID;

} // namespace MIPS



class MIPSAssembler {
    public:
        typedef MIPS::RegisterID RegisterID;
        typedef MIPS::FPRegisterID FPRegisterID;

        //typedef AssemblerBufferWithConstantPool<2048, 4, 4, MIPSAssembler> MIPSBuffer;
        typedef WTF::Vector<int, 64> Jumps;

    /* The maximum reach for a pc relative branch (18 bits signed offset)
     * NOTE: The 18-bit signed offset is added to the address of the
     * instruction following the branch (not the branch itself), in the
     * branch delay slot, to form a PC-relative effective target address.
     */
    static const int MIPS_MAX_BRANCH_OFFSET = 128*1024;
    static const int MIPS_MIN_BRANCH_OFFSET = (-128)*1024+4;
    static const int MIPS_OUT_OF_OFFSET_FLAG = (-1);

    static const int MIPS_MAX_IMMEDIATE = 128*1024;
    static const int MIPS_MIN_IMMEDIATE = (-128)*1024;
    /**************************************************************
     * Private #defines for opcode encodings.
     **************************************************************/

    static const int MIPS_RS_SHIFT   = 21;     /* source register */
    static const int MIPS_I_RD_SHIFT = 16;     /* I-Type dest register */
    static const int MIPS_R_RD_SHIFT = 11;     /* R-Type dest register */
    static const int MIPS_R_RT_SHIFT = 16;     /* R-Type rhs register */
    static const int MIPS_SA_SHIFT   = 6;      /* shift amount */

    typedef enum {
        MIPS_NOP_OPCODE   = (0x00),         /* nop */
        MIPS_BKPT_OPCODE  = (0x0d),         /* break */

        MIPS_ADD_OPCODE   = (0x20),           /* reg32 = reg32 + reg32. */
        MIPS_ADDU_OPCODE  = (0x21),         /* reg32 = reg32 + reg32. No overflow. */
        MIPS_ADDI_OPCODE  = (0x08 << 26),   /* reg32 = reg32 + imm32. */
        MIPS_ADDIU_OPCODE = (0x09 << 26),   /* reg32 = reg32 + imm32. No overflow.*/
        MIPS_SUB_OPCODE   = (0x22),         /* reg32 = reg32 - reg32. */
        MIPS_SUBU_OPCODE  = (0x23),         /* reg32 = reg32 - reg32. Unsigned. */
        MIPS_MUL_OPCODE   = (0x70000002),   /* reg32 = reg32 * reg32. */
        MIPS_MULT_OPCODE  = (0x18),         /* HI, LO = reg32 * reg32. */
        MIPS_MULTU_OPCODE = (0x19),         /* HI, LO = reg32U * reg32U. */
        MIPS_MADD_OPCODE  = (0x70000000),   /* HI,LO = (HI,LO)+(reg32*reg32). */
        MIPS_MADDU_OPCODE = (0x70000001),   /* HI,LO = (HI,LO)+(reg32*reg32), unsigned.*/
        MIPS_DIV_OPCODE   = (0x1a),         /* Div32 */
        
        MIPS_AND_OPCODE	  = (0x24),         /* reg32 = reg32 AND reg32. */
        MIPS_ANDI_OPCODE  = (0x0c << 26),   /* reg32 = reg32 AND imm32.*/
        MIPS_OR_OPCODE	  = (0x25),         /* reg32 = reg32 OR reg32. */
        MIPS_ORI_OPCODE	  = (0x0d << 26),   /* reg32 = reg32 OR imm32.*/
        MIPS_NOR_OPCODE   = (0x27),         /* reg32 = reg32 NOR reg32. */
        MIPS_XOR_OPCODE	  = (0x26),         /* reg32 = reg32 XOR reg32. */
        MIPS_XORI_OPCODE  = (0x0e << 26),   /* reg32 = reg32 XOR imm32.*/
        
        MIPS_SLL_OPCODE	  = (0x00),         /* Shift Left Logical */
        MIPS_SLLV_OPCODE  = (0x04),         /* Shift Left Logical. Shift amount in reg.*/
        MIPS_SRL_OPCODE	  = (0x02),         /* Shift Right Logical */
        MIPS_SRLV_OPCODE  = (0x06),         /* Shift Right Logical. Shift amount in reg*/
        MIPS_SRA_OPCODE	  = (0x03),         /* Shift Right Arithmetic */
        MIPS_SRAV_OPCODE  = (0x07),         /* Shift Right Arithmetic. Shift amount in reg*/

        MIPS_MOVN_OPCODE  = (0x0b),         /* Move Conditional on Not Zero. */
        MIPS_MOVZ_OPCODE  = (0x0a),         /* Move Conditional on Zero. */
        MIPS_MFHI_OPCODE  = (0x10),         /* Move From HI Register. */
        MIPS_MFLO_OPCODE  = (0x12),         /* Move Form LO Register. */
        MIPS_MTHI_OPCODE  = (0x11),         /* Move To HI Register. */
        MIPS_MTLO_OPCODE  = (0x13),         /* Move To LO Register. */

        MIPS_SLT_OPCODE   = (0x2a),         /* Set on Less Than. */
        MIPS_SLTU_OPCODE  = (0x2b),         /* Set on Less Than Unsigend. */
        MIPS_SLTI_OPCODE  = (0x0a << 26),   /* Set on Less Than Immediate. */
        MIPS_SLTIU_OPCODE = (0x0b << 26),   /* Set on Less Than Immediate Unsigned. */

        MIPS_BEQ_OPCODE   = (0x04 << 26),   /* Branch on Equal. */
        MIPS_BNE_OPCODE   = (0x05 << 26),   /* Branch on Not Equal. */
        MIPS_BLTZ_OPCODE  = (0x01 << 26),   /* Branch on Less Than 0. */
        MIPS_BGEZ_OPCODE  = ((0x01 << 26) | (0x01 << 16)),   /* Branch on Greater Than or Equal to 0. */
        //MIPS_BGEZAL_OPCODE=  (0x11 << 16),  /*Branch on Greater Than or Equal to 0 and Link*/
        MIPS_BGTZ_OPCODE  = (0x07 << 26),   /* Branch on Greater Than 0. */
        MIPS_BLEZ_OPCODE  = (0x06 << 26),   /* Branch on Less Than or Equal to 0. */
        //MIPS_BLTZ_OPCODE  = (0x00 << 16),   /* Branch on Less Than 0. */
        MIPS_BLTZAL_OPCODE= ((0x01 << 26) | (0x10 << 16)),  /*Branch on Less Than 0 and Link. */
        MIPS_BGEZAL_OPCODE= ((0x01 << 26) | (0x11 << 16)), /*Branch on Greater Than or Equal to 0 and link*/
        MIPS_J_OPCODE     = (0x02 << 26),   /* Jump. */
        MIPS_JAL_OPCODE   = (0x03 << 26),   /* Jump and Link. */
        MIPS_JALR_OPCODE  = (0x09),         /* Jump and Link Register. */
        MIPS_JR_OPCODE    = (0x08),         /* Jump Register. */

        MIPS_LD_OPCODE    = (0x37 << 26),   /* Load 64 bit value. */
        MIPS_SD_OPCODE    = (0x3f << 26),   /* Store 64 bit value. */
        MIPS_LW_OPCODE    = (0x23 << 26),   /* Load 32 bit value (signed). */
        MIPS_SW_OPCODE    = (0x2b << 26),   /* Store 32 bit value (signed). */
        MIPS_LHU_OPCODE   = (0x25 << 26),   /* Load unsigned 16 bit value. */
        MIPS_LH_OPCODE    = (0x21 << 26),   /* Load signed 16 bit value. */
        MIPS_SH_OPCODE    = (0x29 << 26),   /* Store 16 bit value. */
        MIPS_LB_OPCODE    = (0x20 << 26),   /* Load signed 8 bit value. */
        MIPS_SB_OPCODE    = (0x28 << 26),   /* Store 8 bit value. */

        MIPS_LUI_OPCODE   = (0x0f << 26),   /* Load Upper Immediate. */
        
        MIPS_BC1F_OPCODE  = (0x11<<26 | 0x8<<21),           /* Branch on FP False */
        MIPS_BC1T_OPCODE  = (0x11<<26 | 0x8<<21 | 0x1<<16)  /* Branch on FP True */
    } OpcodeID;

    // Float Point
        
    static const int MIPS_FN_SHIFT = 16; /* fn register */
    static const int MIPS_FS_SHIFT = 11; /* fs register */
    static const int MIPS_FD_SHIFT = 6;  /* fd register */

    typedef enum {
        MIPS_SWC1_OPCODE  = (0x39 << 26), /* Store Word from Floating Point */
        MIPS_SDC1_OPCODE  = (0x3d << 26), /* Store Doubleword from Floating Point */
        MIPS_LWC1_OPCODE  = (0x31 << 26), /* Load Word to Floating Point */
        MIPS_LDC1_OPCODE  = (0x35 << 26), /* Load Doubleword to Floating Point */
                
        MIPS_FP_COP1      = (0x11 << 26),
        MIPS_FP_S_FMT     = (0x10 << 21), /* Single */
        MIPS_FP_D_FMT     = (0x11 << 21), /* Double */
        MIPS_FP_W_FMT     = (0x14 << 21), /* Word */
        
        MIPS_FP_NEG_OPCODE  = (0x7), /* Floating Point Negate */
        MIPS_FP_ADD_OPCODE  = (0)  , /* Floating Point Add */
        MIPS_FP_SUB_OPCODE  = (0x1), /* Floating Point Sub */
        MIPS_FP_MUL_OPCODE  = (0x2), /* Floating Point Mul */
        MIPS_FP_DIV_OPCODE  = (0x3), /* Floating Point Div */
        MIPS_FP_CMP_OPCODE  = (0x3 << 4), /* Floating Point Compare */
        MIPS_FP_MOV_OPCODE  = (0x6), /* Floating Point Move */
        
        MIPS_FP_CVTD_OPCODE = (0x21), /* Floating Point Convert to Double */
        MIPS_FP_CVTS_OPCODE = (0x20), /* Floating Point Convert to Single */
        
        MIPS_FP_TRUNCW_OPCODE = (0xd), /*Floating Point Truncate to Word Fixed Point*/
        
        MIPS_MFC1_OPCODE    = (0),         /* Move 32-bit GPR from 32-bit FPR */
        MIPS_MTC1_OPCODE    = (0x4 << 21) /* Move 32-bit GPR to 32-bit FPR */
    } FPOpcodeID;

        typedef enum {
            //ConditionO,
            //ConditionNO,
            ConditionLZ,
            ConditionGZ,
            ConditionLEZ,
            ConditionGEZ,

            ConditionB,
            ConditionA,
            ConditionBE,
            ConditionAE,

            ConditionE,
            ConditionNE,

            ConditionL,
            ConditionG,
            ConditionGE,
            ConditionLE,
        } Condition;
/*
        enum {
            padForAlign8  = 0x00,
            padForAlign16 = 0x0000,
            padForAlign32 = 0xee120070,
        };
*/        

        MIPSAssembler() { }

        class JmpSrc {
            friend class MIPSAssembler;
        public:
            JmpSrc()
                : m_offset(-1)
                , m_latePatch(false)
            {
            }

            void enableLatePatch() { m_latePatch = true; }
        private:
            JmpSrc(int offset)
                : m_offset(offset)
                , m_latePatch(false)
            {
            }

            int m_offset : 31;
            int m_latePatch : 1;
        };

        class JmpDst {
            friend class MIPSAssembler;
        public:
            JmpDst()
                : m_offset(-1)
                , m_used(false)
            {
            }

            bool isUsed() const { return m_used; }
            void used() { m_used = true; }
        private:
            JmpDst(int offset)
                : m_offset(offset)
                , m_used(false)
            {
                ASSERT(m_offset == offset);
            }

            int m_offset : 31;
            int m_used : 1;
        };

        // Instruction formating

        void emit_R_Inst(OpcodeID op, RegisterID rd, RegisterID rs, RegisterID rt)
        {
            ASSERT(rd <= MIPS::r31 && rs <= MIPS::r31 && rt <= MIPS::r31);
            m_buffer.putInt(op | ((rs & 0x1F) << MIPS_RS_SHIFT) | ((rd & 0x1F) << MIPS_R_RD_SHIFT) | ((rt & 0x1F) << MIPS_R_RT_SHIFT));
        }

        void emit_I_Inst(OpcodeID op, RegisterID rd, RegisterID rs, short imm)
        {
            ASSERT(rd <= MIPS::r31 && rs <= MIPS::r31);
            m_buffer.putInt(op | ((rs & 0x1F) << MIPS_RS_SHIFT) | ((rd & 0x1F) << MIPS_I_RD_SHIFT) | (imm & 0xFFFF));
        }

        void emit_J_Inst(OpcodeID op, int target)
        {
            ASSERT((target < MIPS_MAX_BRANCH_OFFSET && target > MIPS_MIN_BRANCH_OFFSET));
            ASSERT((target % 4) == 0);
            m_buffer.putInt(op | ((target >> 2) & 0x3FFFFFF));
        }

        void nop()
        {
            m_buffer.putInt(0);
        }

        void bkpt(int code = 0)
        {
            m_buffer.putInt(MIPS_BKPT_OPCODE);
        }

        void store32(RegisterID rd, RegisterID rs, short offset = 0)
        {
            emit_I_Inst(MIPS_SW_OPCODE, rd, rs, offset);
        }
        
        void store16(RegisterID rd, RegisterID rs, short offset = 0)
        {
            emit_I_Inst(MIPS_SH_OPCODE, rd, rs, offset);
        }

        void store8(RegisterID rd, RegisterID rs, short offset = 0)
        {
            emit_I_Inst(MIPS_SB_OPCODE, rd, rs, offset);
        }

        void load32(RegisterID rd, RegisterID rs, short offset = 0)
        {
            emit_I_Inst(MIPS_LW_OPCODE, rd, rs, offset);
        }

        void load16(RegisterID rd, RegisterID rs, short offset = 0)
        {
            emit_I_Inst(MIPS_LH_OPCODE, rd, rs, offset);
        }

        void load8(RegisterID rd, RegisterID rs, short offset = 0)
        {
            emit_I_Inst(MIPS_LB_OPCODE, rd, rs, offset);
        }

        void lui(RegisterID rd, short imm)
        {
            emit_I_Inst(MIPS_LUI_OPCODE, rd, MIPS::zero, imm);
        }

        void moveImm32(RegisterID rd, int imm)
        {
            lui(rd, ((imm >> 16) & 0xFFFF));
            ori(rd, rd, (imm & 0xFFFF));
        }

        void move(RegisterID rd, RegisterID rs)
        {
            or32(rd, rs, MIPS::zero);
        }

        void and32(RegisterID rd, RegisterID rs, RegisterID rt)
        {
            emit_R_Inst(MIPS_AND_OPCODE, rd, rs, rt);
        }

        void andi(RegisterID rd, RegisterID rs, short imm)
        {
            emit_I_Inst(MIPS_ANDI_OPCODE, rd, rs, imm);
        }

        void eor(RegisterID rd, RegisterID rs, RegisterID rt)
        {
            emit_R_Inst(MIPS_XOR_OPCODE, rd, rs, rt);
        }

        void eori(RegisterID rd, RegisterID rs, short imm)
        {
            emit_I_Inst(MIPS_XORI_OPCODE, rd, rs, imm);
        }

        void or32(RegisterID rd, RegisterID rs, RegisterID rt)
        {
            emit_R_Inst(MIPS_OR_OPCODE, rd, rs, rt);
        }

        void ori(RegisterID rd, RegisterID rs, short imm)
        {
            emit_I_Inst(MIPS_ORI_OPCODE, rd, rs, imm);
        }

        void nor(RegisterID rd, RegisterID rs, RegisterID rt)
        {
            emit_R_Inst(MIPS_NOR_OPCODE, rd, rs, rt);
        }
        
        void not32(RegisterID rd, RegisterID rs)
        {
            emit_R_Inst(MIPS_NOR_OPCODE, rd, rs, MIPS::zero);
        }

        void sub(RegisterID rd, RegisterID rs, RegisterID rt)
        {
            emit_R_Inst(MIPS_SUBU_OPCODE, rd, rs, rt);
        }

        void subi(RegisterID rd, RegisterID rs, short imm)
        {
            addi(rd, rs, -imm);
        }

        void add(RegisterID rd, RegisterID rs, RegisterID rt)
        {
            emit_R_Inst(MIPS_ADDU_OPCODE, rd, rs, rt);
        }

        void addi(RegisterID rd, RegisterID rs, short imm)
        {
            emit_I_Inst(MIPS_ADDIU_OPCODE, rd, rs, imm);
        }

        void sll(RegisterID rd, RegisterID rs, short shift_amount)
        {
            m_buffer.putInt(MIPS_SLL_OPCODE | (rd << MIPS_R_RD_SHIFT) | (rs << MIPS_R_RT_SHIFT) | (shift_amount & 0x1F) << MIPS_SA_SHIFT);
        }

        void sllv(RegisterID rd, RegisterID rs, RegisterID sa)
        {
            emit_R_Inst(MIPS_SLLV_OPCODE, rd, rs, sa);
        }

        void srl(RegisterID rd, RegisterID rs, short shift_amount)
        {
            m_buffer.putInt(MIPS_SRL_OPCODE | (rd << MIPS_R_RD_SHIFT) | (rs << MIPS_R_RT_SHIFT) | (shift_amount & 0x1F) << MIPS_SA_SHIFT);
        }

        void srlv(RegisterID rd, RegisterID rs, RegisterID sa)
        {
            emit_R_Inst(MIPS_SRLV_OPCODE, rd, rs, sa);
        }
       
        void mul(RegisterID rd, RegisterID rs, RegisterID rt)
        {
            emit_R_Inst(MIPS_MUL_OPCODE, rd, rs, rt);
        }

        void mult(RegisterID rdhi, RegisterID rdlo, RegisterID rs, RegisterID rt)
        {
            emit_R_Inst(MIPS_MULT_OPCODE, MIPS::zero, rs, rt);
            emit_R_Inst(MIPS_MFHI_OPCODE, rdhi, MIPS::zero, MIPS::zero);
            emit_R_Inst(MIPS_MFLO_OPCODE, rdlo, MIPS::zero, MIPS::zero);
        }

        void push(RegisterID rs)
        {
            store32(rs, MIPS::sp);
            subi(MIPS::sp, MIPS::sp, 4);
        }

        void pop(RegisterID rd)
        {
            load32(rd, MIPS::sp);
            addi(MIPS::sp, MIPS::sp, 4);
        }

        inline void poke(RegisterID rs)
        {
            store32(rs, MIPS::sp);
        }

        inline void peek(RegisterID rd)
        {
            load32(rd, MIPS::sp);
        }

        // General helpers

        int size()
        {
            return m_buffer.size();
        }
/*
        void ensureSpace(int insnSpace, int constSpace)
        {
            m_buffer.ensureSpace(insnSpace, constSpace);
        }
*/        

        JmpDst label()
        {
            return JmpDst(m_buffer.size());
        }

        JmpDst align(int alignment)
        {
            while (!m_buffer.isAligned(alignment))
                nop();

            return label();
        }
        
        static Condition getOppositedCond(Condition cond)
        {
            switch(cond) {
                //case ConditionO:  return ConditionNO;
                //case ConditionNO: return ConditionO;
                case ConditionLZ:  return ConditionGEZ;
                case ConditionGZ: return ConditionLEZ;
                case ConditionLEZ:  return ConditionGZ;
                case ConditionGEZ: return ConditionLZ;
                case ConditionE:  return ConditionNE;
                case ConditionNE: return ConditionE;
                case ConditionA:  return ConditionBE;
                case ConditionB:  return ConditionAE;
                case ConditionG:  return ConditionLE;
                case ConditionL:  return ConditionGE;
                case ConditionBE: return ConditionA;
                case ConditionAE: return ConditionB;
                case ConditionLE: return ConditionG;
                case ConditionGE: return ConditionL;
                default:
                                  ASSERT(0);
            }
        }

        void slt(RegisterID rd, RegisterID rs, RegisterID rt)
        {
            emit_R_Inst(MIPS_SLT_OPCODE, rd, rs, rt);
        }

        void sltu(RegisterID rd, RegisterID rs, RegisterID rt)
        {
            emit_R_Inst(MIPS_SLTU_OPCODE, rd, rs, rt);
        }

        void slti(RegisterID rd, RegisterID rs, short value)
        {
            emit_I_Inst(MIPS_SLTI_OPCODE, rd, rs, value);
        }

        void sltiu(RegisterID rd, RegisterID rs, short value)
        {
            emit_I_Inst(MIPS_SLTIU_OPCODE, rd, rs, value);
        }

        void bne(RegisterID rs, RegisterID rt, short target, bool insertNop = true)
        {
            emit_I_Inst(MIPS_BNE_OPCODE, rs, rt, target);
            if (insertNop)
                nop();
        }

        void beq(RegisterID rs, RegisterID rt, short target, bool insertNop = true)
        {
            emit_I_Inst(MIPS_BEQ_OPCODE, rs, rt, target);
            if (insertNop)
                nop();
        }

        void branch(Condition cond, RegisterID rs, RegisterID rt, short target, bool insertNop = true)
        {
            RegisterID tmp;

            switch(cond)
            {
                //case ConditionO:
                //case ConditionNO:
                //    ASSERT(0);

                case ConditionE:
                case ConditionNE:
                    nop();
                    emit_I_Inst(cond == ConditionE ? MIPS_BEQ_OPCODE : MIPS_BNE_OPCODE, rs, rt, target);
                    break;

                case ConditionLZ:
                    nop();
                    emit_I_Inst(MIPS_BLTZ_OPCODE, MIPS::zero, rs, target);
                    break;
                case ConditionGZ:
                    nop();
                    emit_I_Inst(MIPS_BGTZ_OPCODE, MIPS::zero, rs, target);
                    break;
                case ConditionLEZ:
                    nop();
                    emit_I_Inst(MIPS_BLEZ_OPCODE, MIPS::zero, rs, target);
                    break;
                case ConditionGEZ:
                    nop();
                    emit_I_Inst(MIPS_BGEZ_OPCODE, MIPS::zero, rs, target);
                    break;
                case ConditionA:
                case ConditionG:
                    tmp = rs;
                    rs = rt;
                    rt = tmp;
                case ConditionB:
                case ConditionL:
                    if (cond == ConditionA || cond == ConditionB) {
                        emit_R_Inst(MIPS_SLTU_OPCODE, MIPS::at, rs, rt);
                    } else {
                        emit_R_Inst(MIPS_SLT_OPCODE, MIPS::at, rs, rt);
                    }
                    emit_I_Inst(MIPS_BNE_OPCODE, MIPS::at, MIPS::zero, target);
                    break;
                case ConditionAE:
                case ConditionGE:
                    tmp = rs;
                    rs = rt;
                    rt = tmp;
                case ConditionBE:
                case ConditionLE:
                    if (cond == ConditionAE || cond == ConditionBE) {
                        emit_R_Inst(MIPS_SLTU_OPCODE, MIPS::at, rt, rs);
                    } else {
                        emit_R_Inst(MIPS_SLT_OPCODE, MIPS::at, rt, rs);
                    }
                    emit_I_Inst(MIPS_BEQ_OPCODE, MIPS::at, MIPS::zero, target);
                    break;
                default:
                    ASSERT(0);
            }
            if (insertNop)
                nop();
        }

        JmpSrc call(bool insertNop = true)
        {
            int s = size() + 8;
            nop();
            nop();
            emit_R_Inst(MIPS_JALR_OPCODE, MIPS::zero, MIPS::at, MIPS::zero);
            if (insertNop)
                nop();
            return JmpSrc(s);
        }

        JmpSrc call(RegisterID target, bool insertNop = true)
        {
            int s = size();
            emit_R_Inst(MIPS_JALR_OPCODE, MIPS::ra, target, MIPS::zero);
            if (insertNop)
                nop();
            return JmpSrc(s);
        }

        JmpSrc call(void* target, bool insertNop = true)
        {
            int s = size();
            emit_J_Inst(MIPS_JAL_OPCODE, reinterpret_cast<uint32_t>(target));
            if (insertNop)
                nop();
            return JmpSrc(s);
        }

        JmpSrc jmp(bool insertNop = true)
        {
            int s = size() + 8;
            nop();
            nop();
            emit_R_Inst(MIPS_JR_OPCODE, MIPS::zero, MIPS::at, MIPS::zero);
            if (insertNop)
                nop();
            return JmpSrc(s);
        }

        JmpSrc jmp(RegisterID target, bool insertNop = true)
        {
            int s = size();
            emit_R_Inst(MIPS_JR_OPCODE, MIPS::zero, target, MIPS::zero);
            if (insertNop)
                nop();
            return JmpSrc(s);
        }

        JmpSrc jmp(void* target, bool insertNop = true)
        {
            int s = size();
            emit_J_Inst(MIPS_J_OPCODE, reinterpret_cast<uint32_t>(target));
            if (insertNop)
                nop();
            return JmpSrc(s);
        }

        void ret(bool insertNop = true)
        {
            emit_R_Inst(MIPS_JR_OPCODE, MIPS::zero, MIPS::ra, MIPS::zero);
            if (insertNop)
                nop();
        }

        void* executableCopy(ExecutablePool* allocator)
        {            
            char* data = reinterpret_cast<char*>(m_buffer.executableCopy(allocator));
            for(Jumps::iterator it = m_jumps.begin(); it != m_jumps.end(); ++it) {
                uint32_t* inst = reinterpret_cast<uint32_t*>(data + *it);
                ASSERT(needLatePatch(inst));
                repatchLatePatch(inst);
            }
            return data;
        }

        // Patching helpers

        static void markNeedLatePatch(void* code, int offset)
        {
            int32_t* loc = reinterpret_cast<int32_t*>(code);

            ASSERT(offset < MIPS_MIN_BRANCH_OFFSET || offset > MIPS_MAX_BRANCH_OFFSET);
            
            *(loc - 2) = MIPS_OUT_OF_OFFSET_FLAG;
            *(loc - 1) = offset;
        }

        static bool needLatePatch(void* code)
        {
            int32_t* loc = reinterpret_cast<int32_t*>(code);

            return (*(loc - 2) == MIPS_OUT_OF_OFFSET_FLAG);
        }

        static void repatchLatePatch(void* code)
        {
            ASSERT(needLatePatch(code));

            uint32_t* loc = reinterpret_cast<uint32_t*>(code);
            uint32_t offset = *(loc - 1);
            uint16_t lo16 = offset;
            uint16_t hi16 = offset >> 16;

            *(loc - 2) = (MIPS_LUI_OPCODE | (MIPS::at << MIPS_I_RD_SHIFT) | (hi16 & 0xFFFF));
            *(loc - 1) = (MIPS_ORI_OPCODE | (MIPS::at << MIPS_RS_SHIFT) | (MIPS::at << MIPS_I_RD_SHIFT) | (lo16 & 0xFFFF));
            *(loc) = (MIPS_JR_OPCODE | (MIPS::at << MIPS_R_RD_SHIFT));
        }

        static void patchToLoadImm32(void* code, uint32_t value)
        {
            uint32_t* loc = reinterpret_cast<uint32_t*>(code);
            uint16_t lo16 = value;
            uint16_t hi16 = value >> 16;

            *(loc - 2) = ((*(loc - 2) & 0xFFFF0000) | (hi16 & 0xFFFF));
            *(loc - 1) = ((*(loc - 1) & 0xFFFF0000) | (lo16 & 0xFFFF));
        }

        static void patchToTarget(void* code, void* address, bool link = false)
        {
            uint32_t* loc = reinterpret_cast<uint32_t*>(code);
            uint16_t lo16 = reinterpret_cast<uint32_t>(address);
            uint16_t hi16 = reinterpret_cast<uint32_t>(address) >> 16;
            
            *(loc - 2) = (MIPS_LUI_OPCODE | (MIPS::at << MIPS_I_RD_SHIFT) | (hi16 & 0xFFFF));
            *(loc - 1) = (MIPS_ORI_OPCODE | (MIPS::at << MIPS_RS_SHIFT) | (MIPS::at << MIPS_I_RD_SHIFT) | (lo16 & 0xFFFF));
            if (link)
                *(loc) = (MIPS_JALR_OPCODE | (MIPS::at << MIPS_RS_SHIFT) | (MIPS::ra << MIPS_R_RD_SHIFT));
            else
                *(loc) = (MIPS_JR_OPCODE | (MIPS::at << MIPS_R_RD_SHIFT));
        }

        static void patchToOffset(void* code, int offset, bool link = false)
        {
            uint32_t* loc = reinterpret_cast<uint32_t*>(code);
            ASSERT((offset >= MIPS_MIN_BRANCH_OFFSET) && (offset <= MIPS_MAX_BRANCH_OFFSET));
            *(loc - 2) = 0;
            *(loc - 1) = 0;
            if (link)
                *(loc) = MIPS_BGEZAL_OPCODE | (MIPS::zero << MIPS_RS_SHIFT) | (offset & 0xFFFF);
            else
                *(loc) = MIPS_BEQ_OPCODE | (offset & 0xFFFF);
        }
        
        // Patch pointers

        static void repatchInt32(void* from, int32_t value)
        {
            patchToLoadImm32(from, (uint32_t)value); 

            ExecutableAllocator::cacheFlush(reinterpret_cast<uint32_t*>(from) - 2, sizeof(int) * 2);
        }

        static void repatchPointer(void* from, void* to)
        {
            patchToLoadImm32(from, reinterpret_cast<uint32_t>(to));
            ExecutableAllocator::cacheFlush(reinterpret_cast<uint32_t*>(from) - 2, sizeof(int) * 2);
        }

        static void repatchLoadPtrToLEA(void* from)
        {
            ASSERT(0); //FIXME
#if 0            
            // On arm, this is a patch from LDR to ADD. It is restricted conversion,
            // from special case to special case, altough enough for its purpose
            ARMWord* insn = reinterpret_cast<ARMWord*>(from);
            ASSERT((*insn & 0x0ff00f00) == 0x05900000);

            *insn = (*insn & 0xf00ff0ff) | 0x02800000;
            ExecutableAllocator::cacheFlush(insn, sizeof(ARMWord));
            ExecutableAllocator::cacheFlush(reinterpret_cast<uint32_t*>(from) - 4, sizeof(int) * 4);
#endif            
        }

        // Linkers

        static void linkPointer(void* code, JmpDst where, void* value)
        {
            ASSERT(0);//FIXME need to verify
            patchToLoadImm32(reinterpret_cast<char*>(code) + where.m_offset, reinterpret_cast<uint32_t>(value)); 
        }

        void linkJump(JmpSrc from, JmpDst to)
        {
            ASSERT(to.m_offset != -1);
            ASSERT(from.m_offset != -1);

            uint32_t* location = reinterpret_cast<uint32_t*>(reinterpret_cast<intptr_t>(m_buffer.data()) + from.m_offset);
            uint32_t* target = reinterpret_cast<uint32_t*>(reinterpret_cast<intptr_t>(m_buffer.data()) + to.m_offset);
            int offset = target - location - 1;
            if (offset <= MIPS_MAX_BRANCH_OFFSET && offset >= MIPS_MIN_BRANCH_OFFSET) {
                patchToOffset(location, offset);
            } else {
                if (!needLatePatch(location))
                    m_jumps.append(from.m_offset);
                markNeedLatePatch(location, offset);
            }
        }

        static void linkJump(void* code, JmpSrc from, void* to)
        {
            ASSERT(!(reinterpret_cast<intptr_t>(code) & 3));
            ASSERT(from.m_offset != -1);
            ASSERT(reinterpret_cast<intptr_t>(to) & 3);

            uint32_t* location = (uint32_t*)(reinterpret_cast<intptr_t>(code) + from.m_offset);
            patchToTarget(location, to);
        }
        
        static void relinkJump(void* from, void* to)
        {
            ASSERT(!(reinterpret_cast<intptr_t>(from) & 3));
            ASSERT(reinterpret_cast<intptr_t>(to) & 3);
            patchToTarget(from, to);
            ExecutableAllocator::cacheFlush(reinterpret_cast<uint32_t*>(from) - 2, 2 * sizeof(uint32_t));
        }

        static void linkCall(void* code, JmpSrc from, void* to)
        {
            linkJump(code, from, to);
        }

        static void relinkCall(void* from, void* to)
        {
            relinkJump(from, to);
        }

        // Address operations

        static void* getRelocatedAddress(void* code, JmpSrc jump)
        {
            ASSERT(jump.m_offset != -1);

            return reinterpret_cast<void*>(reinterpret_cast<ptrdiff_t>(code) + jump.m_offset);
        }
    
        static void* getRelocatedAddress(void* code, JmpDst destination)
        {
            ASSERT(destination.m_offset != -1);

            return reinterpret_cast<void*>(reinterpret_cast<ptrdiff_t>(code) + destination.m_offset);
        }
    
        static int getDifferenceBetweenLabels(JmpDst src, JmpDst dst)
        {
            return dst.m_offset - src.m_offset;
        }
    
        static int getDifferenceBetweenLabels(JmpDst src, JmpSrc dst)
        {
            return dst.m_offset - src.m_offset;
        }
    
        static int getDifferenceBetweenLabels(JmpSrc src, JmpDst dst)
        {
            return dst.m_offset - src.m_offset;
        }

        static unsigned getCallReturnOffset(JmpSrc call)
        {
            return call.m_offset + sizeof(int);
        }

        // Constant pool hnadlers

        static int placeConstantPoolBarrier(int offset)
        {
            UNUSED_PARAM(offset);
            ASSERT_NOT_REACHED();
            return 0;
        }

        AssemblerBuffer m_buffer;
        Jumps   m_jumps;
    };

} // namespace JSC

#endif // ENABLE(ASSEMBLER) && PLATFORM(MIPS)

#endif // MIPSAssembler_h

