/*
 * Copyright (C) 2008 Apple Inc.
 * Copyright (C) 2009 University of Szeged
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef MacroAssemblerMIPS_h
#define MacroAssemblerMIPS_h

#include <wtf/Platform.h>

#if ENABLE(ASSEMBLER) && PLATFORM(MIPS)

#include "MIPSAssembler.h"
#include "AbstractMacroAssembler.h"

namespace JSC {

class MacroAssemblerMIPS : public AbstractMacroAssembler<MIPSAssembler> {
public:
   
    static const Scale ScalePtr = TimesFour;

    enum Condition {
        Equal = MIPSAssembler::ConditionE,
        NotEqual = MIPSAssembler::ConditionNE,
        Above = MIPSAssembler::ConditionA,
        AboveOrEqual = MIPSAssembler::ConditionAE,
        Below = MIPSAssembler::ConditionB,
        BelowOrEqual = MIPSAssembler::ConditionBE,
        GreaterThan = MIPSAssembler::ConditionG,
        GreaterThanOrEqual = MIPSAssembler::ConditionGE,
        LessThan = MIPSAssembler::ConditionL,
        LessThanOrEqual = MIPSAssembler::ConditionLE,
        Overflow = 0xFF, //MIPSAssembler::ConditionE, //FIXME
        Signed = MIPSAssembler::ConditionLZ,
        Zero = MIPSAssembler::ConditionE,
        NonZero = MIPSAssembler::ConditionNE
    };
   
    enum DoubleCondition {
        DoubleEqual = MIPSAssembler::ConditionE,
        DoubleGreaterThan = MIPSAssembler::ConditionG,
        DoubleGreaterThanOrEqual = MIPSAssembler::ConditionGE,
        DoubleLessThan = MIPSAssembler::ConditionL,
        DoubleLessThanOrEqual = MIPSAssembler::ConditionLE,
    };

    static const RegisterID stackPointerRegister = MIPS::sp;
    static const RegisterID linkRegister  = MIPS::ra;
    static const RegisterID DR0 = MIPS::t7; /* it's used for load immediate */
    static const RegisterID DR1 = MIPS::t8;
    static const RegisterID AR0 = MIPS::t9; /* it's used for load & store */

    void add32(RegisterID src, RegisterID dest)
    {
        m_assembler.add(dest, dest, src);
    }

    void add32(Imm32 imm, Address address)
    {        
        load32(address, DR0);
        move(imm, DR1);
        add32(DR1, DR0);
        store32(DR0, address);
    }

    void add32(Imm32 imm, RegisterID dest)
    {
        move(imm, DR0);
        m_assembler.add(dest, dest, DR0);
    }

    void add32(Address src, RegisterID dest)
    {
        load32(src, DR0);
        add32(DR0, dest);
    }

    void and32(RegisterID src, RegisterID dest)
    {
        m_assembler.and32(dest, dest, src);
    }

    void and32(Imm32 imm, RegisterID dest)
    {
        move(imm, DR0);
        and32(DR0, dest);
    }

    void lshift32(Imm32 imm, RegisterID dest)
    {
        m_assembler.sll(dest, dest, imm.m_value);
    }

    void lshift32(RegisterID shift_amount, RegisterID dest)
    {
        m_assembler.sllv(dest, dest, shift_amount);
    }

    void mul32(RegisterID src, RegisterID dest)
    {
        if (src == dest) {
            move(src, DR0);
            src = DR0;
        }
        m_assembler.mul(dest, dest, src);
    }

    void mul32(Imm32 imm, RegisterID src, RegisterID dest)
    {
        move(imm, DR0);
        m_assembler.mul(dest, src, DR0);
    }

    void not32(RegisterID dest)
    {
        m_assembler.not32(dest, dest);
    }

    void or32(RegisterID src, RegisterID dest)
    {
        m_assembler.or32(dest, dest, src);
    }

    void or32(Imm32 imm, RegisterID dest)
    {
        move(imm, DR0);
        m_assembler.or32(dest, dest, DR0);
    }

    void rshift32(RegisterID shift_amount, RegisterID dest)
    {
        m_assembler.srlv(dest, dest, shift_amount);
    }

    void rshift32(Imm32 imm, RegisterID dest)
    {
        m_assembler.srl(dest, dest, imm.m_value);
    }

    void sub32(RegisterID src, RegisterID dest)
    {
        m_assembler.sub(dest, dest, src);
    }

    void sub32(Imm32 imm, RegisterID dest)
    {
        move(imm, DR0);
        m_assembler.sub(dest, dest, DR0);
    }

    void sub32(Imm32 imm, Address address)
    {
        move(imm, DR0);
        load32(address, DR1);
        m_assembler.sub(DR1, DR1, DR0);
        store32(DR1, address);
    }

    void sub32(Address src, RegisterID dest)
    {
        load32(src, DR0);
        sub32(DR0, dest);
    }

    void xor32(RegisterID src, RegisterID dest)
    {
        m_assembler.eor(dest, dest, src);
    }

    void xor32(Imm32 imm, RegisterID dest)
    {
        move(imm, DR0);
        m_assembler.eor(dest, dest, DR0);
    }

    void load32(ImplicitAddress address, RegisterID dest)
    {        
        m_assembler.load32(dest, address.base, address.offset);
    }

    void load32(BaseIndex address, RegisterID dest)
    { 
        m_assembler.move(AR0, address.index);
        m_assembler.sll(AR0, AR0, address.scale);
        m_assembler.add(AR0, AR0, address.base);
        m_assembler.load32(dest, AR0, address.offset);
    }

    DataLabel32 load32WithAddressOffsetPatch(Address address, RegisterID dest)
    {        
        DataLabel32 label(this);
        m_assembler.load32(dest, address.base, address.offset);
        return label;
    }

    Label loadPtrWithPatchToLEA(Address address, RegisterID dest)
    {
        Label label(this);
        m_assembler.load32(dest, address.base, address.offset);
        return label;
    }

    void load16(BaseIndex address, RegisterID dest)
    {
        m_assembler.move(DR0, address.index);
        m_assembler.sll(DR0, DR0, address.scale);
        m_assembler.add(DR0, DR0, address.base);
        m_assembler.load16(dest, DR0, address.offset);
    }

    void store32(RegisterID src, ImplicitAddress address)
    {
        m_assembler.store32(src, address.base, address.offset);
    }

    void store32(RegisterID src, BaseIndex address)
    {
        m_assembler.sll(DR0, address.index, address.scale);
        m_assembler.add(DR0, DR0, address.base);
        m_assembler.store32(src, DR0, address.offset);
    }

    void store32(Imm32 imm, ImplicitAddress address)
    {
        move(imm, DR0);
        store32(DR0, address);
    }

    void store32(RegisterID src, void* address)
    {
        m_assembler.moveImm32(DR0, reinterpret_cast<int>(address));
        m_assembler.store32(src, DR0);
    }

    void store32(Imm32 imm, void* address)
    {
        m_assembler.moveImm32(DR0, reinterpret_cast<int>(address));
        m_assembler.moveImm32(DR1, imm.m_value);
        m_assembler.store32(DR1, DR0);
    }

    DataLabel32 store32WithAddressOffsetPatch(RegisterID src, Address address)
    {
        DataLabel32 label(this);
        m_assembler.store32(src, address.base, address.offset);
        return label;
    }

    void pop(RegisterID dest)
    {
        m_assembler.load32(dest, stackPointerRegister);
        m_assembler.addi(stackPointerRegister, stackPointerRegister, 4);
    }

    void push(RegisterID src)
    {
        m_assembler.store32(src, stackPointerRegister);
        m_assembler.subi(stackPointerRegister, stackPointerRegister, 4);
    }

    void push(Address address)
    {
        load32(address, DR0);
        push(DR0);
    }

    void push(Imm32 imm)
    {
        move(imm, DR0);
        push(DR0);
    }

    void move(Imm32 imm, RegisterID dest)
    {
        m_assembler.lui(dest, ((imm.m_value >> 16) & 0xFFFF));
        m_assembler.ori(dest, dest, (imm.m_value & 0xFFFF));
    }

    void move(RegisterID src, RegisterID dest)
    {
        m_assembler.move(dest, src);
    }

    void move(ImmPtr imm, RegisterID dest)
    {        
        m_assembler.moveImm32(dest, reinterpret_cast<int>(imm.m_value));
    }

    void swap(RegisterID reg1, RegisterID reg2)
    {
        m_assembler.eor(reg1, reg1, reg2);
        m_assembler.eor(reg2, reg2, reg1);
        m_assembler.eor(reg1, reg1, reg2);
    }

    void signExtend32ToPtr(RegisterID src, RegisterID dest)
    {
        if (src != dest)
            move(src, dest);
    }

    void zeroExtend32ToPtr(RegisterID src, RegisterID dest)
    {
        if (src != dest)
            move(src, dest);
    }

    Jump branch32(Condition cond, RegisterID left, RegisterID right)
    {
        m_assembler.branch(m_assembler.getOppositedCond(MIPSCondition(cond)), left, right, 5);
        return Jump(m_assembler.jmp());
    }

    Jump branch32(Condition cond, RegisterID left, Imm32 right)
    {
        move(right, DR0);
        return branch32(cond, left, DR0);
    }

    Jump branch32(Condition cond, RegisterID left, Address right)
    {
        load32(right, DR0);
        return branch32(cond, left, DR0);
    }

    Jump branch32(Condition cond, Address left, RegisterID right)
    {
        load32(left, DR0);
        return branch32(cond, DR0, right);
    }

    Jump branch32(Condition cond, Address left, Imm32 right)
    {
        load32(left, DR0);
        move(right, DR1);
        return branch32(cond, DR0, DR1);
    }

    Jump branch32(Condition cond, BaseIndex left, Imm32 right)
    {
        load32(left, DR0);
        move(right, DR1);
        return branch32(cond, DR0, DR1);
    }

    Jump branch16(Condition cond, BaseIndex left, RegisterID right)
    {
        UNUSED_PARAM(cond);
        UNUSED_PARAM(left);
        UNUSED_PARAM(right);
        ASSERT_NOT_REACHED();
        return jump();
    }

    Jump branch16(Condition cond, BaseIndex left, Imm32 right)
    {
        load16(left, DR0);
        move(right, DR1);
        return branch32(cond, DR0, DR1);
    }

    Jump branchTest32(Condition cond, RegisterID reg, RegisterID mask)
    {
        ASSERT((cond == Zero) || (cond == NonZero));
        m_assembler.and32(DR0, reg, mask);
        return branch32(cond, DR0, MIPS::zero);
    }

    Jump branchTest32(Condition cond, RegisterID reg, Imm32 mask = Imm32(-1))
    {
        ASSERT((cond == Zero) || (cond == NonZero));
        if (mask.m_value != -1) {
            move(mask, DR0);
            m_assembler.and32(DR0, reg, DR0);
            return branch32(cond, DR0, MIPS::zero);
        } else
            return branch32(cond, reg, MIPS::zero);
    }

    Jump branchTest32(Condition cond, Address address, Imm32 mask = Imm32(-1))
    {
        load32(address, DR1);
        return branchTest32(cond, DR1, mask);
    }

    Jump branchTest32(Condition cond, BaseIndex address, Imm32 mask = Imm32(-1))
    {
        load32(address, DR1);
        return branchTest32(cond, DR1, mask);
    }

    Jump jump()
    {
        return Jump(m_assembler.jmp());
    }

    void jump(RegisterID target)
    {
        m_assembler.jmp(target);
    }

    void jump(Address address)
    {
        load32(address, DR0);
        m_assembler.jmp(DR0);
    }

    Jump branchAdd32(Condition cond, RegisterID src, RegisterID dest)
    {
        ASSERT_NOT_REACHED(); //FIXME
        ASSERT((cond == Overflow) || (cond == Signed) || (cond == Zero) || (cond == NonZero));
        add32(src, dest);
        switch(cond) {
            case Overflow:
                m_assembler.eor(DR0, src, dest);        //check if sign differ
                m_assembler.slt(DR0, DR0, MIPS::zero);  // DR0 = 1 if differ
                m_assembler.bne(DR0, MIPS::zero, 8, false);    // signs of src dest different, no overflow
                m_assembler.eor(DR0, dest, src);        //sign of sum src different
                m_assembler.slt(DR0, DR0, MIPS::zero);  //DR0 = 1 if sum has different
                m_assembler.beq(DR0, MIPS::zero, 5);     //no overflow
                return jump();
            case Signed:
            case Zero:
            case NonZero:
                return branch32(cond, dest, MIPS::zero);
        }
    }

    Jump branchAdd32(Condition cond, Imm32 imm, RegisterID dest)
    {
        ASSERT((cond == Overflow) || (cond == Signed) || (cond == Zero) || (cond == NonZero));
        move(imm, DR1);
        return branchAdd32(cond, DR1, dest);
    }

    Jump branchMul32(Condition cond, RegisterID src, RegisterID dest)
    {
        ASSERT_NOT_REACHED(); //FIXME
        ASSERT((cond == Overflow) || (cond == Signed) || (cond == Zero) || (cond == NonZero));
        if (cond == Overflow) {
            if (src == dest) {
                move(src, DR0);
                src = DR0;
            }
            m_assembler.mult(DR1, dest, dest, src);
            dest = DR1;
            cond = NonZero;
        }
        else
            m_assembler.mul(dest, dest, src);
        return branchTest32(cond, dest);
    }

    Jump branchMul32(Condition cond, Imm32 imm, RegisterID src, RegisterID dest)
    {
        ASSERT((cond == Overflow) || (cond == Signed) || (cond == Zero) || (cond == NonZero));
        move(imm, DR0);
        if (cond == Overflow) {
            m_assembler.mult(DR1, dest, src, DR0);
            dest = DR1;
            cond = NonZero;
        } else
            m_assembler.mul(dest, src, DR0);
        return branchTest32(cond, dest);
    }

    Jump branchSub32(Condition cond, RegisterID src, RegisterID dest)
    {
        ASSERT_NOT_REACHED(); //FIXME
        ASSERT((cond == Overflow) || (cond == Signed) || (cond == Zero) || (cond == NonZero));
        sub32(src, dest);
        switch(cond) {
            case Overflow:
                m_assembler.eor(DR0, src, dest);        //check if sign differ
                m_assembler.slt(DR0, DR0, MIPS::zero);  // DR0 = 1 if differ
                m_assembler.beq(DR0, MIPS::zero, 8);    // signs of src dest same, no overflow
                m_assembler.eor(DR0, dest, src);        //sign of sum src different
                m_assembler.slt(DR0, DR0, MIPS::zero);  //DR0 = 1 if sum has different
                m_assembler.bne(DR0, MIPS::zero, 5);     //no overflow
                m_assembler.nop();
                return jump();
            case Signed:
            case Zero:
            case NonZero:
                return branch32(cond, dest, MIPS::zero);
        }
    }

    Jump branchSub32(Condition cond, Imm32 imm, RegisterID dest)
    {
        move(imm, DR1);
        return branchSub32(cond, DR1, dest);
    }

    void breakpoint()
    {
        m_assembler.bkpt(0);
    }

    Call nearCall()
    {
        return Call(m_assembler.call(), Call::LinkableNear);
    }

    Call call(RegisterID target)
    {
        return Call(m_assembler.call(target), Call::None);
    }

    Call call(Address address)
    {
        load32(address, DR0);
        return Call(m_assembler.call(DR0), Call::None);
    }

    void ret()
    {
        m_assembler.ret();
    }

    void set32(Condition cond, RegisterID left, RegisterID right, RegisterID dest)
    {
        ASSERT((cond == Equal) || (cond == NotEqual));
        if (cond == Equal) {
            m_assembler.eor(dest, left, right);
            m_assembler.sltiu(dest, dest, 1);
        } else {
            m_assembler.eor(dest, left, right);
            m_assembler.sltu(dest, MIPS::zero, dest);
        }
    }

    void set32(Condition cond, RegisterID left, Imm32 right, RegisterID dest)
    {
        move(right, DR0);
        set32(cond, left, DR0, dest);
    }

    void setTest32(Condition cond, Address address, Imm32 mask, RegisterID dest)
    {
        ASSERT((cond == NonZero) || (cond == Zero));
        load32(address, DR0);
        if (mask.m_value != -1) {
            move(mask, DR1);
            m_assembler.and32(DR0, DR0, DR1);
        }
        set32(cond == NonZero ? NotEqual : Equal, DR0, MIPS::zero, dest);
    }

    void add32(Imm32 imm, RegisterID src, RegisterID dest)
    {
        move(imm, DR0);
        m_assembler.add(dest, src, DR0);
    }

    void add32(Imm32 imm, AbsoluteAddress address)
    {
        move(imm, DR0);
        move(Imm32(reinterpret_cast<int>(address.m_ptr)), AR0);
        m_assembler.load32(DR1, AR0);
        m_assembler.add(DR1, DR1, DR0);
        m_assembler.store32(AR0, DR1);
    }

    void sub32(Imm32 imm, AbsoluteAddress address)
    {
        move(imm, DR0);
        move(Imm32(reinterpret_cast<int>(address.m_ptr)), AR0);
        m_assembler.load32(DR1, AR0);
        m_assembler.sub(DR1, DR1, DR0);
        m_assembler.store32(AR0, DR1);
    }

    void load32(void* address, RegisterID dest)
    {
        m_assembler.moveImm32(AR0, reinterpret_cast<int>(address));
        m_assembler.load32(dest, AR0);
    }

    Jump branch32(Condition cond, AbsoluteAddress left, RegisterID right)
    {
        load32(left.m_ptr, DR0);
        return branch32(cond, DR0, right);
    }

    Jump branch32(Condition cond, AbsoluteAddress left, Imm32 right)
    {
        load32(left.m_ptr, DR0);
        move(right, DR1);
        return branch32(cond, DR0, DR1);
    }

    Call call()
    {
        return Call(m_assembler.call(), Call::Linkable);
    }

    Call tailRecursiveCall()
    {
        return Call::fromTailJump(jump());
    }

    Call makeTailRecursiveCall(Jump oldJump)
    {
        return Call::fromTailJump(oldJump);
    }

    DataLabel32 moveWithPatch(Imm32 imm, RegisterID dest)
    {
        move(imm, dest);
        return DataLabel32(this);
    }

    DataLabelPtr moveWithPatch(ImmPtr initialValue, RegisterID dest)
    {
        move(Imm32(initialValue), dest);
        return DataLabelPtr(this);
    }

    Jump branchPtrWithPatch(Condition cond, RegisterID left, DataLabelPtr& dataLabel, ImmPtr initialRightValue = ImmPtr(0))
    {
        dataLabel = moveWithPatch(initialRightValue, DR0);
        Jump jump = branch32(cond, left, DR0);
        jump.enableLatePatch();
        return jump;
    }

    Jump branchPtrWithPatch(Condition cond, Address left, DataLabelPtr& dataLabel, ImmPtr initialRightValue = ImmPtr(0))
    {
        load32(left, DR1);
        dataLabel = moveWithPatch(initialRightValue, DR0);
        Jump jump = branch32(cond, DR0, DR1);
        jump.enableLatePatch();
        return jump;
    }

    DataLabelPtr storePtrWithPatch(ImmPtr initialValue, ImplicitAddress address)
    {
        DataLabelPtr dataLabel = moveWithPatch(initialValue, DR0);
        store32(DR0, address);
        return dataLabel;
    }

    DataLabelPtr storePtrWithPatch(ImplicitAddress address)
    {
        return storePtrWithPatch(ImmPtr(0), address);
    }

    // Floating point operators
    bool supportsFloatingPoint() const
    {
        return false;
    }

    bool supportsFloatingPointTruncate() const
    {
        return false;
    }

    void loadDouble(ImplicitAddress address, FPRegisterID dest)
    {
        UNUSED_PARAM(address);
        UNUSED_PARAM(dest);
        ASSERT_NOT_REACHED();
    }

    void storeDouble(FPRegisterID src, ImplicitAddress address)
    {
        UNUSED_PARAM(src);
        UNUSED_PARAM(address);
        ASSERT_NOT_REACHED();
    }

    void addDouble(FPRegisterID src, FPRegisterID dest)
    {
        UNUSED_PARAM(src);
        UNUSED_PARAM(dest);
        ASSERT_NOT_REACHED();
    }

    void addDouble(Address src, FPRegisterID dest)
    {
        UNUSED_PARAM(src);
        UNUSED_PARAM(dest);
        ASSERT_NOT_REACHED();
    }

    void subDouble(FPRegisterID src, FPRegisterID dest)
    {
        UNUSED_PARAM(src);
        UNUSED_PARAM(dest);
        ASSERT_NOT_REACHED();
    }

    void subDouble(Address src, FPRegisterID dest)
    {
        UNUSED_PARAM(src);
        UNUSED_PARAM(dest);
        ASSERT_NOT_REACHED();
    }

    void mulDouble(FPRegisterID src, FPRegisterID dest)
    {
        UNUSED_PARAM(src);
        UNUSED_PARAM(dest);
        ASSERT_NOT_REACHED();
    }

    void mulDouble(Address src, FPRegisterID dest)
    {
        UNUSED_PARAM(src);
        UNUSED_PARAM(dest);
        ASSERT_NOT_REACHED();
    }

    void convertInt32ToDouble(RegisterID src, FPRegisterID dest)
    {
        UNUSED_PARAM(src);
        UNUSED_PARAM(dest);
        ASSERT_NOT_REACHED();
    }

    Jump branchDouble(DoubleCondition cond, FPRegisterID left, FPRegisterID right)
    {
        UNUSED_PARAM(cond);
        UNUSED_PARAM(left);
        UNUSED_PARAM(right);
        ASSERT_NOT_REACHED();
        return jump();
    }

    // Truncates 'src' to an integer, and places the resulting 'dest'.
    // If the result is not representable as a 32 bit value, branch.
    // May also branch for some values that are representable in 32 bits
    // (specifically, in this case, INT_MIN).
    Jump branchTruncateDoubleToInt32(FPRegisterID src, RegisterID dest)
    {
        UNUSED_PARAM(src);
        UNUSED_PARAM(dest);
        ASSERT_NOT_REACHED();
        return jump();
    }

protected:
    MIPSAssembler::Condition MIPSCondition(Condition cond)
    {
        return static_cast<MIPSAssembler::Condition>(cond);
    }

private:
    friend class LinkBuffer;
    friend class RepatchBuffer;

    static void linkCall(void* code, Call call, FunctionPtr function)
    {
        MIPSAssembler::linkCall(code, call.m_jmp, function.value());
    }

    static void repatchCall(CodeLocationCall call, CodeLocationLabel destination)
    {
        MIPSAssembler::relinkCall(call.dataLocation(), destination.executableAddress());
    }

    static void repatchCall(CodeLocationCall call, FunctionPtr destination)
    {
        MIPSAssembler::relinkCall(call.dataLocation(), destination.executableAddress());
    }

};

}

#endif

#endif // MacroAssemblerMIPS_h
