/*
 * Copyright (C) 2008 Pleyo.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Pleyo nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PLEYO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL PLEYO OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "IntRect.h"
#include "FrameView.h"
#include "Frame.h"
#include "GraphicsContext.h"
#include "Page.h"
#include "EventHandler.h"
#include "FocusController.h"
#include "HitTestResult.h"
#include "HitTestRequest.h"
#include "RenderTheme.h"
#include <MainThread.h>
#include <KeyboardCodes.h>
#include <PlatformKeyboardEvent.h>
#include <PlatformMouseEvent.h>
#include <PlatformWheelEvent.h>
#include "SelectionController.h"
#include <SharedTimer.h>
#include "Editor.h"
#include "owb-config.h"
#include "PopupMenu.h"
#include "CString.h"
#include "FileIO.h"
#include "WebFrame.h"
#include "WebView.h"
#include "CurrentTime.h"
#include <directfb.h>
#include "WebViewPrivate.h"
#include "WebViewDelegate.h"

#include <HTMLTextAreaElement.h>
#include <HTMLInputElement.h>
#include <HTMLNames.h>
#include <RenderTreeAsText.h>
#include <RenderView.h>
#include <PlatformString.h>
#include <CString.h>
#include <NodeList.h>

#if PLATFORM(DFB_CAIRO)
#include <cairo-directfb.h>
#endif

#include <FL/Fl_Window.H>

#if ENABLE(ACTVILA)
#include <acnavi.h>
#endif

using namespace WebCore;

WebViewPrivate::WebViewPrivate(WebView *webView)
    : m_webView(webView)
    , isInitialized(false)
    , m_inDrawing(false)
    , m_repaintTimer(this, &WebViewPrivate::repaintTimerFired)
{
#if PLATFORM(DFB_CAIRO)
    m_cairoSurface = NULL;
#endif
    m_context = NULL;
    m_cachedRoot = NULL;  
	m_lastFocused = NULL;
	m_domtree_version = 0;
	m_check_domtree_version = true;    
    const char* env = getenv("VA_WEBKIT_MOUSE_SIMULATE");
    if (env && strcasecmp(env, "true") == 0)
        mEnableMouseSimulate = true;
    else
        mEnableMouseSimulate = false;
}

void WebViewPrivate::fireWebKitEvents()
{
    WTF::dispatchFunctionsFromMainThread();
    fireTimerIfNeeded();
    //usleep(10000);
}

void WebViewPrivate::onExpose(BalEventExpose event)
{
    Frame* frame = core(m_webView->mainFrame());
    if (!frame || !m_context)
        return;

    m_inDrawing = true;

    if(!isInitialized) {
        isInitialized = true;
        frame->view()->resize(m_rect.width(), m_rect.height());
        frame->view()->forceLayout();
        frame->view()->adjustViewSize();
        m_webView->updateFocusedAndActiveState();
    }

    GraphicsContext ctx(m_context);
    ctx.setBalExposeEvent(&event);
    IntRect rect(m_webView->dirtyRegion());
    if (frame->contentRenderer() && frame->view() && !rect.isEmpty()) {
        frame->view()->layoutIfNeededRecursive();
        IntRect dirty = rect;
        IntRect d = rect;
        if (!m_hUpdateRect.isEmpty()) {
            frame->view()->paint(&ctx, m_hUpdateRect);
            dirty.unite(m_hUpdateRect);
            m_hUpdateRect = IntRect();
        }
        if (!m_vUpdateRect.isEmpty()) {
            frame->view()->paint(&ctx, m_vUpdateRect);
            dirty.unite(m_vUpdateRect);
            m_vUpdateRect = IntRect();
        }
	    //printf("%s : dirty = {%d,  %d, %d, %d}\n", __FUNCTION__, dirty.x(), dirty.y(), dirty.width(), dirty.height());
        frame->view()->paint(&ctx, d);

        //if (m_scrollSurface)
        {
            //FIXME
        }

	    //drawFocusRing(ctx, dirty);
	 
        updateView(m_webView->viewWindow(), dirty);
        m_webView->clearDirtyRegion();
		
		if (mEnableMouseSimulate) {
			    CacheBuilder& builder = getCacheBuilder();
    			WebCore::Node* oldFocusNode = builder.currentFocus();    		
    			WebCore::IntRect oldBounds = oldFocusNode ?
        					oldFocusNode->getRect() : WebCore::IntRect(0,0,0,0);
    			DBG_NAV_LOGD("m_lastFocused=%p oldFocusNode=%p"
        			" m_lastFocusedBounds={%d,%d,%d,%d} oldBounds={%d,%d,%d,%d}\n",
        				m_lastFocused, oldFocusNode,
        				m_lastFocusedBounds.x(), m_lastFocusedBounds.y(), m_lastFocusedBounds.width(), m_lastFocusedBounds.height(),
        				oldBounds.x(), oldBounds.y(), oldBounds.width(), oldBounds.height());

    				unsigned latestVersion = 0;
					if (m_check_domtree_version) {
        			    // as domTreeVersion only increment, we can just check the sum to see 
        			    // whether we need to update the frame cache
        			    for (Frame* frame = getCoreMainFrame(); frame; frame = frame->tree()->traverseNext()) {
            				latestVersion += frame->document()->domTreeVersion();
        		    	}
					}  else
						latestVersion = m_domtree_version;

			    if (m_lastFocused != oldFocusNode || m_lastFocusedBounds != oldBounds || 
					(m_check_domtree_version && latestVersion != m_domtree_version)) {
			        m_lastFocused = oldFocusNode;
			        m_lastFocusedBounds = oldBounds;
			        DBG_NAV_LOGD("call updateFrameCache m_domtree_version=%d latest=%d\n",
			            m_domtree_version, latestVersion);
			        m_domtree_version = latestVersion;
			        updateFrameCache();
			    }	
			}
    }
    m_inDrawing = false;
}

#if DUMP_NAV_CACHE
static void debugDump(CachedRoot* root)
{   
    if (root)
        root->mDebug.print();
}
#endif

void WebViewPrivate::setCharInput(const char* text)
{
    if (text) {
        Node* focusNode = getCacheBuilder().currentFocus();
        if (!focusNode) {
            DBG_NAV_LOG("!focus");
            return;
        }
        bool isTextArea = focusNode->hasTagName(WebCore::HTMLNames::textareaTag);
        bool isTextField = (focusNode->hasTagName(WebCore::HTMLNames::inputTag) 
                && ((WebCore::HTMLInputElement*)(focusNode))->isTextField());

        String t = String::fromUTF8(text);

        if (isTextArea)
            reinterpret_cast<WebCore::HTMLTextAreaElement*>(focusNode)->setValue(t);
        else if (isTextField)
            reinterpret_cast<WebCore::HTMLInputElement*>(focusNode)->setValue(t);
    }
}

bool WebViewPrivate::displaySoftKeyboard()
{
    WebViewDelegate* softKeyboard = m_webView->webViewDelegate();
    if (!softKeyboard) {
        DBG_NAV_LOG("!softkeyboard");
        return false;
    }
    Node* focusNode = getCacheBuilder().currentFocus();
    if (!focusNode) {
        DBG_NAV_LOG("!focus");
        return false;
    }
    bool isTextArea = focusNode->hasTagName(WebCore::HTMLNames::textareaTag);
    bool isTextField = (focusNode->hasTagName(WebCore::HTMLNames::inputTag) 
            && ((WebCore::HTMLInputElement*)(focusNode))->isTextField());

    if (!isTextArea && !isTextField) {
        DBG_NAV_LOG("not text area");
        return false;
    }
#if 1    
#if 0
    AcNavi_StartCharInput();    
#else
    bool isPassword = false;
    WebCore::String initText;
    if (isTextField) {
        isPassword = ((WebCore::HTMLInputElement*)(focusNode))->inputType() == WebCore::HTMLInputElement::PASSWORD;
        initText = ((WebCore::HTMLInputElement*)(focusNode))->value();
    } else
        initText = ((WebCore::HTMLTextAreaElement*)(focusNode))->value();

    const char* text = softKeyboard->displaySoftKeyboard(m_webView, initText.utf8().data(), isPassword);
    setCharInput(text);
#endif
#else
    WebCore::IntRect scrollViewRect;
    getVisibleRect(scrollViewRect);
	softKeyboard->displaySoftKeyboard(m_webView,focus->bounds().x() - scrollViewRect.x() ,focus->bounds().y() - scrollViewRect.y(),
	                               focus->bounds().width(),focus->bounds().height());
#endif
    return true;
}

int WebViewPrivate::onKeyDown(BalEventKey event)
{
    PlatformKeyboardEvent keyboardEvent(&event);

    //deal with soft keyboard
    if(keyboardEvent.windowsVirtualKeyCode() == VK_RETURN
       && !keyboardEvent.shiftKey()
       && !keyboardEvent.ctrlKey()
       && !keyboardEvent.altKey()
       && displaySoftKeyboard())
        return 1;

    Frame* mainFrame = getCoreMainFrame();
    Frame* frame = m_webView->page()->focusController()->focusedOrMainFrame();
    if (!frame || !mainFrame)
        return 1;

    EventHandler* eventHandler = frame->eventHandler();
    Node* focusNode = getCacheBuilder().currentFocus();
    if (focusNode)
        eventHandler = focusNode->document()->frame()->eventHandler();

	int keyevent_ret = 0;
    if ((keyevent_ret = eventHandler->keyEvent(keyboardEvent)))
    {
        if(VK_TAB == keyboardEvent.windowsVirtualKeyCode()){
            Node* node = getCacheBuilder().currentFocus();
            moveMouseOnElement(static_cast<WebCore::Element*>(node));
        }
        return keyevent_ret;
    }

	if (mEnableMouseSimulate
       && !keyboardEvent.shiftKey()
       && !keyboardEvent.ctrlKey()
       && !keyboardEvent.altKey()
       && !moveFocus(keyboardEvent.windowsVirtualKeyCode(), 1, false, false, m_lastFocused, &m_lastFocusedBounds)) {		
		return 0;
    }

    FrameView* view = mainFrame->view();
    SelectionController::EAlteration alteration;
    if (keyboardEvent.shiftKey())
        alteration = SelectionController::EXTEND;
    else
        alteration = SelectionController::MOVE;

    switch(keyboardEvent.windowsVirtualKeyCode())
    {
        case VK_DOWN:
            if (!frame->eventHandler()->scrollOverflow(ScrollDown, ScrollByLine))
                view->scroll(ScrollDown, ScrollByLine);
            return 1;
        case VK_UP:
            if (!frame->eventHandler()->scrollOverflow(ScrollUp, ScrollByLine))
                view->scroll(ScrollUp, ScrollByLine);
            return 1;
        case VK_RIGHT:
            if (!frame->eventHandler()->scrollOverflow(ScrollRight, ScrollByLine))
                view->scroll(ScrollRight, ScrollByLine);
            return 1;
        case VK_LEFT:
            if (!frame->eventHandler()->scrollOverflow(ScrollLeft, ScrollByLine))
                view->scroll(ScrollLeft, ScrollByLine);
            return 1;
        case VK_NEXT:
            if (!mainFrame->eventHandler()->scrollOverflow(ScrollDown, ScrollByDocument))
                view->scroll(ScrollDown, ScrollByPage);
            return 1;
        case VK_PRIOR:
            if (!mainFrame->eventHandler()->scrollOverflow(ScrollUp, ScrollByDocument))
                view->scroll(ScrollUp, ScrollByPage);
            return 1;
        case VK_HOME:
            mainFrame->selection()->modify(alteration, SelectionController::BACKWARD, DocumentBoundary, true);
            return 1;
        case VK_END:
            mainFrame->selection()->modify(alteration, SelectionController::FORWARD, DocumentBoundary, true);
            return 1;
        case VK_ESCAPE:
            //FIXME
            return 1;
        case VK_F8:
            {
                WebCore::CString renderDump = WebCore::externalRepresentation(mainFrame->contentRenderer()).utf8();
                printf("\n\n%s\n\n", renderDump.data());
            }
            break;
	case VK_F9:
            printf(" mainFrame = %p , focusFrame = %p\n", mainFrame, frame);
            for(Frame* child = mainFrame->tree()->traverseNext(); child; child = child->tree()->traverseNext())
                printf("subFrame = %p\n", child);
		#if DUMP_NAV_CACHE
			debugDump(getFrameCache(DontAllowNewer));
		#endif
		break;
		
        case VK_BROWSER_BACK:
        case VK_F1:
        {
            m_webView->goBack();            
            if (m_context && mainFrame->contentRenderer() && mainFrame->view())
            {
                GraphicsContext ctx(m_context);
                mainFrame->view()->layoutIfNeededRecursive();
                IntRect dirty(0, 0, m_rect.width(), m_rect.height());
                mainFrame->view()->paint(&ctx, dirty);
            }
            return 1;
        }
        case VK_MENU:
        {
		    const char* szMainWebSite = getenv("VA_WEBKIT_HOMEPAGE");
		    if (szMainWebSite)
		        m_webView->mainFrame()->loadURL(szMainWebSite);
		    else
		        m_webView->mainFrame()->loadURL("http://www.youtube.com/xl");
        	return 1;
        }
        case VK_BROWSER_FORWARD:
        case VK_F2:
        {
            m_webView->goForward();
            if (m_context && mainFrame->contentRenderer() && frame->view())
            {
                GraphicsContext ctx(m_context);
                mainFrame->view()->layoutIfNeededRecursive();
                IntRect dirty(0, 0, m_rect.width(), m_rect.height());
                mainFrame->view()->paint(&ctx, dirty);
            }
            return 1;
        }
        case VK_F3:        
            m_webView->zoomPageIn();
            return 1;
        case VK_F4:
            m_webView->zoomPageOut();
            return 1;
            
        case VK_F5:
        {
            m_webView->refresh();            
            if (m_context && mainFrame->contentRenderer() && mainFrame->view())
            {
                GraphicsContext ctx(m_context);
                mainFrame->view()->layoutIfNeededRecursive();
                IntRect dirty(0, 0, m_rect.width(), m_rect.height());
                mainFrame->view()->paint(&ctx, dirty);
            }
            return 1;
        }
        case VK_F6:
        {
        	focusFlashplayer();
        	return 0;
        }

        default:
            break;
    }

    return 1;
}

int WebViewPrivate::onKeyUp(BalEventKey event)
{
   PlatformKeyboardEvent keyboardEvent(&event);

   Frame* frame = m_webView->page()->focusController()->focusedOrMainFrame();
    if (!frame)
        return 1;

    EventHandler* eventHandler = frame->eventHandler();
    Node* focusNode = getCacheBuilder().currentFocus();
    if (focusNode)
        eventHandler = focusNode->document()->frame()->eventHandler();

	int keyevent_ret = 0;
    if ((keyevent_ret = eventHandler->keyEvent(keyboardEvent)))
    {
    	//printf("@@@@@@%s:%d %s() return %d.\n", __FILE__, __LINE__, __FUNCTION__, keyevent_ret);
        return keyevent_ret;
    }
        
	return 0;
}

void WebViewPrivate::onMouseMotion(BalEventMotion event)
{
    Frame* frame = core(m_webView->mainFrame());
    if (!frame)
        return;
    frame->eventHandler()->mouseMoved(PlatformMouseEvent(&event));
}

void WebViewPrivate::onMouseButtonDown(BalEventButton event)
{
    Frame* frame = core(m_webView->mainFrame());

    if (!frame)
        return;

    frame->eventHandler()->handleMousePressEvent(PlatformMouseEvent(&event));
}

void WebViewPrivate::onMouseButtonUp(BalEventButton event)
{
    Frame* focusedFrame = m_webView->page()->focusController()->focusedFrame();
    if (!focusedFrame)
        return;
//     if (focusedFrame && focusedFrame->editor()->canEdit()) {
//         ;
//     }

    focusedFrame->eventHandler()->handleMouseReleaseEvent(PlatformMouseEvent(&event));
}

void WebViewPrivate::onScroll(BalEventScroll event)
{
    Frame* frame = core(m_webView->mainFrame());
    if (!frame)
        return;
    PlatformWheelEvent wheelEvent(&event);
    frame->eventHandler()->handleWheelEvent(wheelEvent);
}

void WebViewPrivate::onResize(BalResizeEvent event)
{
    Frame* frame = core(m_webView->mainFrame());
    if (!frame)
        return;
    m_rect.setWidth(event.bounds.w);
    m_rect.setHeight(event.bounds.h);
    setViewWindow(m_webView->viewWindow());

    frame->view()->resize(event.bounds.w, event.bounds.h);
    frame->view()->forceLayout();
    frame->view()->adjustViewSize();
}

void WebViewPrivate::onQuit(BalQuitEvent)
{
    NotImplemented();
}

void WebViewPrivate::onUserEvent(BalUserEvent)
{
    NotImplemented();
}

void WebViewPrivate::setViewWindow(BalWidget* viewWindow)
{
    Fl_Window* win = reinterpret_cast<Fl_Window*>(viewWindow);

#if PLATFORM(DFB_CAIRO)
    if (m_context)
    {
        cairo_destroy(m_context);
        m_context = NULL;
    }
    if (m_cairoSurface)
    {
        cairo_surface_destroy(m_cairoSurface);
        m_cairoSurface = NULL;
    }

    if (win != NULL)
    {
        IDirectFBSurface* surface = reinterpret_cast<IDirectFBSurface*>(win->getClientSurface());

        if (surface)
        {
            m_cairoSurface = cairo_directfb_surface_create(fl_dfb, surface);
            if (!m_cairoSurface)
            {
                printf("failed to create cairo surface\n");
                return;
            }
            m_context = cairo_create(m_cairoSurface);
        }
    }
#else
    if (m_context)
    {
        m_context->Release(m_context);
        m_context = NULL;
    }
    if (win)
    {
        IDirectFBSurface* surface = reinterpret_cast<IDirectFBSurface*>(win->getClientSurface());
        if (surface)
        {
            m_context = surface;
            m_context->AddRef(m_context);
        }
    }
#endif
}

void WebViewPrivate::popupMenuHide()
{
    //BOOKMARKLET_INSTALL_PATH
}


void WebViewPrivate::popupMenuShow(void *popupInfo)
{
    printf("%s: TODO\n", __FILE__);
    NotImplemented();
}

void WebViewPrivate::updateView(BalWidget *surf, IntRect rect)
{
#if 0
    if (!surf || rect.isEmpty())
        return;
    rect.intersect(m_rect);

    DFBRegion dfbRegion = {rect.x(), rect.y(), rect.x() + rect.width() - 1, rect.x() + rect.height() - 1};
    surf->Flip(surf, NULL/*FIXME &dfbRegion*/, DSFLIP_BLIT);
#endif
}

void WebViewPrivate::repaintTimerFired(WebCore::Timer<WebViewPrivate>* timer)
{
    ASSERT(!m_inDrawing);
    ASSERT(!m_repaintRegion.isEmpty());
    sendExposeEvent(m_repaintRegion);
    m_repaintRegion.setX(0);
    m_repaintRegion.setY(0);
    m_repaintRegion.setWidth(0);
    m_repaintRegion.setHeight(0);
}

void WebViewPrivate::sendExposeEvent(IntRect r)
{
    if (m_inDrawing) {
        m_repaintRegion.unite(r);
        if (!m_repaintTimer.isActive())
            m_repaintTimer.startOneShot(0.0);
        return;
    }

    m_webView->addToDirtyRegion(r);
    BalWidget* view = m_webView->viewWindow();
    if (view)
    {
    //printf("%s : {%d, %d, %d, %d}, view = %p, damage = %d\n", __FUNCTION__, r.x(), r.y(), r.width(), r.height(), view, view->damage());
        view->damage(FL_DAMAGE_CHILD, r.x(), r.y(), r.width(), r.height());
    //printf("%s : {%d, %d, %d, %d}, view = %p, damage = %d\n", __FUNCTION__, r.x(), r.y(), r.width(), r.height(), view, view->damage());
    }
}

void WebViewPrivate::repaint(const WebCore::IntRect& windowRect, bool contentChanged, bool immediate, bool repaintContentOnly)
{
   //printf("%s : %d : {%d, %d, %d, %d} %d, %d, %d\n", __FUNCTION__, __LINE__,
   //         windowRect.x(), windowRect.y(), windowRect.width(), windowRect.height(),
   //         contentChanged, immediate, repaintContentOnly
   //         );

    IntRect rect = windowRect;
    rect.intersect(m_rect);
    
    if (!rect.isEmpty() && contentChanged)
        sendExposeEvent(rect);

#if 0
    if (contentChanged) 
    {
        m_webView->addToDirtyRegion(rect);
        /*
        Frame* focusedFrame = m_webView->page()->focusController()->focusedFrame();
        if (focusedFrame)
        {
            Scrollbar* hBar = focusedFrame->view()->horizontalScrollbar();
            Scrollbar* vBar = focusedFrame->view()->verticalScrollbar();
            
            // TODO : caculate the scroll delta and test this.
            //if (dx && hBar)
            if (hBar)
                m_webView->addToDirtyRegion(IntRect(focusedFrame->view()->windowClipRect().x() + hBar->x(), focusedFrame->view()->windowClipRect().y() + hBar->y(), hBar->width(), hBar->height()));
            //if (dy && vBar)
            if (vBar)
                m_webView->addToDirtyRegion(IntRect(focusedFrame->view()->windowClipRect().x() + vBar->x(), focusedFrame->view()->windowClipRect().y() + vBar->y(), vBar->width(), vBar->height()));
        }*/
    }
    if (!repaintContentOnly)
        sendExposeEvent(rect);
    if (immediate)
    {
        if (repaintContentOnly)
            m_webView->updateBackingStore(core(m_webView->topLevelFrame())->view());
        else
            sendExposeEvent(rect);
    }
#endif    
}

void WebViewPrivate::scrollBackingStore(WebCore::FrameView* view, int dx, int dy, const WebCore::IntRect& scrollViewRect, const WebCore::IntRect& clipRect)
{
#if 0
    printf("%s : %d, %d, {%d, %d, %d, %d} {%d, %d, %d, %d}\n",
            __FUNCTION__, 
            dx, dy, 
            scrollViewRect.x(), scrollViewRect.y(), scrollViewRect.width(), scrollViewRect.height(),
            clipRect.x(), clipRect.y(), clipRect.width(), clipRect.height());

    Frame* frame = core(m_webView->mainFrame());
    LiteBox* widget = m_webView->viewWindow();
    IntRect dirtyRect;

    if (!frame || !m_context || !widget)
        return;
    
    do {
        int svWidth = scrollViewRect.width();
        int svHeight = scrollViewRect.height();
        DFBRectangle sr;
        int dstX, dstY;
        int dirtyX, dirtyY, dirtyW, dirtyH;

        if (abs(dx) >= svWidth || abs(dy) >= svHeight || (dx != 0 && dy != 0)) {
            dirtyRect = clipRect;
            dirtyRect.intersect(scrollViewRect);
            break;
        }
        if (dx != 0) {
            sr.y = dstY = dirtyY = 0;
            sr.h = dirtyH = svHeight;
            if (dx < 0) {
                dstX = 0;
                sr.x = dirtyW = -dx;
                sr.w = dirtyX = svWidth + dx;
            } else {
                sr.w = svWidth - dx;
                sr.x = dirtyX = 0;
                dirtyW = dstX = dx;
            }
        } else {
            sr.x = dstX = dirtyX = 0;
            sr.w = dirtyW = svWidth;
            if (dy < 0) {
                dstY = 0;
                sr.y = dirtyH = -dy;
                sr.h = dirtyY = svHeight + dy;
            } else {
                sr.h = svHeight - dy;
                sr.y = dirtyY = 0;
                dirtyH = dstY = dy;
            }
        }

        printf("%s: dirty {%d, %d, %d, %d}, blit {%d, %d, %d, %d} --> {%d, %d}\n", 
                __FUNCTION__, 
                dirtyX, dirtyY, dirtyW, dirtyH,
                sr.x, sr.y, sr.w, sr.h,
                dstX, dstY
              );

        widget->surface->SetSrcBlendFunction(widget->surface, DSBF_ONE);
        widget->surface->SetDstBlendFunction(widget->surface, DSBF_ZERO);
        widget->surface->SetBlittingFlags(widget->surface, DSBLIT_NOFX);
        widget->surface->SetClip(widget->surface, NULL);
        widget->surface->Blit(widget->surface, widget->surface, &sr, dstX, dstY);
        
        dirtyRect = IntRect(dirtyX, dirtyY, dirtyW, dirtyH);
        dirtyRect.intersect(scrollViewRect);
    } while(0);

    if (frame->contentRenderer() && frame->view() && !dirtyRect.isEmpty()) {
        GraphicsContext ctx(m_context);

        frame->view()->layoutIfNeededRecursive();
        frame->view()->paint(&ctx, dirtyRect);
    }
    widget->surface->Flip(widget->surface, NULL, (DFBSurfaceFlipFlags)0);

#else
    IntRect updateRect = clipRect;
    updateRect.intersect(scrollViewRect);

    sendExposeEvent(updateRect);
#endif

}

void WebViewPrivate::runJavaScriptAlert(WebFrame* frame, const char* message)
{
    printf("Javascript Alert: %s (from frame %p)\n", message, frame);
}

static android::CachedFrame::Direction KeyToDirection(int keyCode)
{
    switch (keyCode) {
        case VK_RIGHT:
            DBG_NAV_LOGD("keyCode=%s\n", "right");
            return android::CachedFrame::RIGHT;
        case VK_LEFT:
            DBG_NAV_LOGD("keyCode=%s\n", "left");
            return android::CachedFrame::LEFT;
        case VK_DOWN:
            DBG_NAV_LOGD("keyCode=%s\n", "down");
            return android::CachedFrame::DOWN;
        case VK_UP:
            DBG_NAV_LOGD("keyCode=%s\n", "up");
            return android::CachedFrame::UP;
        default:
            DBG_NAV_LOG("------- bad key sent to WebView::moveFocus\n");
            return android::CachedFrame::UNINITIALIZED;
    }
}

void WebViewPrivate::getVisibleRect(WebCore::IntRect& rect)
{
	rect = m_webView->page()->mainFrame()->view()->visibleContentRect();
}
int WebViewPrivate::getScaledMaxXScroll()
{
	return m_webView->page()->mainFrame()->view()->layoutWidth();
}
int WebViewPrivate::getScaledMaxYScroll()
{
	return m_webView->page()->mainFrame()->view()->layoutHeight();
}

CachedRoot* WebViewPrivate::getFrameCache(FrameCachePermission allowNewer)
{
	return m_cachedRoot;
    /*
    m_viewImpl->gRecomputeFocusMutex.lock();
    bool recomputeInProgress = m_viewImpl->m_recomputeEvents.size() > 0;
    m_viewImpl->gRecomputeFocusMutex.unlock();
    if (allowNewer != AllowNewest && recomputeInProgress)
        return m_frameCacheUI;
    if (allowNewer == DontAllowNewer && m_viewImpl->m_lastGeneration < m_generation)
        return m_frameCacheUI;
    DBG_NAV_LOGD("%s", "m_viewImpl->m_updatedFrameCache == true");
    bool hadFocus = m_frameCacheUI && m_frameCacheUI->currentFocus();
    m_viewImpl->gFrameCacheMutex.lock();
    OutOfFocusFix fix = DoNothing;
    if (allowNewer != DontAllowNewer)
        fix = fixOutOfDateFocus(m_viewImpl->m_useReplay);
    delete m_frameCacheUI;
    delete m_navPictureUI;
    m_viewImpl->m_updatedFrameCache = false;
    m_frameCacheUI = m_viewImpl->m_frameCacheKit;
    m_navPictureUI = m_viewImpl->m_navPictureKit;
    m_viewImpl->m_frameCacheKit = 0;
    m_viewImpl->m_navPictureKit = 0;
    m_viewImpl->gFrameCacheMutex.unlock();
    if (hadFocus && (!m_frameCacheUI || !m_frameCacheUI->currentFocus()))
        viewInvalidate(); // redraw in case focus ring is still visible
    if (fix == UpdateTextEntry)
        updateTextEntry();
    else if (fix == ClearTextEntry)
        clearTextEntry();
    return m_frameCacheUI;
    */
}

static void getFrameScrollOffset(WebCore::Frame* frame, int* x, int* y)
{
    ASSERT(x);
    ASSERT(y);
    *x = 0;
    *y = 0;
    if (!frame || !frame->view())
        return;
    do {
		IntSize offset = frame->view()->scrollOffset();
		*x += offset.width();
		*y += offset.height();
    } while((frame = frame->tree()->parent()) != NULL);    
}

void WebViewPrivate::drawFocusRing(GraphicsContext& ctx, WebCore::IntRect& dirty)
{
    Node* focusNode = getCacheBuilder().currentFocus();
    if (!focusNode) {
        DBG_NAV_LOG("!focus");
        return;
    }
    if (!focusNode->document()) {
        DBG_NAV_LOG("!document");
        return;
    }
    RenderObject* renderer = focusNode->renderer();
    if (!renderer) {
        DBG_NAV_LOG("!renderer");
        return;
    }
    if (RenderStyle* style = renderer->style()) {
        if (style->visibility() != WebCore::VISIBLE) {
            DBG_NAV_LOG("!visible");
            return;
        }
    }

    int offsetX, offsetY;
    getFrameScrollOffset(focusNode->document()->frame(), &offsetX, &offsetY);
    IntRect boundingRect = renderer->absoluteClippedOverflowRect();

#if 0    
#ifndef NDEBUG
    printf("\n\n*********************************\n");
    printf("renderer = %p, node = %p, boundingRect = {%d, %d, %d, %d}, offsetX = %d, offsetY = %d\n", renderer, focusNode, boundingRect.x(), boundingRect.y(), boundingRect.width(), boundingRect.height(), offsetX, offsetY);
    renderer->showTreeForThis();
    printf("\n\n*********************************\n");
#endif
#endif    

    boundingRect.move(-offsetX, -offsetY);
    boundingRect.inflate(-2);    
    if (!boundingRect.intersects(dirty)) {
        DBG_NAV_LOG("!dirtyRect");
        return;
    }

    cairo_t* cr = ctx.platformContext();
    Color ringColor = RenderTheme::focusRingColor();
    float r, g, b, a;
    ringColor.getRGBA(r, g, b, a);

    cairo_save(cr);
    cairo_set_line_width(cr, 2.0f);
    cairo_set_source_rgba(cr, r, g, b, a);
    cairo_set_dash(cr, 0, 0, 0);
    cairo_rectangle(cr, boundingRect.x(), boundingRect.y(), boundingRect.width(), boundingRect.height());
    cairo_stroke(cr);
    cairo_restore(cr);

    return;

#if 0
    const android::CachedRoot* root = getFrameCache(AllowNewer);
    if (!root) {
        DBG_NAV_LOG("!root");
        //m_followedLink = false;
        return;
    }
    const android::CachedFrame* frame;
    const android::CachedNode* node = root->currentFocus(&frame);
    if (!node || !frame) {
        DBG_NAV_LOG("!node || !frame");
        //m_followedLink = false;
        return;
    }
    if (!node->hasFocusRing()) {
        DBG_NAV_LOG("!node->hasFocusRing()");
        return;
    }
    const WTF::Vector<WebCore::IntRect>& rings = node->focusRings();
    unsigned ringsCount = rings.size();
	
    if (!ringsCount) {
        DBG_NAV_LOG("!rings.size()");
        return;
    }

    cairo_t* cr = ctx.platformContext();
    Color ringColor(WebCore::focusRingColor());
    float red, green, blue, alpha;
    ringColor.getRGBA(red, green, blue, alpha);
    int offsetX, offsetY;
    //IntSize offsetIntSize = frame->view()->scrollOffset();
    cairo_save(cr);
    cairo_set_source_rgba(cr, red, green, blue, 0.6);

    getFrameScrollOffset(static_cast<WebCore::Frame*>(frame->framePointer()), &offsetX, &offsetY);
    DBG_NAV_LOGD("dirty = {%d,%d,%d,%d}\n", dirty.x(), dirty.y(), dirty.width(), dirty.height());
    for (unsigned i = 0; i < ringsCount; i++) {
        DBG_NAV_LOGD("rings[%d] = {%d %d %d %d}\n", i, rings[i].x() - offsetX, rings[i].y() - offsetY, rings[i].width(), rings[i].height());  
        cairo_rectangle(cr, rings[i].x() - offsetX, rings[i].y() - offsetY, rings[i].width(), rings[i].height());
    }

    cairo_set_operator(cr, CAIRO_OPERATOR_OVER);
    cairo_fill(cr);
    cairo_restore(cr);
#endif

#if 0
    IntRect focusRect = IntRect(0, 0, 0, 0);
    
    for (unsigned i = 0; i < ringsCount; i++)
        focusRect.unite(rings[i]);
        
    if (!dirty.intersects(focusRect)) {
	
	//DBG_NAV_LOGD("{%d,%d,%d,%d} {%d,%d,%d,%d}\n", dirty.x(), dirty.y(), dirty.width(), dirty.height(),
	//	focusRect.x(), focusRect.y(), focusRect.width(), focusRect.height());
     //   return;
    }
	DBG_NAV_LOGD("{%d,%d,%d,%d} {%d,%d,%d,%d}\n", dirty.x(), dirty.y(), dirty.width(), dirty.height(),
		focusRect.x(), focusRect.y(), focusRect.width(), focusRect.height());
    cairo_t* cr = ctx.platformContext();
    Color ringColor(WebCore::focusRingColor());
	
    cairo_save(cr);
    cairo_push_group(cr);
	
    cairo_new_path(cr);
    cairo_rectangle(cr, focusRect.x(), focusRect.y(), focusRect.width(), focusRect.height());
	
    //int radius = (ctx.focusRingWidth()- 1) / 2;
    //for (unsigned i = 0; i < ringsCount; i++)
//		cairo_rectangle(cr, rings[i].x(), rings[i].y(), rings[i].width(), rings[i].height());
	
       // addPath(Path::createRoundedRectangle(rings[i], FloatSize(radius, radius)));

    // Force the alpha to 50%.  This matches what the Mac does with outline rings.
    cairo_set_source_rgba(cr, ringColor.red(), ringColor.green(), ringColor.blue(), 127);
    cairo_set_line_width(cr,8);

    cairo_set_operator(cr, CAIRO_OPERATOR_OVER);
    cairo_stroke_preserve(cr);

    cairo_set_operator(cr, CAIRO_OPERATOR_CLEAR);
    cairo_set_fill_rule(cr, CAIRO_FILL_RULE_WINDING);
    cairo_fill(cr);

    cairo_pop_group_to_source(cr);
    cairo_set_operator(cr, CAIRO_OPERATOR_OVER);
    cairo_paint(cr);
    cairo_restore(cr);
#endif	
	
}

bool WebViewPrivate::focusIsTextArea(FrameCachePermission allowNewer)
{
    CachedRoot* root = getFrameCache(allowNewer);
    if (!root) {
        DBG_NAV_LOG("!root");
        return false;
    }
    const CachedNode* focus = root->currentFocus();
    if (!focus)
        return false;
    return focus->isTextArea() || focus->isTextField();
}
	
bool WebViewPrivate::finalKitFocus(WebCore::Frame* frame, WebCore::Node* node,
    int x, int y, bool donotChangeDOMFocus)
{
    CacheBuilder& builder = getCacheBuilder();
    if (!frame || builder.validNode(frame, NULL) == false)
        frame = getCoreMainFrame();	
    WebCore::Node* oldFocusNode = builder.currentFocus();
    // mouse event expects the position in the window coordinate
    IntSize offsetIntSize = m_webView->page()->mainFrame()->view()->scrollOffset();
	x -= offsetIntSize.width();
	y -= offsetIntSize.height();
    m_mousePos = WebCore::IntPoint(x, y);
    // validNode will still return true if the node is null, as long as we have
    // a valid frame.  Do not want to make a call on frame unless it is valid.
    WebCore::PlatformMouseEvent mouseEvent(m_mousePos, m_mousePos,
        WebCore::NoButton, WebCore::MouseEventMoved, 1, false, false, false,
        false, WTF::currentTime());
	DBG_NAV_LOGD("sendMouseMoveEvent {%d, %d}\n", x, y);					
	
    frame->eventHandler()->mouseMoved(mouseEvent);
    bool valid = builder.validNode(frame, node);
    if (!donotChangeDOMFocus) {
        WebCore::Document* oldDoc = oldFocusNode ? oldFocusNode->document() : 0;
        if (!node) {
            if (oldFocusNode)
                oldDoc->setFocusedNode(0);
            return false;
        } else if (!valid) {
            DBG_NAV_LOGD("sendMarkNodeInvalid node=%p", node);
            //sendMarkNodeInvalid(node);
            if (oldFocusNode)
                oldDoc->setFocusedNode(0);
            return false;
        }
        // If we jump frames (docs), kill the focus on the old doc
        if (oldFocusNode && node->document() != oldDoc) {
            oldDoc->setFocusedNode(0);
        }
        if (!node->isTextNode())
            static_cast<WebCore::Element*>(node)->focus(false);
        if (node->document()->focusedNode() != node) {
            // This happens when Element::focus() fails as we may try to set the
            // focus to a node which WebCore doesn't recognize as a focusable node.
            // So we need to do some extra work, as it does in Element::focus(),
            // besides calling Document::setFocusedNode.
            if (oldFocusNode) {
                // copied from clearSelectionIfNeeded in FocusController.cpp
                WebCore::SelectionController* s = oldDoc->frame()->selection();
                if (!s->isNone())
                    s->clear();
            }
            //setFocus on things that WebCore doesn't recognize as supporting focus
            //for instance, if there is an onclick element that does not support focus
            node->document()->setFocusedNode(node);
        }
    } else {   // !donotChangeDOMFocus
        if (!node || !valid)
            return false;
    }
    
    builder.setLastFocus(node);
    m_lastFocused = node;
    m_lastFocusedBounds = node->getRect();
    DBG_NAV_LOGD("setFocusedNode node=%p {%d, %d, %d, %d}\n", m_lastFocused, 
					m_lastFocusedBounds.x(), m_lastFocusedBounds.y(),
					m_lastFocusedBounds.width(), m_lastFocusedBounds.height());
    return true;
}

bool WebViewPrivate::moveFocus(int keyCode, int count, bool ignoreScroll, bool inval,
            void* lastSentFocus, const WebCore::IntRect* lastSentBounds)
{	
    if (keyCode != VK_RIGHT && keyCode != VK_LEFT && keyCode != VK_DOWN && keyCode != VK_UP) {
        return true;
    }
    CachedRoot* root = getFrameCache(AllowNewer);

    if (!root) {
        return true;
    }

    fixupFocus(root, &getCacheBuilder(), true);

    const android::CachedFrame* cachedFrame, * oldFrame = 0;
    const CachedNode* focus = root->currentFocus(&oldFrame);
    WebCore::IntPoint focusLocation = root->focusLocation();

    DBG_NAV_LOGD("old focus %d (nativeNode=%p) focusLocation={%d, %d}",
        focus ? focus->index() : 0,
        focus ? focus->nodePointer() : 0, focusLocation.x(), focusLocation.y());		

    android::CachedFrame::Direction direction = KeyToDirection(keyCode);	
    WebCore::IntRect visibleRect;
    getVisibleRect(visibleRect);
    DBG_NAV_LOGD("getVisibleRect %d,%d,%d,%d",
        visibleRect.x(), visibleRect.y(), visibleRect.width(), visibleRect.height());
    root->setVisibleRect(visibleRect);
    int xMax = visibleRect.width() / 2;         //getScaledMaxXScroll();
    int yMax = visibleRect.height() / 2;        //getScaledMaxYScroll();
    root->setMaxScroll(xMax, yMax);
    CachedHistory savedHistory = *root->rootHistory();
    bool oldNodeIsTextArea = focusIsTextArea(DontAllowNewer);
    const CachedNode* cachedNode = 0;
    int dx = 0;
    int dy = 0;
    int counter = count;
    //if (!focus || !focus->isInput() || !m_followedLink)
    //    root->setScrollOnly(m_followedLink);
    while (--counter >= 0) {
        WebCore::IntPoint scroll = WebCore::IntPoint(0, 0);
        cachedNode = root->moveFocus(direction, &cachedFrame, &scroll);
        dx += scroll.x();
        dy += scroll.y();
    }

#if 0
    const android::CachedFrame* scrolledFrame = cachedFrame ? cachedFrame : oldFrame;
    if ((dx || dy) && scrolledFrame)
    {    
          WebCore::Frame* frame = static_cast<WebCore::Frame*>(scrolledFrame->framePointer());
          while(frame) {
	   	IntSize scrolledSize(0, 0);

	   	if (dx && !frame->eventHandler()->scrollOverflow(dx > 0 ? ScrollRight : ScrollLeft, ScrollByLine))
			scrolledSize.setWidth(dx);
		if (dy && !frame->eventHandler()->scrollOverflow(dy > 0 ? ScrollDown : ScrollUp, ScrollByLine))
			scrolledSize.setHeight(dy);
		if (!scrolledSize.isEmpty()) {
			frame->view()->scrollBy(scrolledSize);
			break;
		}
		frame = frame->tree()->parent();
          }
#else
    if (dx || dy) {
        IntSize scrollRect;
        WebCore::Frame* frame = getCoreMainFrame();

        if (dx && !frame->eventHandler()->scrollOverflow(dx > 0 ? ScrollRight : ScrollLeft, ScrollByLine))
            scrollRect.setWidth(dx);        
        if (dy && !frame->eventHandler()->scrollOverflow(dy > 0 ? ScrollDown : ScrollUp, ScrollByLine))
            scrollRect.setHeight(dy);

        if (scrollRect.width() || scrollRect.height()) {
    	   frame->view()->scrollBy(scrollRect);
	       if (!ignoreScroll && !cachedNode)
	          return moveFocus(keyCode, count, true, inval, lastSentFocus, lastSentBounds);
        }
    }
#endif	

    DBG_NAV_LOGD("new focus %d (nativeNode=%p) focusLocation={%d, %d}",
        cachedNode ? cachedNode->index() : 0,
        cachedNode ? cachedNode->nodePointer() : 0, root->focusLocation().x(),
            root->focusLocation().y());
    // If !m_heightCanMeasure (such as in the browser), we want to scroll no
    // matter what
    bool result = false;
    if (cachedNode) {
        WebCore::IntPoint pos;
        root->setCachedFocus((android::CachedFrame*) cachedFrame, (CachedNode*) cachedNode);
        root->getSimulatedMousePosition(&pos);
	
        if (lastSentFocus == cachedNode->nodePointer() && lastSentBounds &&
                *lastSentBounds == cachedNode->bounds()) {
        	finalKitFocus((WebCore::Frame*)cachedFrame->framePointer(), (WebCore::Node*)cachedNode->nodePointer(), pos.x(), pos.y(), false);
            //sendFinalFocus((WebCore::Frame*) cachedFrame->framePointer(),
             //   (WebCore::Node*) cachedNode->nodePointer(), pos.x(), pos.y());
        } else {
        	finalKitFocus((WebCore::Frame*)cachedFrame->framePointer(), (WebCore::Node*)cachedNode->nodePointer(), pos.x(), pos.y(), false);
            /*setFocusData(root->generation(),
                (WebCore::Frame*) cachedFrame->framePointer(),
                (WebCore::Node*) cachedNode->nodePointer(), pos.x(), pos.y(),
                true);
            sendKitFocus();
            if (inval)
                viewInvalidate();
            MoveFocusParams params;
            params.d.d.m_trigger = CommonParams::MoveFocusParams;
            params.d.d.m_generation = m_generation;
            params.c.setFocus(focus, oldFrame, root, focusLocation);
            params.m_sentFocus = cachedNode->nodePointer();
            params.m_sentBounds = cachedNode->bounds();
            params.m_visibleRect = visibleRect;
            params.m_history = savedHistory;
            DBG_NAV_LOGD("history.mDidFirstLayout=%s",
                    params.m_history.didFirstLayout() ? "true" : "false");
            params.m_xMax = xMax;
            params.m_yMax = yMax;
            params.d.m_keyCode = keyCode;
            params.d.m_count = count;
            params.d.m_ignoreScroll = ignoreScroll;
            m_replay.add(params.d.d, sizeof(params));
            */
        }
    } else {
        if (visibleRect.intersects(root->focusBounds()) == false) {
            //setFocusData(root->generation(), 0, 0, 0, 0, true);
            //sendKitFocus(); // will build cache and retry
        }
		/*
        FirstMoveFocusParams params;
        params.d.m_trigger = CommonParams::FirstMoveFocusParams;
        params.d.m_generation = m_generation;
        params.m_keyCode = keyCode;
        params.m_count = count;
        params.m_ignoreScroll = ignoreScroll;
        m_replay.add(params.d, sizeof(params));
		
        int docHeight = root->documentHeight();
        int docWidth = root->documentWidth();
        if (visibleRect.bottom() + dy > docHeight)
            dy = docHeight - visibleRect.bottom();
        else if (visibleRect.y() + dy < 0)
            dy = -visibleRect.y();
        if (visibleRect.right() + dx > docWidth)
            dx = docWidth - visibleRect.right();
        else if (visibleRect.x() < 0)
            dx = -visibleRect.x();		
        result = direction == android::CachedFrame::LEFT ? dx >= 0 :
            direction == android::CachedFrame::RIGHT ? dx <= 0 :
            direction == android::CachedFrame::UP ? dy >= 0 : dy <= 0;
            */
            result = true;
    }
	/*
    if (focusIsTextArea(DontAllowNewer))
        updateTextEntry();
    else if (oldNodeIsTextArea)
        clearTextEntry();
        */
    return result;
}

 void WebViewPrivate::didFirstLayout()
 {
 	m_check_domtree_version = false;
 	updateFrameCache();
 }
 
 void WebViewPrivate::progressFinishedNotification()
 {
 	m_check_domtree_version = true;
 	updateFrameCache();
    if (!focusedNode())
        m_webView->setInitialFocus(true); 	
 }
 
void WebViewPrivate::fixupFocus(android::CachedRoot* root, android::CacheBuilder* builder, bool followDOM)
{
    const android::CachedFrame* focusFrame = 0;
    const CachedNode* focus = root->currentFocus(&focusFrame);
    WebCore::IntPoint focusLocation = root->focusLocation();
    WebCore::Node* domFocus = builder->currentFocus();

    DBG_NAV_LOGD("focus = %p, domFocus = %p, nodePointer = %p\n", focus, domFocus, focus ? focus->nodePointer() : 0);

    if (focus && domFocus && focus->nodePointer() != domFocus) {
        DBG_NAV_LOGD("focus out of date{cachedFocus = %p, dom focus = %p}\n", focus->nodePointer(), domFocus);
        if (followDOM)
            root->setFocus(domFocus);
        else
            finalKitFocus((WebCore::Frame*)focusFrame->framePointer(), 
                    (WebCore::Node*)focus->nodePointer(), focusLocation.x(), focusLocation.y(), false);
    }
}

void WebViewPrivate::updateFrameCache()
{
    //printf("%s : old cached root = %p, m_lastFocused = %p\n", __FUNCTION__, m_cachedRoot, m_lastFocused);
    //if (m_lastFocused)
    //    printf(" focusBounds = {%d %d %d %d}\n", m_lastFocusedBounds.x(), m_lastFocusedBounds.y(),
    //            m_lastFocusedBounds.width(), m_lastFocusedBounds.height());
    if (!mEnableMouseSimulate)
        return;

	if (m_cachedRoot) {
		delete m_cachedRoot;
		m_cachedRoot = NULL;
	}
	
    CacheBuilder* builder = &getCacheBuilder();
	
    m_cachedRoot = new CachedRoot();
    m_cachedRoot ->init(getCoreMainFrame(), &m_cachedHistory);
    builder->buildCache(m_cachedRoot);

    if (m_lastFocused && builder->validNode(getCoreMainFrame(), m_lastFocused)) {
        m_cachedRoot->setFocus(m_lastFocused);
    }
    fixupFocus(m_cachedRoot, builder, true);
#if 0
{
        WebCore::IntRect visibleRect;
        const android::CachedFrame* foundFrame = 0;
        int offsetX, offsetY;
        int x, y;
        WebCore::IntRect nodeRect;
        const android::CachedNode* found;


        getVisibleRect(visibleRect);
        m_cachedRoot->setVisibleRect(visibleRect);
        android::CacheBuilder::GetGlobalOffset(m_lastFocused, &offsetX, &offsetY);
        nodeRect = m_lastFocused->getRect();
        nodeRect.move(offsetX, offsetY);
        found = m_cachedRoot->findAt(nodeRect, &foundFrame, &x, &y);
        m_cachedRoot->setCachedFocus(const_cast<android::CachedFrame*>(foundFrame), const_cast<android::CachedNode*>(found));
        if (found) {
            WebCore::Frame* framePointer = foundFrame ? (WebCore::Frame*) foundFrame->framePointer() : 0;
            WebCore::Node* nodePointer = (WebCore::Node*) found->nodePointer();
            finalKitFocus(framePointer, nodePointer, x, y, false);
        }
}
#endif    
}

void WebViewPrivate::focusFlashplayer()
{
	Frame* mainFrame = getCoreMainFrame();
	Element* fp = NULL;

	#if 1	// get node by node's TagName.
	
	RefPtr<NodeList> node_list = mainFrame->document()->getElementsByTagName("embed");
	printf("get embed num = %d\n",node_list->length());

	if(node_list->length() == 0){
		return;
	}

	Node* node = node_list->item(0);
	// TODO: get the biggest item of node_list. I means that the value of width*height should be the biggest.
	fp = static_cast<WebCore::Element*>(node);

	#else	// get node by node's id.

	fp = mainFrame->document()->getElementById("main.player.video");
	
	#endif
	
	if(fp == NULL){
		printf("get flashplayer div fail!\n");
		return;
	}

	int area =  fp->offsetWidth() * fp->offsetHeight();
	if(area < 100){
		printf("focus flashplayer fail for its size is too small.\n");
		return;
	}

	moveMouseOnElement(fp);
}

void WebViewPrivate::moveMouseOnElement(Element* element)
{
	if(element == NULL){
		printf("move Mouse on Element fail for elment == NULL!\n");
		return;
	}

	Frame* mainFrame = getCoreMainFrame();

	IntSize offsetIntSize = mainFrame->view()->scrollOffset();
	int real_x = element->offsetLeft() - offsetIntSize.width();
	int real_y = element->offsetTop() - offsetIntSize.height();
	Element* e = element;
	while((e = e->offsetParent()) != NULL){
		real_x += e->offsetLeft();
		real_y +=  e->offsetTop();
	}

	//test
	//printf("target node : left=%d, top=%d, w=%d, h=%d\n", real_x, real_y, element->offsetWidth(), element->offsetHeight());
	
	int x = 0;
	int y = 0;
	WebCore::IntPoint pos(x, y);
	WebCore::PlatformMouseEvent mouseEvent(pos, pos,
		WebCore::NoButton, WebCore::MouseEventMoved, 1, false, false, false,
		false, WTF::currentTime());
	mainFrame->eventHandler()->handleMouseMoveEvent(mouseEvent);

	x = real_x + element->offsetWidth()/2;
	y = real_y + element->offsetHeight()/2;
	WebCore::IntPoint pos2(x, y);
	WebCore::PlatformMouseEvent mouseEvent2(pos2, pos2,
		WebCore::NoButton, WebCore::MouseEventMoved, 1, false, false, false,
		false, WTF::currentTime());
	mainFrame->eventHandler()->handleMouseMoveEvent(mouseEvent2);

	printf("moveMouseOn, left = %d, top = %d. \n", x, y);
}
