/*
 * Copyright (C) 2008 Pleyo.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Pleyo nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY PLEYO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL PLEYO OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef WebViewPrivate_H
#define WebViewPrivate_H

#include "WebCachedRoot.h"
#include "WebCachedHistory.h"
#include "WebCachedNode.h"
#include "WebCacheBuilder.h"
#include "WebFrameLoaderClient.h"

#include "WebView.h"
#include "IntRect.h"
#include "FrameView.h"
#include "Frame.h"
#include "BALBase.h"
#include "NotImplemented.h"
#include "Timer.h"

#if PLATFORM(DFB_CAIRO)
#include <cairo.h>
#endif

using namespace android;

enum FrameCachePermission {
    DontAllowNewer,
    AllowNewer,
    AllowNewest
};

enum OutOfFocusFix {
    DoNothing,
    ClearTextEntry,
    UpdateTextEntry
};

class WebViewPrivate {
public:
    WebViewPrivate(WebView *webView);
    ~WebViewPrivate()
    {
        //NotImplemented();
    }

    void show()
    {
        //NotImplemented();
    }
    
    void hide()
    {
        //NotImplemented();
    }

    void setFrameRect(WebCore::IntRect r)
    {
        m_rect = r;
    }

    BalRectangle frameRect()
    {
        BalRectangle r = {m_rect.x(), m_rect.y(), m_rect.width(), m_rect.height()};
        return r;
    }
    
    BalWidget *createWindow(BalRectangle r)
    {
        WebCore::IntRect rect(r);
        if(rect != m_rect)
            m_rect = rect;
        return 0;
    }

    void initWithFrameView(WebCore::FrameView *frameView)
    {
        //NotImplemented();
    }

    void clearDirtyRegion()
    {
        m_backingStoreDirtyRegion.setX(0);
        m_backingStoreDirtyRegion.setY(0);
        m_backingStoreDirtyRegion.setWidth(0);
        m_backingStoreDirtyRegion.setHeight(0);
    }

    BalRectangle dirtyRegion()
    {
        BalRectangle r = {m_backingStoreDirtyRegion.x(), 
            m_backingStoreDirtyRegion.y(),
            m_backingStoreDirtyRegion.width(),
            m_backingStoreDirtyRegion.height()
        };
    
        return r;
    }

    void addToDirtyRegion(const BalRectangle& dirtyRect)
    {
        m_backingStoreDirtyRegion.unite(dirtyRect);
    }

    void onExpose(BalEventExpose event);
    int  onKeyDown(BalEventKey event);
    int  onKeyUp(BalEventKey event);
    void onMouseMotion(BalEventMotion event);
    void onMouseButtonDown(BalEventButton event);
    void onMouseButtonUp(BalEventButton event);
    void onScroll(BalEventScroll event);
    void onResize(BalResizeEvent event);
    void onQuit(BalQuitEvent);
    void onUserEvent(BalUserEvent);
    void popupMenuHide();
    void popupMenuShow(void *popupInfo);
    void setViewWindow(BalWidget* viewWindow);

    void sendExposeEvent(WebCore::IntRect);
    
    void repaint(const WebCore::IntRect&, bool contentChanged, bool immediate = false, bool repaintContentOnly = false);
    
    void scrollBackingStore(WebCore::FrameView*, int dx, int dy, const WebCore::IntRect& scrollViewRect, const WebCore::IntRect& clipRect);

    void fireWebKitEvents();
    void runJavaScriptAlert(WebFrame* frame, const char* message);

    void setCharInput(const char* input);

    void focusFlashplayer();
    void moveMouseOnElement(Element* element);
    
private:
    void updateView(BalWidget* widget, WebCore::IntRect rect);
    WebCore::IntRect m_rect;
    WebView* m_webView;
    bool isInitialized;
    WebCore::IntRect m_scrollUpdateRect;
    WebCore::IntRect m_hUpdateRect;
    WebCore::IntRect m_vUpdateRect;

    WebCore::IntPoint m_backingStoreSize;
    WebCore::IntRect m_backingStoreDirtyRegion;

#if PLATFORM(DFB_CAIRO)
    cairo_surface_t* m_cairoSurface;
    cairo_t*        m_context;
#else
    IDirectFBSurface* m_context;
#endif

private: //FIX bug which FLTK can not call damage when in draw function
    bool m_inDrawing;
    WebCore::IntRect m_repaintRegion;
    WebCore::Timer<WebViewPrivate> m_repaintTimer;
    void repaintTimerFired(WebCore::Timer<WebViewPrivate>*);

//////// nav implement  //FIXME  should move it into WebView
private:
    bool mEnableMouseSimulate;

public:
    bool enableMouseSimulate() { return mEnableMouseSimulate; }

    CachedRoot* m_cachedRoot;
    CachedHistory m_cachedHistory;

	WebCore::Node* m_lastFocused;
    WebCore::IntRect m_lastFocusedBounds;
	unsigned m_domtree_version;
	bool m_check_domtree_version;
	int m_maxXScroll;
    int m_maxYScroll;
    int m_scrollOffsetX; // webview.java's current scroll in X
    int m_scrollOffsetY; // webview.java's current scroll in Y
    WebCore::IntPoint m_mousePos;

    void progressFinishedNotification();
    void didFirstLayout();
 
    void updateFrameCache();
    void fixupFocus(android::CachedRoot*, android::CacheBuilder*, bool);
    CachedRoot* getFrameCache(FrameCachePermission allowNewer);

    bool focusIsTextArea(FrameCachePermission allowNewer);

    void drawFocusRing(GraphicsContext& ctx, WebCore::IntRect& dirty);
		
    bool finalKitFocus(WebCore::Frame* frame, WebCore::Node* node,
        int x, int y, bool donotChangeDOMFocus);
    
    bool moveFocus(int keyCode, int count, bool ignoreScroll, bool inval,
        void* lastSentFocus, const WebCore::IntRect* lastSentBounds);

    void getVisibleRect(WebCore::IntRect& rect);
	int getScaledMaxXScroll();
    int getScaledMaxYScroll();

	WebCore::Frame* getCoreMainFrame()
	{
		WebFrame* mainWebFrame = m_webView->mainFrame();
		return core(mainWebFrame);
	}
	WebFrameLoaderClient* getLoader()
	{
		WebFrame* mainWebFrame = m_webView->mainFrame();
		WebCore::Frame* mainFrame = core(mainWebFrame);
		WebCore::FrameLoader* loader = mainFrame->loader();
		return (WebFrameLoaderClient*)(loader->client());				
	}
	CacheBuilder& getCacheBuilder()
	{
		return getLoader()->getCacheBuilder();
	}

    bool displaySoftKeyboard();

    WebCore::Node* focusedNode()
    {
        return getCacheBuilder().currentFocus();
    }
};

#endif // WebViewPrivate_H

