
#include <stdio.h>
#include <string.h>

#include <WebKit.h>
#include <directfb.h>

#include "Fl_SoftKeyboard_Youtube_JP.h"

#include "Fl_ImageCache.h"

#include "MainApplication.h"
#include "MainWebViewDelegate.h"
#include "Fl_WebViewWindow.h"

#include <FL/fl_draw.H>
#include <FL/x.H>

extern "C"
{
#include <HiraToKanji.h>
}

#define KB_CHARACTER		0
#define KB_SWITCH 			1
#define KB_COMP				8
#define KB_KANJI_CONFIRM	15
#define KB_CLEAR			22
#define KB_CONFIRM			23
#define KB_BACKSPACE		36
#define KB_RETURN			37


/*
   key num is 44.
   key map =>
        left key           main key             right key

          0
          1,	           2, 3, 4, 5, 6, 7,		     8,
                        9,10,11,12,13,14,
          15,         16,17,18,19,20,21,       22,
          23,         24,25,26,27,28,29,
                       30,31,32,33,34,35,       36,
          37,         38,39,40,41,42,43
   */


/*
	Hiragana ( 3040 - 309f) 
	3040 ?                
	3050                 
	3060                 
	3070                 
	3080                 
	3090     ? ? ? ? ? ? ? a b f g ? 
 */
const int Fl_SoftKeyboard_Youtube_JP::Hiragana1Map[Fl_SoftKeyboard_Youtube_JP::SCANCODES] = {
	FL_CHARACTERSET_KEY,
	FL_SWITCH_KEY,		0x3042, 0x3044, 0x3046, 0x3048, 0x304a, FL_DAKUTEN,  	FL_COMP_KEY,
						0x3041, 0x3043, 0x3045, 0x3047, 0x3049, FL_HANDAKUTEN, 
	FL_KANJI_SWITCH,		0x304b, 0x304d, 0x304f, 0x3051, 0x3053, 0x30fc, 			FL_AllCLEAR_KEY,
	FL_CONFIRM,			0x3055, 0x3057, 0x3059, 0x305b, 0x305d, 0x3000, 
						0x305f, 0x3061, 0x3064, 0x3066, 0x3068, 0x3063, 			FL_BackSpace,
	FL_RETURN_KEY,		0x30FB, 0xff0a, 0xff01, 0xff1f, FL_Left, FL_Right
};

const int Fl_SoftKeyboard_Youtube_JP::Hiragana2Map[Fl_SoftKeyboard_Youtube_JP::SCANCODES] = {
	FL_CHARACTERSET_KEY,
	FL_SWITCH_KEY,		0x306a, 0x306b, 0x306c, 0x306d, 0x306e, FL_DAKUTEN,  	FL_COMP_KEY,
						0x306f, 0x3072, 0x3075, 0x3078, 0x307b, FL_HANDAKUTEN, 
	FL_KANJI_SWITCH,		0x307e, 0x307f, 0x3080, 0x3081, 0x3082, 0x30fc, 			FL_AllCLEAR_KEY,
	FL_CONFIRM,			0x3084, 0x3086, 0x3088, 0x3083, 0x3085, 0x3087, 
						0x3089, 0x308a, 0x308b, 0x308c, 0x308d, 0x3000, 			FL_BackSpace,
	FL_RETURN_KEY,		0x308f, 0x3092, 0x3093, 0x308e, FL_Left, FL_Right
};

/*
	Katakana ( 30a0 - 30ff) 
	30a0 94                
	30b0                 
	30c0                 
	30d0                 
	30e0                 
	30f0        95 96 97 98 99 ` c d 90 
*/
const int Fl_SoftKeyboard_Youtube_JP::Katakana1Map[Fl_SoftKeyboard_Youtube_JP::SCANCODES] = {
	FL_CHARACTERSET_KEY,
	FL_SWITCH_KEY,		0x30a2, 0x30a4, 0x30a6, 0x30a8, 0x30aa, FL_DAKUTEN,  	FL_COMP_KEY,
						0x30a1, 0x30a3, 0x30a5, 0x30a7, 0x30a9, FL_HANDAKUTEN, 
	0,					0x30ab, 0x30ad, 0x30af, 0x30b1, 0x30b3, 0x30fc, 			FL_AllCLEAR_KEY,
	0,					0x30b5, 0x30b7, 0x30b9, 0x30bb, 0x30bd, 0x3000,
						0x30bf, 0x30c1, 0x30c4, 0x30c6, 0x30c8, 0x30c3,			FL_BackSpace,
	FL_RETURN_KEY,		0x30FB, 0xff0a, 0xff01, 0xff1f, FL_Left, FL_Right
};

const int Fl_SoftKeyboard_Youtube_JP::Katakana2Map[Fl_SoftKeyboard_Youtube_JP::SCANCODES] = {
	FL_CHARACTERSET_KEY,
	FL_SWITCH_KEY,		0x30ca, 0x30cb, 0x30cc, 0x30cd, 0x30ce, FL_DAKUTEN,  	FL_COMP_KEY,
						0x30cf, 0x30d2, 0x30d5, 0x30d8, 0x30db, FL_HANDAKUTEN, 
	0,					0x30de, 0x30df, 0x30e0, 0x30e1, 0x30e2, 0x30fc, 			FL_AllCLEAR_KEY,
	0,					0x30e4, 0x30e6, 0x30e8, 0x30e3, 0x30e5, 0x30e7,
						0x30e9, 0x30ea, 0x30eb, 0x30ec, 0x30ed, 0x3000,			FL_BackSpace,
	FL_RETURN_KEY,		0x30ef, 0x30f2, 0x30f3, 0x30ee, FL_Left, FL_Right
};

const int Fl_SoftKeyboard_Youtube_JP::NormalMap[Fl_SoftKeyboard_Youtube_JP::SCANCODES] = {
	FL_CHARACTERSET_KEY,
	0,					'a', 'b', 'c', 'd', 'e', 'f',  		FL_COMP_KEY,
						'g', 'h', 'i', 'j', 'k', 'l', 
	0,					'm', 'n', 'o', 'p', 'q', 'r', 		FL_AllCLEAR_KEY,
	0,					's', 't', 'u', 'v', 'w', 'x', 
						'y', 'z', '.', '/', ':', '@', 		FL_BackSpace,
	FL_RETURN_KEY,	 	'-', '_', ',', ' ', FL_Left, FL_Right
};

const int Fl_SoftKeyboard_Youtube_JP::ShiftMap[Fl_SoftKeyboard_Youtube_JP::SCANCODES] = {
	FL_CHARACTERSET_KEY,	
	0,					'A', 'B', 'C', 'D', 'E', 'F',  		FL_COMP_KEY,
						'G', 'H', 'I', 'J', 'K', 'L', 
	0,					'M', 'N', 'O', 'P', 'Q', 'R', 		FL_AllCLEAR_KEY,
	0,					'S', 'T', 'U', 'V', 'W', 'X', 
						'Y', 'Z', '.', '/', ':', '@', 		FL_BackSpace,
	FL_RETURN_KEY,		'-', '_', ',', ' ', FL_Left, FL_Right
};

const int Fl_SoftKeyboard_Youtube_JP::NumberMap[Fl_SoftKeyboard_Youtube_JP::SCANCODES] = {
	FL_CHARACTERSET_KEY,
	0,					'1', '2', '3', '4', '5', '6',  		FL_COMP_KEY,
						'7', '8', '9', '0', '.', '/', 
	0,					':', '@', '-', '_', ',', ' ', 		FL_AllCLEAR_KEY,
	0,					FL_Left, FL_Right, 0, 0, 0, 0, 
						0, 0, 0, 0, 0, 0,			FL_BackSpace,
	FL_RETURN_KEY, 		0, 0, 0, 0, 0, 0
};

const int Fl_SoftKeyboard_Youtube_JP::SymbolMap[Fl_SoftKeyboard_Youtube_JP::SCANCODES] = {
	FL_CHARACTERSET_KEY,
	0,					'!' , '"', '#', '$', '%', '&',  		FL_COMP_KEY,
						'\'', '*', '+', ',', '-', '.', 
	0,					'/' , ':', ';', '<', '=', '>', 		FL_AllCLEAR_KEY,
	0,					'?' , '@', '\\', '^', '_', '`', 
						'|', '~', '(', ')', '[', ']', 		FL_BackSpace,
	FL_RETURN_KEY,		'{', '}', FL_Left, FL_Right, 0, 0
};

const Fl_SoftKeyboard_Youtube_JP::LayoutRect Fl_SoftKeyboard_Youtube_JP::KeyLayoutRect[Fl_SoftKeyboard_Youtube_JP::SCANCODES] = {
	{73, 76, 400, 30},	
	{22, 122, 103, 47},	{131, 122, 50, 30}, {182, 122, 50, 30}, {233, 122, 50, 30}, {284, 122, 50, 30}, {335, 122, 50, 30}, {386, 122, 50, 30},	{442, 122, 82, 61},
						{131, 153, 50, 30}, {182, 153, 50, 30}, {233, 153, 50, 30}, {284, 153, 50, 30}, {335, 153, 50, 30}, {386, 153, 50, 30}, 
	{22, 170, 103, 45},	{131, 184, 50, 30}, {182, 184, 50, 30}, {233, 184, 50, 30}, {284, 184, 50, 30}, {335, 184, 50, 30}, {386, 184, 50, 30},	{442, 184, 82, 61},
	{22, 216, 103, 45}, 	{131, 215, 50, 30}, {182, 215, 50, 30}, {233, 215, 50, 30}, {284, 215, 50, 30}, {335, 215, 50, 30}, {386, 215, 50, 30}, 
						{131, 246, 50, 30}, {182, 246, 50, 30}, {233, 246, 50, 30}, {284, 246, 50, 30}, {335, 246, 50, 30}, {386, 246, 50, 30},	{442, 246, 82, 61},
	{22, 262, 103, 45},	{131, 277, 50, 30}, {182, 277, 50, 30}, {233, 277, 50, 30}, {284, 277, 50, 30}, {335, 277, 50, 30}, {386, 277, 50, 30}
};

const Fl_SoftKeyboard_Youtube_JP::LayoutState Fl_SoftKeyboard_Youtube_JP::LayoutStates[8] = {
	{"im-jp/Hiragana1.png", Hiragana1Map},
	{"im-jp/Katakana1.png", Katakana1Map},
	{"im-jp/panel_lower.png", NormalMap},
	{"im-jp/panel_upper.png",  ShiftMap},
	{"im-jp/panel_num.png", NumberMap},
	{"im-jp/panel_kigou.png", SymbolMap},

	// for switch.
	{"im-jp/Hiragana2.png", Hiragana2Map},
	{"im-jp/Katakana2.png", Katakana2Map},
};

const Fl_SoftKeyboard_Youtube_JP::DakutenTable Fl_SoftKeyboard_Youtube_JP::DakutenTransTable[Fl_SoftKeyboard_Youtube_JP::DAKUTEN_COUNT] = 
{
	// hira
	{0x304b, 0x304c, 0}, {0x304d, 0x304e, 0}, {0x304f, 0x3050, 0}, {0x3051, 0x3052, 0},
	{0x3053, 0x3054, 0}, {0x3055, 0x3056, 0}, {0x3057, 0x3058, 0}, {0x3059, 0x305a, 0},
	{0x305b, 0x305c, 0}, {0x305d, 0x305e, 0}, {0x305f, 0x3060, 0}, {0x3061, 0x3062, 0},
	{0x3064, 0x3065, 0}, {0x3066, 0x3067, 0}, {0x3068, 0x3069, 0}, {0x306f, 0x3070, 0x3071},
	{0x3072, 0x3073, 0x3074}, {0x3075, 0x3076, 0x3077}, {0x3078, 0x3079, 0x307a}, {0x307b, 0x307c, 0x307d},

	// kata
	{0x30ab, 0x30ac, 0}, {0x30ad, 0x30ae, 0}, {0x30af, 0x30b0, 0}, {0x30b1, 0x30b2, 0},
	{0x30b3, 0x30b4, 0}, {0x30b5, 0x30b6, 0}, {0x30b7, 0x30b8, 0}, {0x30b9, 0x30ba, 0},
	{0x30bb, 0x30bc, 0}, {0x30bd, 0x30be, 0}, {0x30bf, 0x30c0, 0}, {0x30c1, 0x30c2, 0},
	{0x30c4, 0x30c5, 0}, {0x30c6, 0x30c7, 0}, {0x30c8, 0x30c9, 0}, {0x30cf, 0x30d0, 0x30d1},
	{0x30d2, 0x30d3, 0x30d4}, {0x30d5, 0x30d6, 0x30d7}, {0x30d8, 0x30d9, 0x30da}, {0x30db, 0x30dc, 0x30dd}
};

Fl_SoftKeyboard_Youtube_JP::Fl_SoftKeyboard_Youtube_JP(int x, int y, int w, int h, const char* title)
	: Fl_SoftKeyboard(x, y, SoftKeyboardWidth, SoftKeyboardHeight, "SoftKeyboard")
	, mMode(HiraganaMode)
	, mLayout(0)
	, mLayoutImage(0)
	, mGroup(0)    
	, mFocus(0)    
	, mInput(0)
	, mInterrupted(false)
	, mBeSwitched(false)
	, mHiraStatus(HIRA_STATUS_UNINPUT)
	, mLastInputCharacter(0)
	, mTransStatus(0)
{
	mGroup = new Fl_Group(20,20,505,41);
	//mGroup->align(Fl_Align(FL_ALIGN_CENTER|FL_ALIGN_INSIDE));
	
	mInput = new Fl_InputKB(20, 20, 505, 41);
	mInput->box(FL_UP_BOX);
	mInput->textsize(20);

	mGroup->end();

	mInputHira = new Fl_InputKB(20, 26, 100, 28);
	mInputHira->box(FL_FLAT_BOX);
	mInputHira->textsize(20);
	mInputHira->textcolor(FL_WHITE);
	mInputHira->color(FL_BLUE);
	mInputHira->selection_color(FL_BLACK);
	mInputHira->setMaxWidth(505);
	mInputHira->setXLimit(20+505);
	mInputHira->hide();

	mInputKanji = new Fl_InputKB(20, 26, 100, 28);
	mInputKanji->box(FL_FLAT_BOX);
	mInputKanji->textsize(20);
	mInputKanji->color(FL_YELLOW);
	mInputKanji->setMaxWidth(505);
	mInputKanji->setXLimit(20+505);
	mInputKanji->hide();

	hide();
	clear_border();
	set_modal();
	changeMode(HiraganaMode);        
	end();
}

Fl_SoftKeyboard_Youtube_JP::~Fl_SoftKeyboard_Youtube_JP()
{
	if (mLayoutImage != NULL)
		mLayoutImage->release();

	if(mGroup != NULL)
		delete mGroup;

	if(mInputHira != NULL)
		delete mInputHira;

	if(mInputKanji != NULL)
		delete mInputKanji;
}

void Fl_SoftKeyboard_Youtube_JP::showSoftKeyboard(const int x, const int y)
{
	if (visible_r())
		return ;

	if(!KlbInit()){
		printf("init klb fail!\n");
		return;
	}

	mInterrupted = false;

	if (x != -1 || y != -1)
		this->position(x == -1 ? this->x() : x, y == -1 ? this->y() : y);

	//mGroup->hide();   mark this for Youtube

	this->show(); 
}

const char* Fl_SoftKeyboard_Youtube_JP::showSoftKeyboard(const char* initText, const bool isSecret, const int x, const int y)
{
	if (visible_r())
		return NULL;

	if(!KlbInit()){
		printf("init klb fail!\n");
		return NULL;
	}

	mInterrupted = false;

	if (x != -1 || y != -1)
		this->position(x == -1 ? this->x() : x, y == -1 ? this->y() : y);
	if (isSecret)
		mInput->input_type(FL_INPUT_READONLY | FL_SECRET_INPUT);
	else
		mInput->input_type(FL_INPUT_READONLY);   

	if (initText) {
		mInput->value(initText);
		mInput->position(strlen(initText));
	}
	else
		mInput->value("");    	

	mInput->box(box());
	mGroup->show();

	this->show();
	Fl_Window* g = Fl::grab();
	if (g)
		Fl::grab(this);

	while(visible_r() && !mInterrupted && !fl_application_exited()) {
		Fl::wait();
	}

	if (g)
		Fl::grab(g);

	this->hide();

	if (mInterrupted || fl_application_exited())
		return NULL;

	KlbFinish();
	return (mInput->value());
}

void Fl_SoftKeyboard_Youtube_JP::hideSoftKeyboard()
{
	if (!visible_r())
		return;

	KlbFinish();

	mInterrupted = true;
}

void Fl_SoftKeyboard_Youtube_JP::draw()
{
	if (damage() & ~FL_DAMAGE_CHILD) {
		//draw_box(box(), 0, 0, w(), h(), color()); // draw box with x/y = 0	
		if (mLayoutImage) {
			mLayoutImage->draw(0, 0);
		}

		if (mFocus != NO_FOCUS ) {        
			const LayoutRect& focusRect = KeyLayoutRect[mFocus];
			fl_color(0,255,255); //Set Focus color
			fl_rect(focusRect.x, focusRect.y, focusRect.w, focusRect.h);
		}
	}

	draw_children();
}

int Fl_SoftKeyboard_Youtube_JP::handle(int event)
{
	switch (event){
		case FL_KEYDOWN:
			//case FL_SHORTCUT:
			{
				int key = Fl::event_key();

				if (key == FL_Enter && mFocus != NO_FOCUS) {
					int scancode = mLayout->mMap[mFocus];
					if(mGroup->visible())
						processSoftKeyInput(scancode);
					else
						processWebWiewInput(scancode);	
				} else if (key == FL_Left) {
					if(mFocus==0){
						changeMode(-1);
					}else{
						setFocus(moveLeft(mFocus));                	
					}
				} else if (key == FL_Right) {
					if(mFocus==0){
						changeMode(1);
					}else{
						setFocus(moveRight(mFocus)); 
					}
				} else if (key == FL_Down) {
					setFocus(nextLine(mFocus));
				} else if (key == FL_Up) {
					setFocus(previousLine(mFocus));
				} else if (key == FL_Home) {
					this->hide();            
				} else {
					return Fl_Window::handle(event);
				}
				return 1;
			}
		case FL_PUSH:        
		case FL_MOVE:
			{
				int focus = mouseAt(Fl::event_x(), Fl::event_y());
				setFocus(focus);
				if (mFocus != NO_FOCUS && event == FL_PUSH) {
					int scancode = mLayout->mMap[mFocus];
					if(mGroup->visible())
						processSoftKeyInput(scancode);
					else
						processWebWiewInput(scancode);	
				}
				return 1;
			} 
		default:
			return Fl_Window::handle(event);	
	}
}

void Fl_SoftKeyboard_Youtube_JP::changeMode(int n)
{
	int mode = mMode + n;
	if(mode < 0)
		mode = 5;
	else if(mode > 5)
		mode = mode % 6;
	
	printf("switch mode = %d.\n", mode);

	if(mMode == HiraganaMode && mHiraStatus != HIRA_STATUS_UNINPUT){
		fixHira();
	}
	
	mMode = mode;

	if(mMode == HiraganaMode || mMode == KataKanaMode){
		if(mBeSwitched){
			printf("Switched Status.\n");
			mode += 6;
		}
	}
	mLayout = &LayoutStates[mode];
	if (mLayoutImage)
		mLayoutImage->release();

	mLayoutImage = Fl_ImageCache::getInstance()->get(mLayout->mImageFile);
	mFocus = 0;

	mLastInputCharacter = 0;

	redraw();
}

int Fl_SoftKeyboard_Youtube_JP::mouseAt(int x, int y)
{
	if (!mLayout)
		return NO_FOCUS;

	for(int i = 0; i <SCANCODES; i++) {
		const LayoutRect& rect = KeyLayoutRect[i];
		if (x >= rect.x 
				&& x < (rect.x + rect.w)
				&& y >= rect.y
				&& y < (rect.y + rect.h)
				&& mLayout->mMap[i] != 0){
			return i;
		}
	}
	return NO_FOCUS;
}

int Fl_SoftKeyboard_Youtube_JP::nextLine(int focus)
{
	if (focus == NO_FOCUS)
		return NO_FOCUS;

	if(mFocus == KB_CHARACTER)
		return 2; //Choose next focus for CharacterSet 

	const LayoutRect& focusRect = KeyLayoutRect[focus];
	const int centerX = focusRect.x + focusRect.w / 2;
	int f = mouseAt(centerX, focusRect.y + focusRect.h + focusRect.h / 2);
	if (f != NO_FOCUS){
		return f;
	}

	if(KB_SWITCH == focus){	// katakana model.
		return KB_RETURN;
	}
	else if(KB_RETURN == focus){
		if(mLayout->mMap[KB_SWITCH] != 0){
			return KB_SWITCH;
		}
		else {
			return focus;
		}
	}
	else if(KB_BACKSPACE == focus){
		return KB_COMP;
	}
	else {
		return KB_CHARACTER;
	}
}

int Fl_SoftKeyboard_Youtube_JP::previousLine(int focus)
{ 
	if (focus == NO_FOCUS)
		return NO_FOCUS;

	if(KB_SWITCH == focus){
		return KB_RETURN;
	}
	else if(KB_COMP == focus){
		return KB_BACKSPACE;
	}
	else if(KB_RETURN == focus){
		if(mLayout->mMap[KB_CONFIRM] != 0){
			return KB_CONFIRM;
		}
		else if(mLayout->mMap[KB_SWITCH] != 0){
			return KB_SWITCH;
		}
	}
	else if(KB_CHARACTER == focus){
		if(mLayout->mMap[38] != 0){
			return 38;
		}
		else if(mLayout->mMap[30] != 0){
			return 30;
		}
		else if(mLayout->mMap[24] != 0){
			return 24;
		}
		else if(mLayout->mMap[16] != 0){
			return 16;
		}
		else {
			return KB_CHARACTER;
		}
	}

	const LayoutRect& focusRect = KeyLayoutRect[focus];

	//Multi Lookup
	int f = NO_FOCUS;
	for(int i = 0; i < 2 ; i++)
	{
		const int centerX = focusRect.x + (focusRect.w / 2);
		f = mouseAt(centerX, focusRect.y - focusRect.h / 2  * (i + 1));
		if(f != NO_FOCUS)
			return f;
	}     
	return focus;
}

int Fl_SoftKeyboard_Youtube_JP::moveRight(int focus)
{ 
	if (focus == NO_FOCUS)
		return NO_FOCUS;

	// goto return key.
	if(KB_COMP == focus){
		if(mLayout->mMap[KB_SWITCH] != 0){
			return KB_SWITCH;
		}
		else {
			return KB_RETURN;
		}
	}
	else if(KB_CLEAR == focus){
		if(mLayout->mMap[KB_KANJI_CONFIRM] != 0){
			return KB_KANJI_CONFIRM;
		}
		else if(mLayout->mMap[KB_SWITCH] != 0){
			return KB_SWITCH;
		}
		else {
			return KB_RETURN;
		}
	}
	else if(KB_BACKSPACE == focus){
		return KB_RETURN;
	}
	else if(KB_RETURN == focus){
		if(mLayout->mMap[KB_SWITCH] != 0){
			return 30;
		}
		return 2;
	}
	else if(focus == 14){
		return KB_COMP;
	}
	else if(focus == 29){
		return KB_CLEAR;
	}
	else if(focus == 43){
		return KB_BACKSPACE;
	}
	else if(mMode == NumberMode && FL_Right == mLayout->mMap[focus]){
    		return KB_CLEAR;
    	}
    	else if(FL_Right == mLayout->mMap[focus]){
    		return KB_BACKSPACE;
    	}

	return focus + 1;
}

int Fl_SoftKeyboard_Youtube_JP::moveLeft(int focus)
{ 
	if (focus == NO_FOCUS)
		return NO_FOCUS;

	/// main key.
	if(2 == focus || 9 == focus){
		if(mLayout->mMap[KB_SWITCH] != 0){
			return KB_SWITCH;
		}
		else {
			return KB_RETURN;
		}
	}
	else if(16 == focus){
		if(mLayout->mMap[KB_KANJI_CONFIRM] != 0){
			return KB_KANJI_CONFIRM;
		}
		else if(mLayout->mMap[KB_SWITCH] != 0){
			return KB_SWITCH;
		}
		else {
			return KB_RETURN;
		}
	}
	else if(24 == focus){
		if(mLayout->mMap[KB_CONFIRM] != 0){		// HIRA
			return KB_CONFIRM;
		}
		else if(mLayout->mMap[KB_SWITCH] != 0){	// KATA
			return KB_SWITCH;
		}
		else {
			return KB_RETURN;
		}
	}
	else if(30 == focus){
		return KB_RETURN;
	}
	/// right key ...
	else if(KB_BACKSPACE == focus){
		if(mLayout->mMap[focus-1] == 0){
			return KB_RETURN;
		}
	}
	/// left key ...
	else if(KB_SWITCH == focus){
		return KB_COMP;
	}
	else if(KB_KANJI_CONFIRM == focus || KB_CONFIRM == focus){
		return KB_CLEAR;
	}
	else if(KB_RETURN == focus){
		return KB_BACKSPACE;
	}

	return focus - 1;	
}

void Fl_SoftKeyboard_Youtube_JP::processSoftKeyInput(int scancode)
{
	switch(scancode) {
		case FL_CHARACTERSET_KEY:
			changeMode(1);
			break;
		case FL_BackSpace:
			mLastInputCharacter = 0;
			if(mMode == HiraganaMode)
			{
				if(mHiraStatus == HIRA_STATUS_UNCHANGED){
					mInputHira->cutAndAdjustSize(-1);
					--mHiraSize;
					
					if(0 == mInputHira->size()){
						resetHira();
					}
					redraw();
					setFocus(KB_BACKSPACE);
					return;
				}
				else if(mHiraStatus == HIRA_STATUS_UNDECIDED){
					mHiraStatus = HIRA_STATUS_UNCHANGED;
					mInputKanji->hide();
					mInputHira->show();
					mInputHira->highlightLastCharacter();
					setFocus(KB_BACKSPACE);
					return;
				}
			}
			mInput->cutAsUTF8(-1);
			break;
		case FL_Left:
			if(mMode == HiraganaMode && mHiraStatus != HIRA_STATUS_UNINPUT){
				printf("move left fail for it si hira's undecide or unchanged status.\n");
				return;
			}	
			mInput->moveCursorAsUTF8(-1);
			break;
		case FL_Right:
			if(mMode == HiraganaMode && mHiraStatus != HIRA_STATUS_UNINPUT){
				printf("move left fail for it si hira's undecide or unchanged status.\n");
				return;
			}
			mInput->moveCursorAsUTF8(1);
			break;
		case FL_COMP_KEY:
			if(mMode == HiraganaMode && mHiraStatus != HIRA_STATUS_UNINPUT){
				fixHira();
			}
			this->hide();
			break;
		case FL_AllCLEAR_KEY:
			if(mMode == HiraganaMode && mHiraStatus != HIRA_STATUS_UNINPUT)
			{
				resetHira();
				setFocus(KB_CLEAR);
			}
			mInput->static_value(NULL);
			break;
		case FL_RETURN_KEY:
			hideSoftKeyboard();
			resetHira();
			break;
		case FL_SWITCH_KEY:
			if(mMode != HiraganaMode && mMode != KataKanaMode){
				printf("switch error.\n");
				break;
			}
			{
				int mode = mMode;
				if(mBeSwitched){
					mBeSwitched = false;
				}
				else {
					printf("Switched Status.\n");
					mode += 6;
					mBeSwitched = true;
				}
				mLayout = &LayoutStates[mode];
				if (mLayoutImage)
					mLayoutImage->release();
				mLayoutImage = Fl_ImageCache::getInstance()->get(mLayout->mImageFile);
				mFocus = KB_SWITCH;
				redraw();
			}
			if(mMode == HiraganaMode && mHiraStatus == HIRA_STATUS_UNDECIDED){
				fixHira();
			}
			break;
		case FL_KANJI_SWITCH:
			if(mMode == HiraganaMode)
			{
				const char * ret_str;
				if(mHiraStatus == HIRA_STATUS_UNCHANGED){
					if(0 == mInputHira->size()){
						printf("ignore kanji confirm for the size of input str is 0.\n");
						return;
					}
					if(NULL == (ret_str = convert(CONVERT_UTF8_TO_EUC, mInputHira->value()))){
						printf("utf8 to euc error.\n");
						return;
					}
					if(!KlbSetInputStr((char*)ret_str)){
						printf("KlbSetInputStr fail.\n");
						return;
					}
				}
				else if(mHiraStatus == HIRA_STATUS_UNINPUT){
					return;
				}
				
				if(!KlbConvNext()){
					printf("KlbConvNext fail.\n");
					return;						
				}
				
				const char * kanji = KlbGetOutputStr();
				if(kanji == NULL){
					printf("return for KlbGetOutputStr() == NULL.\n");
					return;
				}
				
				if(NULL == (ret_str = convert(CONVERT_EUC_TO_UTF8, kanji))){
					printf("euc to utf8 error.\n");
					return;
				}
				
				if(mHiraStatus == HIRA_STATUS_UNCHANGED){
					mInputHira->hide();
					mInputKanji->show();
					mInputKanji->move(mOffset);
					setFocus(KB_KANJI_CONFIRM);
				}
				
				mHiraStatus = HIRA_STATUS_UNDECIDED;
				mInputKanji->static_value(NULL);
				mInputKanji->insertAndAdjustSize(ret_str, strlen(ret_str));
				redraw();
			}
			break;
		case FL_CONFIRM:
			if(mMode == HiraganaMode)
			{
				fixHira();
			}
			break;
		// these 2 key are used as default key either.
		case FL_DAKUTEN:
		case FL_HANDAKUTEN:
			if(mMode == HiraganaMode || mMode == KataKanaMode)
			{
				if(mMode == HiraganaMode && mHiraStatus == HIRA_STATUS_UNDECIDED){
					printf("return for it is undecided status in HiraganaMode.\n");
					return;
				}
				int ret_code = transDakuten(scancode);
				if(ret_code <= 0){
					printf("get trans dakuten or handakuten is 0.\n");
					return;
				}
				printf("get transfered code = %x.\n", ret_code);
				char out[5];
				if(convertToUtf8(ret_code, out) < 0){
					return;
				}

				if(mMode == HiraganaMode){
					mInputHira->cutAsUTF8(-1);
					mInputHira->insertAndAdjustSize(out, strlen(out));
					mInputHira->highlightLastCharacter();
					redraw();
				}
				else {
					mInput->cutAsUTF8(-1);
					mInput->insert(out, strlen(out));
				}
				break;
			}
		default:
			{
				mLastInputCharacter = scancode;
				//printf("input %x.\n", scancode);
				mTransStatus = 0;
				
				if(mMode == HiraganaMode && 
					(scancode == 0x30FB || scancode ==  0xff0a ||scancode == 0xff01||
					 scancode == 0xff1f || scancode == 0x3000 || scancode == 0x30fc ))
				{
					if(mHiraStatus == HIRA_STATUS_UNDECIDED){
						printf("cann't input symbol character when hira status is unchanged or undecided.\n");
						return;
					}
					else if(mHiraStatus == HIRA_STATUS_UNINPUT){
						char out[5];
						if(convertToUtf8(scancode, out) < 0){
							return;
						}
						mInput->insert(out, strlen(out));
						return;
					}
				}
				
				// input jp characters.
				if(mMode == HiraganaMode)
				{
					if(mHiraStatus == HIRA_STATUS_UNDECIDED){
						fixHira();

						// mLastInputCharacter will be reset to 0 in fixHira(), so, reset it again.
						mLastInputCharacter = scancode;
						
						// force to update the x coordinate of cursor.
						mInput->updateCursorInfo();
					}
					
					if(mHiraStatus == HIRA_STATUS_UNINPUT){
						mHiraStatus = HIRA_STATUS_UNCHANGED;
						mOffset = 2+ mInput->get_cursor_x();
						mInputHira->move(mOffset);
						mInputHira->show();
						mInput->hide_cursor(1);
						mHiraSize = 0;
						printf("set HIRA input widget offset = %d.\n", mOffset);
					}
					
					if(mHiraSize == MAX_HIRA_SIZE){
						printf("cann't input for the size is more then %d.\n", MAX_HIRA_SIZE);
						return;
					}
					char out[5];
					if(convertToUtf8(scancode, out) < 0){
						return;
					}
					
					int size = strlen(out);
					++mHiraSize;
					mInputHira->insertAndAdjustSize(out, size);
					mInputHira->highlightLastCharacter();
					redraw();
					return;
				}
				
				char out[5];
				if(convertToUtf8(scancode, out) < 0){
					return;
				}
				mInput->insert(out, strlen(out));
			}
	}
}

void Fl_SoftKeyboard_Youtube_JP::processWebWiewInput(int scancode)
{  
	DFBWindowEvent dfbevent;
	memset(&dfbevent,0,sizeof(DFBWindowEvent));
	dfbevent.clazz = DFEC_WINDOW;
	dfbevent.type  = DWET_KEYDOWN;

	MainApplication* app = MainApplication::application();
	WebView* webView = app->webViewWindow()->webView();

	switch(scancode) {
		case FL_BackSpace:
			{
				dfbevent.key_symbol = DIKS_BACKSPACE;
				webView->onKeyDown((const DFBEvent&)dfbevent);
			}
			break;
		case FL_COMP_KEY:
			this->hide();
			break;
		case FL_AllCLEAR_KEY:
			mInput->static_value(NULL);
			break;
		case FL_RETURN_KEY:
			hideSoftKeyboard();
			break;                               
		default:
			{
				char t = (char)scancode;
				if (isprint(t)) {
					dfbevent.key_symbol = (DFBInputDeviceKeySymbol)t;
					webView->onKeyDown((const DFBEvent&)dfbevent);
				}
			}
	}
}

int Fl_SoftKeyboard_Youtube_JP::convertToUtf8(int value, char * dest )
{
	if( value <= 0x7f ){
		*dest = value;
		*(dest+1) = 0;
		return 1;
	}   
	else if( value <= 0x7ff ){
		*dest = ( value >> 6 ) | 0xC0;
		*(dest+1) = ( value & 0x3F ) | 0x80;
		*(dest+2) = 0;
		return 2;
	}   
	else if( value <= 0xffff ){
		*dest = ( value >> 12 ) | 0xE0;
		*(dest+1) = ( ( value >> 6 ) & 0x3F ) | 0x80;
		*(dest+2) = ( value & 0x3F ) | 0x80;
		*(dest+3) = 0;
		return 3;
	}   
	else {
		printf("convert to utf8 error .\n");
		return -1; 
	}   
}

void Fl_SoftKeyboard_Youtube_JP::resetHira()
{
	printf("reset Hiragana status.\n");
	mHiraStatus = HIRA_STATUS_UNINPUT;
	
	mInputHira->static_value(NULL);
	if(mInputHira->visible())
		mInputHira->hide();
	
	mInputKanji->static_value(NULL);
	if(mInputKanji->visible())
		mInputKanji->hide();
	
	mInput->hide_cursor(0);
	mInput->redraw();
	
	mLastInputCharacter = 0;
	mTransStatus = 0;
}

void Fl_SoftKeyboard_Youtube_JP::fixHira()
{
	printf("fix HIRA , current focus = %d.\n", mFocus);
	int focus = mFocus;
	const char * data = NULL;
	if(mHiraStatus == HIRA_STATUS_UNCHANGED){
		data = mInputHira->value();
	}
	else if(mHiraStatus == HIRA_STATUS_UNDECIDED){
		data = mInputKanji->value();
	}
	else {
		return;
	}
	if(data != NULL){
		mInput->insert(data, strlen(data));
	}
	resetHira();
	setFocus(focus);
}

int Fl_SoftKeyboard_Youtube_JP::transDakuten(int type)
{
	if(mLastInputCharacter == 0){
		printf("transDakuten fail for last input character is 0.\n");
		return 0;
	}
	printf("transDakuten (%x).\n", mLastInputCharacter);
	if(mTransStatus == type){
		printf("transDakuten return original char.\n");
		mTransStatus = 0;
		return mLastInputCharacter;
	}
	
	for(int i = 0; i < DAKUTEN_COUNT; ++i){
		if(mLastInputCharacter == DakutenTransTable[i].none){
			if(type == FL_DAKUTEN && DakutenTransTable[i].dakuten != 0){
				mTransStatus = type;
				return DakutenTransTable[i].dakuten;
			}
			else if(type == FL_HANDAKUTEN && DakutenTransTable[i].handakuten != 0){
				mTransStatus = type;
				return DakutenTransTable[i].handakuten;
			}
			else {
				printf("transDakuten fail for unsupport type.\n");
				return 0;
			}
		}
	}
	printf("transDakuten fail for there is not match.\n");
	return 0;
}

const char* Fl_SoftKeyboard_Youtube_JP::convert(ConvertType type, const char* src)
{
	if(NULL == src)
		return NULL;
	
	int src_len = strlen(src);
	int target_len = src_len * 3 + 1;
	char* target = new char[target_len];
	
	target_len = MainApplication::application()->webViewWindow()->webView()->textConvert(src, src_len, target, target_len , "EUC-JP", type);
	if(target_len <= 0){
		delete target;
		return NULL;
	}
	
	mEncodeBuf.assign(target, target_len);

	delete target;
	return mEncodeBuf.c_str();
}

