#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <sys/types.h>
#include <signal.h>
#include <unwind.h>
#include <dlfcn.h>
#define ucontext asm_ucontext
#include <asm/ucontext.h>
#ifdef VA_PLATFORM_X86
#include <execinfo.h>
#endif

#include <FL/fl_ask.H>
#include <FL/x.H>
#include <FL/Fl.H>
#include <string>
using std::string;

#include "MainApplication.h"
#include "Fl_DigitalClock.h"
#include "Fl_SoftKeyboard.h"
#include "Fl_MessageForm.h"
#include "Fl_WebViewWindow.h"
#include "Fl_MessageForm.h"
#include "Fl_ImageCache.h"
#include "Fl_MainMenu.h"
#include "mainUI.h"

#include "WebCookieJar.h"
#include "WebPreferencesInterface.h"
#include "History.h"
#include "Bookmark.h"

#include "cJSON.h"

static const double IDLE_TIMEOUT = 15.0;

static MainApplication* s_application = NULL;

// Used for quitting OWB when receiving a signal.
static void signalQuitCatcher(int /*signum*/)
{
    printf("signal quit received\n");
    fl_exit_application();
    Fl::awake();
}

static void signal_handle(int signum, siginfo_t* info, void*ptr) {
  //static const char *si_codes[3] = {"", "SEGV_MAPERR", "SEGV_ACCERR"};

  size_t i;
  struct ucontext *ucp = (struct ucontext*)ptr;
  
  void *bt[20];
  char **strings;
  size_t sz;
  
  Dl_info dlinfo; 
  int f = 0;
  void **bp = 0;
  void *ip = 0;

  //fprintf(stderr, "Segmentation Fault!\n");
  fprintf(stderr, "info.si_signo = %d\n", signum);
  fprintf(stderr, "info.si_errno = %d\n", info->si_errno);
  //fprintf(stderr, "info.si_code  = %d (%s)\n", info->si_code, si_codes[info->si_code]);
  fprintf(stderr, "info.si_addr  = %p\n", info->si_addr);
#ifndef VA_PLATFORM_X86  
  for(i = 0; i < NGREG; i++)
    fprintf(stderr, "reg[%02d]       = 0x%llx\n", i, ucp->uc_mcontext.sc_regs[i]);
  fprintf(stderr, "reg[pc] = 0x%llx\n", ucp->uc_mcontext.sc_pc);  

#if 0 //defined(VA_PLATFORM_X86)
  ip = (void*)ucp->uc_mcontext.gregs[REG_EIP];
  bp = (void**)ucp->uc_mcontext.gregs[REG_EBP];
#else
  ip = (void*)ucp->uc_mcontext.sc_pc;
  bp = (void**)ucp->uc_mcontext.sc_regs[31];
#endif

//#ifndef VA_PLATFORM_X86
  fprintf(stderr, "Stack trace:\n");
  while(bp && ip) {
    if(!dladdr(ip, &dlinfo))
      break;

    const char *symname = dlinfo.dli_sname;

    fprintf(stderr, "% 2d: %p <%s+%u> (%s)\n",
            ++f,
            ip,
            symname,
            (unsigned)((char*)ip - (char*)(dlinfo.dli_saddr)),
            dlinfo.dli_fname);

    if(dlinfo.dli_sname && !strcmp(dlinfo.dli_sname, "main"))
      break;

    ip = bp[1];
    bp = (void**)bp[0];
  }
#else
  fprintf(stderr, "Stack trace (non-dedicated):\n");
  sz = backtrace(bt, 20);
  strings = backtrace_symbols(bt, sz);

  for(i = 0; i < sz; ++i)
    fprintf(stderr, "addr = %p, %s\n", bt[i], strings[i]);
#endif
  fprintf(stderr, "End of stack trace\n");

  int pid = getpid();
  fprintf(stderr, "Process #%d received signal %d\n", pid, signum);
  fprintf(stderr, "Process #%d being suspended\n", pid);
  kill(pid, SIGSTOP); 
}

static int shortcut_handler(int event)
{
    if (FL_SHORTCUT == event && s_application)
        return s_application->shortcut_handler(event);
    return 0;
}

static void onTimeout(void* v)
{
    MainApplication* app = reinterpret_cast<MainApplication*>(v);
    if (app) {
        app->saveSettings();
        Fl::add_timeout(IDLE_TIMEOUT, onTimeout, app);
    }
}

MainApplication::MainApplication(int argc, char** argv)
{
    assert(s_application == NULL);
    s_application = this;

    fl_open_display();

    signal(SIGINT, &signalQuitCatcher);
    struct sigaction sa;
    sa.sa_sigaction = (void(*)(int, siginfo_t*,void*))(signal_handle);
    sa.sa_flags = SA_SIGINFO;
    sigemptyset(&sa.sa_mask);
    sigaction(SIGSEGV, &sa, NULL);

    mURL = NULL;
    mBrowserMode = false;
    for(int i = 1; i < argc; i++) {
        if (!strcmp("-browser", argv[i]))
            mBrowserMode = true;
        else
            mURL = argv[i];
    }

    mMainUI = new MainUI();
    memset(mMainUI, 0, sizeof(MainUI));
    if (!mBrowserMode) {
        mMainUI->buildFullScreenWindows();
    } else {
        mMainUI->buildWindows();
        mMainUI->mMainGroup->image(Fl_ImageCache::getInstance()->get("bg.png"));
    //{{
    //  fluid    
        mMainUI->mTitleBar->color(fl_rgb_color(0, 0, 0));
        mMainUI->mDigitalClock->color(fl_rgb_color(0, 0, 0));
        mMainUI->mMainMenu->color(fl_rgb_color(0, 0, 0));
    //}}
        Fl::add_handler(::shortcut_handler);
    }

    mCookieJar = WebCookieJar::webCookieJar();
    //mPref = WebPreferences::sharedStandardPreferences();
    mPref = WebPreferencesInterface::defaultInterface();
    mBookmark = new Bookmark();
    mHistory = new History();
    WebHistoryInterface::setDefaultInterface(mHistory);

    //Fl::add_timeout(IDLE_TIMEOUT, onTimeout, this);
}

MainApplication::~MainApplication()
{
    if (mMainUI->mWebViewWindow)
        delete mMainUI->mWebViewWindow;
    if (mMainUI->mMainWindow)
        delete mMainUI->mMainWindow;
    if (mMainUI->mSoftKeyboard)
        delete mMainUI->mSoftKeyboard;
    if (mMainUI->mMessageForm)
        delete mMainUI->mMessageForm;
    delete mMainUI;

    WebHistoryInterface::setDefaultInterface(NULL);
    delete mHistory;
    delete mBookmark;

    assert(s_application == this);
    s_application = NULL;
}

MainApplication* MainApplication::application()
{
    return s_application;
}

void MainApplication::onInit()
{
    Fl_WebViewWindow* viewWindow = webViewWindow();
    const char* uri;

    if (mBrowserMode) {
        loadSettings();
        viewWindow->hide();
        mainWindow()->show();

        if (!mURL && softKeyboard())
            mURL = softKeyboard()->showSoftKeyboard("http://",false,207,88); //be placed in the middle
    }
    uri = mURL;
    if (uri == NULL)
        uri = "http://www.google.com";

    printf("uri = %s \n", uri);
    viewWindow->show();

    viewWindow->webView()->mainFrame()->loadURL(uri);   
    Fl::focus(viewWindow);
}

void MainApplication::onExit()
{
    if (mBrowserMode)
        saveSettings();
}

int MainApplication::run()
{
    int result = 0;

    onInit();

    while(!fl_application_exited())
        result = Fl::wait(0.1);

    onExit();

    return result;
}

Fl_Window* MainApplication::mainWindow()
{
    return mMainUI->mMainWindow;
}

Fl_Group* MainApplication::mainGroup()
{
    return mMainUI->mMainGroup;
}

Fl_WebViewWindow* MainApplication::webViewWindow()
{
    return mMainUI->mWebViewWindow;
}

Fl_SoftKeyboard* MainApplication::softKeyboard()
{
    return mMainUI->mSoftKeyboard;
}

Fl_Message_Form* MainApplication::messageForm()
{
    return mMainUI->mMessageForm;
}

Fl_Box* MainApplication::titleBar()
{
    return mMainUI->mTitleBar;
}

Fl_Box* MainApplication::statusBar()
{
    return mMainUI->mStatusBar;
}

Fl_MainMenu* MainApplication::mainMenu()
{
    return mMainUI->mMainMenu;
}

void MainApplication::updateTitle(const char* title)
{
    if (mMainUI->mTitleBar)
        mMainUI->mTitleBar->label(title);
}

void MainApplication::updateStatus(const char* status)
{
    if (mMainUI->mStatusBar)
        mMainUI->mStatusBar->label(status);
}

bool MainApplication::messageBox(const char* fmt, ...)
{
   bool r;    
   va_list ap;

   va_start(ap, fmt);
   if (mMainUI->mMessageForm)
       r = mMainUI->mMessageForm->ask(fmt, ap);
   else {
       vprintf(fmt, ap);
       r = true;
   }

   va_end(ap);

   return r;
   
}

void MainApplication::messageInfo(const char* fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    if (mMainUI->mMessageForm)
        mMainUI->mMessageForm->message(fmt, ap);
    else
        vprintf(fmt, ap);
    va_end(ap);
}

int MainApplication::shortcut_handler(int event)
{
    if (Fl::event_key() == FL_Home && mMainUI->mMainMenu)
        return mMainUI->mMainMenu->popupMenu();   
    return 0;
}

/**********for rodger test****************
void creat_content_objects(char** out)
{
	cJSON *content_root,*cjHistory,*cjBookmark,*cjCookies,*cjOtherSettings,*Item;
	content_root = cJSON_CreateObject();

	cJSON_AddItemToObject(content_root, "History", cjHistory = cJSON_CreateArray());
	
	
	//for histroy
	char *Title[5] = {"Google","Baidu","Sina","Yahoo","MSN"};
	char *URL[5] = {"http://www.google.com.cn","http://www.baidu.com","http://www.sina.com.cn","http://www.yahoo.com.cn","http://www.msn.com"};
	char *Keys[7] = {"http://www.directfb.org","http://www.ubuntu.org","http://www.wikipedia.org","http://www.fltk.org",
		             "http://www.json.org","http://www.opensource.org","http://nba.tom.com"};
	char *Cookies[7] = {"Name:rodger,Password:123456","Name:rodger,Password:123456","Name:rodger,Password:123456","Name:rodger,Password:123456",
					"Name:rodger,Password:123456","Name:rodger,Password:123456","Name:rodger,Password:123456"};	             

	for(int i = 0; i < 5 ; i++)
	{
		cJSON_AddItemToArray(cjHistory,Item=cJSON_CreateObject());
		cJSON_AddStringToObject(Item, "Title", Title[i]);
		cJSON_AddStringToObject(Item, "URL", URL[i]);
	}
	
	
	cJSON_AddItemToObject(content_root, "Bookmark", cjBookmark = cJSON_CreateArray());
	
	for(int i = 0; i < 5 ; i++)
	{
		cJSON_AddItemToArray(cjBookmark,Item=cJSON_CreateObject());
		cJSON_AddStringToObject(Item, "Title", Title[i]);
		cJSON_AddStringToObject(Item, "URL", URL[i]);
	}
	
	cJSON_AddItemToObject(content_root, "Cookies", cjCookies = cJSON_CreateArray());
	for(int i = 0; i < 7 ; i++)
	{
		cJSON_AddItemToArray(cjCookies,Item=cJSON_CreateObject());
		cJSON_AddStringToObject(Item, "URL", Keys[i]);
		cJSON_AddStringToObject(Item, "Cookie", Cookies[i]);
	}
	
	cJSON_AddItemToObject(content_root, "OtherSettings", cjOtherSettings = cJSON_CreateObject());
	cJSON_AddStringToObject(cjOtherSettings,"Cookies:isEnabled","true");
	cJSON_AddStringToObject(cjOtherSettings,"JavaScript:isEnabled","false");
	
	*out=cJSON_Print(content_root);
	cJSON_Delete(content_root);	
	printf("%s\n",*out);	
	//free(out);		
}
*/

/**********for rodger test****************
void tempfunc(char* pathname)
{	
	FILE *fp;
	char *root_content;
	
	creat_content_objects(&root_content);
	
	if(fp = fopen(pathname,"w+"))
	{
		if(root_content == NULL)
		{
			fclose(fp);
			return;
		}	
		
		fputs(root_content, fp);
		fclose(fp);
	}else{
		printf("Failed2 to loadSettings: pathname = %s \n",pathname);
	}
	
	free(root_content);
}
*/
void MainApplication::processSettingsData(char* root_content)
{
	cJSON *cjRoot,*cjHistory,*cjBookmark,*cjCookies,*cjOtherSettings,*Item;
	
	cjRoot=cJSON_Parse(root_content);

    if (cjRoot == NULL) {
        printf("failed to parse setting data\n");
        return;
    }
	
	if((cjHistory = cJSON_GetObjectItem(cjRoot,"History")) && cjHistory->type == cJSON_Array)
	{
		//process history information
		int arraySize = cJSON_GetArraySize(cjHistory);
		if(NULL != mHistory)
		{
			for(int i = 0 ; i < arraySize ; i++)
			{
				Item = 	cJSON_GetArrayItem(cjHistory,i);
				mHistory->addHistoryEntry(cJSON_GetObjectItem(Item,"URL")->valuestring,cJSON_GetObjectItem(Item,"Title")->valuestring);
			}
		}
	}
	
	if((cjBookmark = cJSON_GetObjectItem(cjRoot,"Bookmark")) && cjBookmark->type == cJSON_Array)
	{
		int arraySize = cJSON_GetArraySize(cjBookmark);
		if(NULL != mBookmark)
		{
			for(int i = 0 ; i < arraySize ; i++)
			{
				Item = 	cJSON_GetArrayItem(cjBookmark,i);
				mBookmark->addItem(cJSON_GetObjectItem(Item,"Title")->valuestring,cJSON_GetObjectItem(Item,"URL")->valuestring);
			}
		}
	}	
	
	if((cjCookies = cJSON_GetObjectItem(cjRoot,"Cookies")) && cjCookies->type == cJSON_Array)
	{
		int arraySize = cJSON_GetArraySize(cjCookies);
		if(NULL != mCookieJar)
		{
			for(int i = 0 ; i < arraySize ; i++)
			{
				Item = 	cJSON_GetArrayItem(cjCookies,i);
				mCookieJar->setCookies(cJSON_GetObjectItem(Item,"URL")->valuestring,cJSON_GetObjectItem(Item,"Cookie")->valuestring);
			}
		}
	}

	if(cjOtherSettings = cJSON_GetObjectItem(cjRoot,"OtherSettings"))
	{	
		if(NULL != mCookieJar)
		{
			char* value = cJSON_GetObjectItem(cjOtherSettings,"Cookies:isEnabled")->valuestring;
            if (value)
            {
			    if(strcmp(value,"false") == 0)
				    mCookieJar->setEnabled(false);
			    if(strcmp(value,"true") == 0)
				    mCookieJar->setEnabled(true);	
            }
		}
		if(NULL != mPref)
		{
			char* value = cJSON_GetObjectItem(cjOtherSettings,"JavaScript:isEnabled")->valuestring;
            if (value)
            {
			    if(strcmp(value,"false") == 0)
				    mPref->setJavaScriptEnabled(false);
			    if(strcmp(value,"true") == 0)
				    mPref->setJavaScriptEnabled(true);	
            }
        }		
	}
	
	//char *out = cJSON_Print(cjRoot);
	cJSON_Delete(cjRoot);
	//printf("%s\n",out);	
}

void MainApplication::getSettingData(char** content_root)
{
	cJSON *cjRoot,*cjHistory,*cjBookmark,*cjCookies,*cjOtherSettings,*Item;
	cjRoot = cJSON_CreateObject();
	
	if(NULL != mHistory && mHistory->countItems())
	{
		cJSON_AddItemToObject(cjRoot, "History", cjHistory = cJSON_CreateArray());
		
		for(int i = 0; i < mHistory->countItems(); i++)
		{
			string Title,URL;
			mHistory->itemAt(i, Title, URL);
			cJSON_AddItemToArray(cjHistory,Item=cJSON_CreateObject());
			cJSON_AddStringToObject(Item, "Title", Title.c_str());
			cJSON_AddStringToObject(Item, "URL", URL.c_str());
		}
	}

	if(NULL != mBookmark && mBookmark->countItems())
	{
		cJSON_AddItemToObject(cjRoot, "Bookmark", cjBookmark = cJSON_CreateArray());
		
		for(int i = 0; i < mBookmark->countItems(); i++)
		{
			string Title,URL;
			mBookmark->itemAt(i, Title, URL);
			cJSON_AddItemToArray(cjBookmark,Item=cJSON_CreateObject());
			cJSON_AddStringToObject(Item, "Title", Title.c_str());
			cJSON_AddStringToObject(Item, "URL", URL.c_str());
		}
	}
	
	if(NULL != mCookieJar)
	{
		vector<string> keys;
		/***********bug here**************/		
		mCookieJar->copyKeysToVector(keys);
		if(keys.size())
		{

			cJSON_AddItemToObject(cjRoot, "Cookies", cjCookies = cJSON_CreateArray());
			for(int i = 0; i < keys.size(); i++)
			{
				cJSON_AddItemToArray(cjCookies,Item=cJSON_CreateObject());
				cJSON_AddStringToObject(Item, "URL", keys[i].c_str());
				cJSON_AddStringToObject(Item, "Cookie", (mCookieJar->cookies(keys[i])).c_str());
			}
			
		}
		
	}
	
	//creat object for other setttings
	cJSON_AddItemToObject(cjRoot, "OtherSettings", cjOtherSettings = cJSON_CreateObject());
	if(NULL != mCookieJar)
	{
		if(mCookieJar->isEnabled())
			cJSON_AddStringToObject(cjOtherSettings,"Cookies:isEnabled","true");
		else	
			cJSON_AddStringToObject(cjOtherSettings,"Cookies:isEnabled","false");	
	}
	
	if(NULL != mPref)
	{
		if(mPref->isJavaScriptEnabled())
			cJSON_AddStringToObject(cjOtherSettings,"JavaScript:isEnabled","true");
		else	
			cJSON_AddStringToObject(cjOtherSettings,"JavaScript:isEnabled","false");	
	}
	
	*content_root = cJSON_Print(cjRoot);
	cJSON_Delete(cjRoot);		
	//printf("%s\n",*content_root);	
}

void MainApplication::loadSettings()
{			
    char pathname[PATH_MAX];
    FILE *fp;
    char *root_content;
    snprintf(pathname, PATH_MAX, "%s/bin/%s", getenv("VA_RUNTIME_ROOT"), "webkit.json");
    
    /**********for rodger test****************
    char *root_test;
    tempfunc(pathname);
    **/
    
    //read setting information from webkit.json
    if(fp = fopen(pathname,"rb"))
    {
    	fseek(fp,0,SEEK_END); 
		long length=ftell(fp);
		
		if(0 == length)
		{
			printf("Warning: webkit.json is NULL\n");
			fclose(fp);
			return;	
		}
		
		fseek(fp,0,SEEK_SET);
		
		//get all data from file
		root_content = (char *)malloc(length + 1);
		fread(root_content, 1, length, fp);
		fclose(fp);
    }else{
    	printf("Failed to loadSettings: pathname = %s \n",pathname);
        return;
    }
    
    processSettingsData(root_content);
    
    free(root_content);
    /**********for rodger test****************
    getSettingData(&root_test);*/
}

void MainApplication::saveSettings()
{
	//creat_content_objects();
	char pathname[PATH_MAX];
	FILE *fp;
	char *root_content;
	
	//get current setting data
	getSettingData(&root_content); 
    if (root_content == NULL)
    {
        printf("faile to get setting data\n");
        return;
    }
	
	snprintf(pathname, PATH_MAX, "%s/bin/%s", getenv("VA_RUNTIME_ROOT"), "webkit.json"); 

	if(fp = fopen(pathname,"w+"))
	{
		fputs(root_content, fp);
		fclose(fp);
	}else{
		printf("Failed to saveSettings: pathname = %s \n",pathname);
	}
	
	free(root_content);
}


